/************************************************************************
 *  (C)opyright 1996/1997 by Khamis Abu El-Komboz                       *
 *  xkhfm 1.0 can be copied and distributed freely for any              *
 *  non-commercial purposes. xkhfm can only be incorporated             *
 *  into commercial software with the permission of the current author. *
 *  Mail: khamis@multix.de                                              *
 ************************************************************************/
/*
 * $Author: khamis $
 * $Id: docdrom.c,v 1.1 1997/01/07 08:39:42 khamis Exp $
 * $Source: /home/khamis/source/xkhfm/khfm/RCS/docdrom.c,v $
 */

#ifdef HAVE_CDROM

#ifndef TEST_CDROM
#define EXTERN extern
#else
#define EXTERN
#endif

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/cdrom.h>
#include <sys/cd_ioctl.h>

#include "dodefs.h"
#include "doglobal.h"
#include "doscr.h"
#include "docfg.h"
#include "doframe.h"
#include "dochoose.h"
#include "domem.h"
#include "docdrom.h"

/* Define the starting track and the device used for accessing the cdrom */

/*
 * Open the device
 */
int cdrom_open (char * cdrom_dev)
{
  return open(cdrom_dev, O_RDONLY);
}

int cdrom_close (int cdrom)
{
  close (cdrom);
}

/*
 * stop playing the cdrom
 */
int cdrom_stop (int cdrom)
{
  return ioctl (cdrom, C_STOPUNIT, NULL);
}

/*
 * avail. tracks on the CD
 */
int cdrom_Tracks (int cdrom)
{
  struct cdrom_audio cd_audio;
  int tracks=0;
  int ret_;

  if (ioctl (cdrom, C_STARTUNIT, NULL) < 0) return -1;
  if (ioctl (cdrom, C_REZERO,    NULL) < 0) return -1;

  cd_audio.addr_track = trackProgram[tracks];
  cd_audio.play       = 0;
  cd_audio.type    = 2;

  while(ret_ > -1)
    {
      if (ioctl (cdrom, C_AUDIOSEARCH, &cd_audio) < 0)
	break;
      tracks ++;
      cd_audio.addr_track = trackProgram[tracks];
    }
  return tracks;
}

/* Kick the disk out */
int cdrom_eject (int cdrom)
{
  return ioctl (cdrom, C_TRAYOPEN, NULL);
}

/*
 * Select the track to play 
 */
int cdrom_play (int cdrom, int trackNum)
{
  struct cdrom_audio cd_audio;
  cd_audio.play = 0;
  cd_audio.type = 2;
  cd_audio.addr_track = trackProgram [trackNum-1];
  
  if (ioctl (cdrom, C_AUDIOSEARCH, &cd_audio) < 0)
    return FALSE;
  
  cd_audio.play=3;
  cd_audio.type=3;
  
  if (ioctl (cdrom, C_PLAYAUDIO, &cd_audio) < 0)
    return FALSE;	
  return TRUE;
}

/* Pause the cd-rom */
int cdrom_pause (int cdrom)
{
  return ioctl (cdrom, C_STILL, NULL);
}

/*
 * Resume playing the cd-rom
 */
int cdrom_resumePlay (int cdrom)
{
  struct cdrom_audio cd_audio;
  cd_audio.play=3;
  cd_audio.type=2;
  
  return ioctl (cdrom, C_PLAYAUDIO, &cd_audio);
}

void cdrom_playing (void)
{
  struct frame_t * f, * wait_f;
  struct input_t * in;
  struct log_t   * lg [2];
  char ** tracks;
  int cdrom = -1, trackNum = -1, i, ret_;
  int width = min (60, max_x-2),
      down  = min (5, max_y-4),
      x     = (max_x-width) / 2,
      y     = (max_y-down ) / 2;

  wait_f = make_frame (x, y, width, down, leer, leer, F_B_UTIL_CHS, NOMODE);
  f_putstr (wait_f, 2, 2, " Reading the songs in the CD, please wait .... ",
	    wait_f->color, NOMODE);
  view_frame (wait_f); fflush_();

  cdrom = cdrom_open (CDROM_DEV);
  if (cdrom >= 0) trackNum = cdrom_Tracks (cdrom);

  hide_frame (wait_f, HIDE_MODE|DELETE_MODE);

  if (cdrom < 0)
    {
	message (" Can't open the CD-ROM device, ;;"
		 " pleace check your device ",
		 " Opining CD-ROM ");
    }
  else if (trackNum <= 0)
    {
	message (" No Soungs found at this CD ", " Search Songs ");
    }

  if (trackNum <= 0) return;

  width = min (60, max_x-2);
  x     = (max_x-width) / 2;
  down  = 7 + trackNum / 6;
  y     = (max_y-down) / 2;
  f = make_frame (x, y, width, down, " CD-Player ", leer,
		  F_B_UTIL_CLR,
		  NOMODE);
  f_putstr  (f, 1, 1, "Songs:", f->color, NOMODE);
  f_putline (f, f->down-1, f->color);
  lg [0] = make_log (f, trackNum, F_B_UTIL_CHS);

  tracks = (char **) getmem (sizeof (char *) * (trackNum + 1));
  for (i=0; i<trackNum; i++)
    {
	char tmp[20];
	sprintf (tmp, " %02i ", i + 1);
	tracks [i] = getmem (strlen(tmp) + 1);
	strcpy (tracks [i], tmp);

        lg [0]->fmt [i] = tracks [i];
        lg [0]->x   [i] = 10 + (i % 6) * 8;
        lg [0]->y   [i] = 2  + i / 6;
    }
  tracks [trackNum] = NULL;

  lg[1] = make_log (f, 4, F_B_UTIL_CHS);
  log_format (lg [1], "play", "stop", "eject", "break");
  log_xy     (lg [1], -1, f->down, -1, f->down, -1, f->down, -1, f->down);
  
  in = make_input (f, 2, CUR_MODE);
  input_log (in, 0, lg [0]);
  input_log (in, 1, lg [1]);
  
  for (ret_ = -1; ret_ != ESC; )
    {
      view_frame (f);
      ret_ = f_getstr (in);
      
      if (ret_ != ESC) switch (lg[1]->pos)
	{
	case 0: cdrom_play  (cdrom, lg[0]->pos + 1); break;
	case 1: cdrom_stop  (cdrom); break;
	case 2: cdrom_eject (cdrom); break;
	case 3: break;
	}
    }
  
  for (i=0; i<trackNum; i++) free (tracks [i]);
  free (tracks);
  
  hide_frame (f, HIDE_MODE|DELETE_MODE);
  del_log    (lg [0]);
  del_log    (lg [1]);
  del_input  (in);
}

#ifdef TEST_CDROM
main ()
{
  int cdrom = cdrom_open  (CDROM_DEV);
  if (cdrom < 0) { perror (CDROM_DEV); exit (1); };
  cdrom_play  (cdrom, 1);
  cdrom_close (cdrom);
}
#endif

#endif /* HAVE_CDROM */

