/************************************************************************
 *  (C)opyright 1996/1997 by Khamis Abu El-Komboz                       *
 *  xkhfm 1.0 can be copied and distributed freely for any              *
 *  non-commercial purposes. xkhfm can only be incorporated             *
 *  into commercial software with the permission of the current author. *
 *  Mail: khamis@multix.de                                              *
 ************************************************************************/
/*
 * $Author: khamis $
 * $Id: dofiles.c,v 1.1 1997/01/07 08:39:42 khamis Exp $
 * $Source: /home/khamis/source/xkhfm/khfm/RCS/dofiles.c,v $
 */

#define GLOBAL extern
#define EXTERN

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>

#ifdef alpha
#include <unistd.h>
#endif

#ifdef DOS
#include <io.h>
#include <dir.h>
#endif

#if defined (HAVE_REALPATH) && defined (alpha) && ! defined (HAVE_READLINK)
#define HAVE_READLINK
#endif
#if defined (HAVE_READLINK) && ! defined (HAVE_REALPATH) 
#define HAVE_REALPATH
#endif

#include "dodefs.h"
#include "doglobal.h"
#include "dolib.h"
#include "dodirent.h"
#include "doscr.h"
#include "docfg.h"
#include "doutil.h"
#include "dostring.h"
#include "dofiles.h"

int exist( char * file)
{
  struct stat st;
  if( MY_stat( file, &st) != 0) return( FALSE);
  return( TRUE);
}

int is_dir (char * dr, int mode)
{
  struct stat st;
  /*
   * mode : DIR_SYMLINK, if file is symbolic link, locks for it's realname and
   *        return its type
   * otherwise for symbolic links will be symbolic type returned
   */
  if( MY_stat( dr, &st) == -1) return( FALSE);
  
  if (S_ISDIR (st.st_mode)) return TRUE ;
  
#ifdef HAVE_REALPATH
  if ((mode & DIR_SYMLINK) && S_ISLNK(st.st_mode))
    {
      char tmp [MAXPATH] ;
#ifdef HAVE_READLINK
      readlink (dr, tmp, sizeof (tmp));
      if (is_dir (tmp, FALSE)) return TRUE ;
#else
      if (is_dir (realpath (dr, tmp), FALSE)) return TRUE ;
#endif
    }
#endif
  
  return FALSE ;
}

int is_file (char *file, int mode)
{
  /* mode = TRUE  : directory isn't a LINK SYMBOL   */
  /* mode = FALSE : directory can a LINK to another */
  struct stat st;
  if (MY_stat( file, &st) != 0) return FALSE ;

#ifdef UNIX
  if (S_ISREG (st.st_mode))
    if (mode && S_ISLNK (st.st_mode)) return FALSE ;
    else return TRUE ;
#else
  if (S_ISREG (st.st_mode)) return TRUE ;
#endif
  return FALSE ;
}

int file_owner( char * file) {
    struct stat st;
    if( MY_stat( file, &st) == -1) return( FALSE);
    return( st.st_uid);
    }

int file_grp( char * file) {
    struct stat st;
    if( MY_stat( file, &st) == -1) return( FALSE);
    return( st.st_gid);
    }

int can_write( char * file) {
    struct stat st;
    if( MY_stat( file, &st) != 0) return( FALSE);
    return( C_W( st.st_uid, st.st_gid, st.st_mode) );
    }

int can_read( char * file) {
    struct stat st;
    if( MY_stat( file, &st) != 0) return( FALSE);
    return( C_R( st.st_uid, st.st_gid, st.st_mode) );
    }

int can_exec( char * file) {
    struct stat st;
    if( MY_stat( file, &st) != 0) return( FALSE);
    return( C_X( st.st_uid, st.st_gid, st.st_mode) );
    }

ushort get_stat( char * file) {
    struct stat st;
    if( MY_stat( file, &st) != 0) return( FALSE);
    return( st.st_mode);
    }

int get_umask (int mode) 
{
  mode_t msk = umask (0777) & 0777; /* Only 8 bits */
  umask ((short) msk);              /* set the umask to it's old value */
#ifdef UNIX
  return (mode ? msk | 0111 : msk);
#else
  return (msk);
#endif
}

/***********************************************************************/
/******** test if a file is ascii-text, data or executable file ********/
/***********************************************************************/

#define FILEBUF	512

int file_type( char * file) {
    unsigned char buf[ FILEBUF];
    int file_h,
	i,
	len,
	not_ascii = 0,
	is__ascii = 0;

    if( can_exec( file)) return( EXEC_FILE);   /* file can execute */

    file_h = open( file, O_RDONLY|O_BINARY);
    if( file_h < 0) return( FALSE);

    len = read( file_h, buf, FILEBUF);

    for( i=0; i<len; i++) {
	if( is_ascii( buf[i] & 0x00ff) ||
	    buf[i] == ENTER  ||
	    buf[i] == CTRL_M ||
	    buf[i] == TAB    ||
	    buf[i] == CTRL_L )
	     is__ascii++;
	else {
	    not_ascii++;
	    break;
	    }
	}
    close( file_h);

    return ( not_ascii > 0 ? DATA_FILE : TEXT_FILE);
    }

long file_size (char * file, int mode)
{
  int files = 0, dirs = 0;
  return file_size_count (file, &files, &dirs, NULL, mode);
}

long file_size_count (char * file,
		      int  * files, 
		      int  * dirs,
		      struct frame_t * fr,
		      int mode)
{

  /***** mode = RECURSIVE_MODE : recursive size <for directories> ******/
  /***** files, is the number of access files/directories	    */
  
  struct stat st;
  long size = 0;
  
  if (MY_stat( file, &st) != 0) return 0;  /* error */
  
  if (S_ISREG( st.st_mode))  		/* only when a regular file */
    {
      (* files) ++;
      return st.st_size;
    }

  if( ! (mode & RECURSIVE_MODE) ) return (0);  /* not recursive size */
  
  if (S_ISDIR( st.st_mode))  /* size of the directory */
    {    
      DIR * dir;
      struct dirent * d_file;
      char tmp[MAXPATH];
      
      if (fr != NULL) 
	{
	  char * p = basename (file);
	  f_putstr (fr, (fr->width - strlen(p))/2, 4, p,
		    fr->color,
		    FILL_MODE|SCR_MODE);
	  fflush_();
	}
      if( (dir = X_opendir( file)) == NULL) return 0; /* error */
      
      dirs[0] ++;				/* increase directories */
      
      while( ( d_file = X_readdir( dir)) != NULL)
	{
	  
	  if (strcmp( d_file->d_name, "." ) == 0 ||
	      strcmp( d_file->d_name, "..") == 0
	      ) continue;
	  
	  sprintf( tmp, "%s/%s", file, d_file->d_name);
	  size += file_size_count (tmp, files, dirs, fr, mode); /* recursive */
	}
      
      X_closedir( dir);
      if (fr != NULL)
	{
	  f_putstr (fr, 1, 4, leer, fr->color, FILL_MODE|SCR_MODE);
	}
    }
  
  return( size);
}
    
int create_dir( char * dr, int mode)
{
  int err_;
  if (is_dir (dr, DIR_SYMLINK)) err_ = FALSE;      /* nothing to do */
  else
    {
#ifdef UNIX
#   ifndef XENIX
      err_ = mkdir (dr, (short) mode);
#   else
      err_ = TRUE;
#   endif
#else
      err_ = mkdir (dr);
#endif
    }
  return (err_);
}

long file_goto( long size, long akt_pos) {
     struct frame_t * f;
     struct input_t * in;
     char str[21];

     int width = min(max_x, 40),
	 down  = 5,
	 x     = (max_x-width)/2,
	 y     = (max_y-down )/2,
	 lk;

     f  = make_frame( x, y, width, down, " Goto ", leer, F_B_DIALOG_CLR, NOMODE);
     in = make_input( f, 1, CUR_MODE|INS_MODE|TYPE_INT);

     f_putstr( f, 1, 2, "Position:", f->color, NOMODE);
     in->x      [0] = 11;
     in->y      [0] = 2;
     in->str    [0] = str;		  /* lock for this string */
     in->str_len[0] = 20;
     in->inp_len[0] = 20;
     sprintf( str, "%li", akt_pos+1);          /* aktual position */

     view_frame( f);
     lk  = f_getstr( in);
     if( lk == ENTER || lk == CTRL_M) {
	 akt_pos = atol(str) - 1;
	 akt_pos = max( 0, min( size-1, akt_pos));
	 }
     del_input ( in);
     hide_frame( f, HIDE_MODE|DELETE_MODE);

     return( akt_pos);
     }

