/************************************************************************
 *  (C)opyright 1996/1997 by Khamis Abu El-Komboz                       *
 *  xkhfm 1.0 can be copied and distributed freely for any              *
 *  non-commercial purposes. xkhfm can only be incorporated             *
 *  into commercial software with the permission of the current author. *
 *  Mail: khamis@multix.de                                              *
 ************************************************************************/
/*
 * $Author: khamis $
 * $Id: dogrep.c,v 1.1 1997/01/07 08:39:42 khamis Exp $
 * $Source: /home/khamis/source/xkhfm/khfm/RCS/dogrep.c,v $
 */

#define GLOBAL extern
#define EXTERN

#include <stdlib.h>
#include <string.h>

#include "dodefs.h"
#include "doglobal.h"
#include "doerror.h"
#include "domem.h"
#include "doscr.h"
#include "dostat.h"
#include "dostring.h"
#include "dotext.h"
#include "docfg.h"
#include "dogrep.h"


/**********************************************************************/
/* grep for a pattern in a list of files <use the 'grep'-command>  		      	      */
/**********************************************************************/

#define FILE_LEN	256
#define PATH_LEN	256
#define FLAG_LEN         80
#define G_BLOCK_LEN	512

/* grep for any context in a file list				 */
/* OUTPUT:   NULL, if no grep or error 			         */
/*           voll a static Pathname with file name               */

char * grep_cmd( void)
{
  struct frame_t * f; /* frame to view the found files allowes the grep-command */
  struct input_t * in;
  struct log_t   * lg;
  static char fnd_path [PATH_LEN];      /* found path */
  
  char get_name[ FILE_LEN] = { 0 },
  get_path[ PATH_LEN] = { 0 },
  get_flag[ FLAG_LEN] = { 0 },
  tmp[ FILE_LEN+PATH_LEN+FLAG_LEN] = { 0 },
  * tmp_file,
  * p = NULL;                      /* pointer to a static pathname */
  int width   = min( 75, max_x),
  down    = min( 20, max_y),
  y       = (max_y-down) / 2,
  x       = (max_x-width) / 2,
  file_h,
  i, j, lk,
  ret_,
  err_ = FALSE;
  
  f = make_frame( x, y, width, down, " Grep ", leer, F_B_UTIL_CLR,
		 NOMODE);
  
  /****************************************************************** 
    INPUT : 1. Pattern
    2. File list
    3. Flags and arguments 
    4. Continue/Break
    *******************************************************************/
  
  in = make_input( f, 4, CUR_MODE | FIL_MODE | INS_MODE);
  lg = make_log  ( f, 2, F_B_UTIL_CHS);
  
  f_putline( f, f->down - 7, f->color);
  f_putstr ( f, 1, f->down - 6, "Pattern  :", f->color, NOMODE);
  f_putline( f, f->down - 5, f->color);
  f_putstr ( f, 1, f->down - 4, "File List:", f->color, NOMODE);
  f_putline( f, f->down - 3, f->color);
  f_putstr ( f, 1, f->down - 2, "Flags    :", f->color, NOMODE);
  f_putline( f, f->down - 1, f->color);
  
  log_format (lg, cont_break[0], cont_break[1]);
  log_xy     (lg, -1, f->down, -1, f->down);
  
  input_string (in, 0, 12, f->down - 6, get_name, FILE_LEN, FILE_LEN);
  input_string (in, 1, 12, f->down - 4, get_path, PATH_LEN, PATH_LEN);
  input_string (in, 2, 12, f->down - 2, get_flag, FLAG_LEN, FLAG_LEN);
  input_log    (in, 3, lg);
  
  view_frame( f);
  
  lk = f_getstr( in);
  
#ifdef DOS
  message( "This command is running only under UNIX ", " Grep ");
#else
  
  f->down -= 6;		  /* view the found files in the same frame */
  
#ifdef DOS				/* if he inserted "X:" --> "X:." */
  i = strlen( get_path);
  if( i > 0 && (get_path[i-1] == ':' || get_path[i-1] == root))
    sprintf (get_path+i, "%s", ".");
#endif
  
  if( (lk != ENTER && lk != CTRL_M) ||
     strlen( get_name) <= 0 ||
     strlen( get_path) <= 0
     ) err_ = TRUE;
  
  if( ! err_)
    {
      int pp[2];
      
      width = min( 60, max_x),	/* coordinates for the viewer frame */
      down  = min( 20, max_y),
      y     = (max_y-down)/2,
      x     = (max_x-width)/2;
      
      tmp_file = tempnam(
#ifdef DOS
			 "\\",
#else
			 "/tmp",
#endif
			 "KH");
      
      file_h = creat( tmp_file, S_IWRITE | S_IREAD); /* TMP. FILE */
      
      f->down -= 2;
      f_putline( f, f->down+1, f->color);
      
      view_frame( f);
      
      sprintf (tmp, "grep %s %s %s", get_flag, get_name, get_path);
      
#ifdef UNIX
      pipe( pp);			/* create pipe */
      
      if( fork() == 0) 
	{			/* child */
	  
	  close( pp[STD_IN]);
	  close(    STD_OUT);
	  dup  ( pp[STD_OUT]);
	  close( pp[STD_OUT]);
	  
	  freopen( options [ERR_DEV], "a+", stderr);
	  
	  
	  err_ = execlp (options [SH_], options [SH_], "-c", tmp, NULL);
	  
	  exit( err_);
	}
      else
	{
	  int buf_len,
	  tmp_pos,
	  s_y = 1;
	  char buf[ G_BLOCK_LEN],
	  tmp[ G_BLOCK_LEN+1];
	  
	  close( pp[STD_OUT]);
	  while( (buf_len = read( pp[STD_IN], buf, G_BLOCK_LEN)) > 0) 
	    {
	      
	      write( file_h, buf, buf_len);
	      
	      for( i=0; i<buf_len; i++)
		{
		  if( buf[i] == 10) 
		    {
		      if( s_y >= f->down) 
			f_scroll( f, UP_MODE | SCR_MODE);
		      else s_y++;
		      
		      tmp[ tmp_pos] = 0;
		      f_putstr( f, 1, s_y, tmp, f->color, SCR_MODE|FILL_MODE);
		      tmp_pos = 0;
		    }
		  else tmp[ tmp_pos++] = buf[i];
		}
	      
	      fflush_();
	      
	    }
	  close( pp[STD_IN]);
	  
	}
      wait( &err_);
#else
      i = 1;
      grep_pattern( f, &i, file_h, get_path, get_name);
#endif
      
      hide_frame( f, HIDE_MODE);
      
      close( file_h);
      
      ret_ = view_file( tmp_file,
		       tmp,
		       x,
		       y,
		       width,
		       down,
		       fnd_path,
		       PATH_LEN-1,
		       NOMODE);
      
      remove( tmp_file);
      
      if( ret_) p = fnd_path;  /* return the found pathname */
      
    }
  
#endif         /* only under UNIX */
  
  del_input( in);
  del_log  ( lg);
  
  hide_frame( f, HIDE_MODE|DELETE_MODE);
  return( p);		/* return pointer to the path */
}

/* by picked a file with pathname, it changes to this file automaticly */

void w_grep( struct win_t * w)
{
  char * p, * f_name;
  int i;
  
  p = grep_cmd();  				/* call grep command */
  
  if( p == NULL) return;
  
  if( ( i = chrindex( p, ":") ) < 0) return;	/* nothing selected */
  
  p[i] = 0;					/* select file name only */
  
  if (strlen (dirname (p)) > 0)			/* change to another dir. */
    if (change_dir (w, dirname (p) ) == -1) return;
  
  w->akt_pos = w->old_pos = max (w->begin_at-1, 0);
  w->anf_pos = 0;
  
  w_readdir( w, ".", NOMODE);			/* reread the dir */
  get_path( w);
  
  f_name = basename (p);			/* file name only */
  
						/* lock for file name
						 * in the directory contents
						 */
  for( i=0; i<w->d_files && strcmp (f_name, w->dr[i]->f) != 0; i++);
  if (i<w->d_files) w->akt_pos = w->old_pos = i; 	/* file found */

  w->fr_out     = TRUE;			/* new output */
  w->rahmen_out = TRUE;
}
