/************************************************************************
 *  (C)opyright 1996/1997 by Khamis Abu El-Komboz                       *
 *  xkhfm 1.0 can be copied and distributed freely for any              *
 *  non-commercial purposes. xkhfm can only be incorporated             *
 *  into commercial software with the permission of the current author. *
 *  Mail: khamis@multix.de                                              *
 ************************************************************************/
/*
 * $Author: khamis $
 * $Id: domenu.c,v 1.1 1997/01/07 08:39:42 khamis Exp $
 * $Source: /home/khamis/source/xkhfm/khfm/RCS/domenu.c,v $
 */

#define GLOBAL extern
#define EXTERN

#include <sys/types.h>
#include <sys/timeb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <fcntl.h>

#ifdef DOS
#include <dos.h>
#include <dir.h>
#include <io.h>
#endif

#include "dodefs.h"
#include "doglobal.h"
#include "domem.h"
#include "dostring.h"
#include "dofiles.h"
#include "doscr.h"
#include "dotext.h"
#include "dostern.h"
#include "dostat.h"
#include "doframe.h"
#include "dochoose.h"
#include "dohist.h"
#include "docfg.h"
#include "domenu.h"

int menu_rdelete (menu_T mn);
int menu_resize  (menu_T);
int menu_close   (menu_T);
int menu_readlevel (int f);

int menu_usermode = FALSE;	/* to enable changing menu entries for user */

/*
 * transform the data structure in a readable menu structur
 */

menu_T menu_first (menu_T mn)
{
  return mn == NULL ? NULL : mn->son;
}

menu_T menu_last (menu_T mn)
{
  menu_T last;
  if (mn == NULL || mn->son == NULL) return NULL;
  for (last = mn->son; last->next != NULL; last = last->next);
  return last;
}

menu_T menu_top (menu_T mn)
{
  menu_T parent;
  if (mn == NULL) return NULL;
  for (parent=mn; parent->parent != NULL; parent = parent->parent);
  return parent;
}

menu_T make_menu (menu_T parent, int key, char * title, char * comm)
{
  menu_T
    first_son = menu_first (parent),
    last_son  = menu_last  (parent),
    mn        = NULL;
  int
    d_len     = (title != NULL ? strlen (title) : 0) + 1,
    c_len     = (comm != NULL ? strlen (comm) : 0) + 1;

  mn       = getcmem (sizeof (struct menu_t)) ;
  mn->key  = key;
  mn->title = getmem (d_len);	/* title */
  mn->comm = getmem (c_len);	/* commando    */

  strcpy (mn->title, (title != NULL) ? title : leer);
  strcpy (mn->comm, (comm != NULL) ? comm : leer);

  mn->parent    = parent;			/* define parent */
  mn->prev      = last_son;			/* insert at end of list */
  mn->next      = NULL;				/* last son */
  if (last_son != NULL) last_son->next = mn;	/* next pointer */
  
  if (parent != NULL)
    {
      if (parent->son == NULL) parent->son = mn;   /* first son */
      parent->cnt   ++;				/* inc. count of sons */
      parent->width = max (parent->width, d_len); /* max. width of sons */
    }

  return mn;
}

int menu_delete (menu_T mn)
{
  /*
   * minimum is one entry on the top of the menu structure
   */
  if (mn->parent->parent->parent == NULL && mn->parent->cnt == 1)
    {
      message (" This is the last top menu in the structure, ;;"
	       " I can't delete it ... !!!",
	       mn->title);
      return FALSE ;
    }
  if (Getyesno (" Do you want realy to delete the menu entry ", NULL,
		mn->title,
		1)
      != 0
      )
    return FALSE ;

  return menu_rdelete (mn);
}

int menu_rdelete (menu_T mn)
{
  menu_T son, next, prev, parent;

  if (mn == NULL) return FALSE;
				/* recursive delete 
				 */
  for (son=mn->son; son != NULL; son = son->next) menu_rdelete (son);

				/* the first son of the parent 
				 */
  parent = mn->parent;
  if (parent != NULL &&		/* parent avail. */
      parent->son == mn		/* same son */
      )
    parent->son = mn->next; 	/* parent points to next son or to NULL */

  next = mn->next;
  prev = mn->prev;
  
  if (prev != NULL) prev->next = next;	/* don't break the list */
  if (next != NULL) next->prev = prev;
  Free (mn->title);
  Free (mn->comm);
  Free (mn);

  if (parent != NULL) menu_resize (parent->son); /* correct attributes */

  return TRUE;
}

/* scroll in the menu entries	[VERTICAL]
 */
int menu_bar (int x,
	      int y,
	      menu_T mn,
	      int * key
	      )
{
  menu_T parent = mn->parent, son;
  struct frame_t  * f;
  struct choose_t * ch;
  char
    hlp_text [] =
	"Ins/F3:New item"
	"#F2:Edit item"
	"#Del:Delete item"
	"#Esc:Break";
  char 
    hlp_choose [] = " Please choose your entry or press F12 to abort ... ";
  int
    lk      = 0,
    i,
    end_    = FALSE,
    restart;

  for (restart = TRUE; restart == TRUE; )
    {
      restart = FALSE;
      f  = make_frame  (min (x, max_x-mn->width-2), y,
			mn->width + 2, mn->cnt + 2,
			leer, leer,
			F_B_MENU_CLR, NOMODE);
      ch = make_choose (f,
			0, F_B_MENU_CHS,
			AUTO_MODE | LET_MODE |
			(menu_usermode ? BREAK_MODE: 0x0) /* user mode */
			);
      choose_help (ch, (menu_usermode ? hlp_text : hlp_choose));
      ch->pos  = mn->pos;
      
      for( i=0, son=mn->son; son != NULL && i<mn->cnt; i++, son = son->next)
	{
	  f_putstr (f, 1, i+1, son->title, f->color, FILL_MODE);
	}  

      view_frame (f);
      
      for (end_ = FALSE; ! end_ && ! restart; )
	{
	  lk = f_choose (ch);
	  
	  for (i=0, son=mn->son; i<ch->pos; son = son->next, i++);
	  
	  switch (lk)
	    {
	    case CUR_LEFT:
	      end_ = TRUE;
	      break;
	      
	    case CUR_RIGHT:
	      if (son->son != NULL)
		{
		  lk = menu_bar (x+2, ch->pos+y+2, son, key);
		  if (key [0] == ENTER || key [0] == CTRL_M) end_ = TRUE;
		}
	      if (lk == CUR_RIGHT) end_ = TRUE;
	      break;
	      
	    case ENTER:
	    case CTRL_M:
	      if (son->son != NULL)	/* recursive call of submenus */
		{
		  menu_bar (x+2, ch->pos+y+2, son, key); /* recursive */
		  if (key [0] == ENTER || key [0] == CTRL_M) end_ = TRUE;
		}
	      else if (son->comm [0] != 0)
		{
		  run (son->comm, SAVE_SCREEN);
		}
	      else
		{
		  * key = son->key;		/* return key */
		  end_  = TRUE;
		}
	      break ;
	      
	    case INSERT:
	    case F3:
	      menu_insert (mn);
	      restart = TRUE;
	      break;
	      
	    case F2:
	      menu_edit (son);
	      restart = TRUE;
	      break ;
	      
	    case DELETE:		/* delete user defined menu item */
	    case REMOVE:
	      menu_delete (son);
	      if (mn->son == NULL)		/* last son */
		end_ = TRUE;
	      else
		restart = TRUE; 		/* rebuild and view */
	      break ;
	      
	    case ESC:
	      end_ = TRUE ;
	      break ;
	    }
	}	/* for (end_ = FALSE; ..) */
      
      mn->pos = ch->pos;

      hide_frame (f, HIDE_MODE | DELETE_MODE);
      del_choose (ch);
    }		/* for (restart ..) */
  return     lk;
} /* ev-chs **/

/*
 * scroll menu [HORIZONTAL]
 */
int menu_choose (menu_T mn)
{
  menu_T son;
  int 
    mn_down     = FALSE,
    mn_x   	= 1,
    mn_y   	= 0,
    mn_width,
    i,
    lk     	= 0,
    key    	= -1,
    end    	= FALSE,
    nextfor	= FALSE;
  struct frame_t * f;
  struct choose_t * ch;
  int f_width = 0;
  char hlp_text [] = " Press ENTER to list the existing entries ... ";

    				/* when values changed, must be viewed */
  for (nextfor=TRUE; nextfor == TRUE; )
    {
      nextfor = FALSE;
      for (son=mn->son, mn_width = 0; son != NULL; son = son->next)
	{
	  mn_width += strlen (son->title) + 1;
	}
      
      f = make_frame (2, 0, min(mn_width+1,max_x), 1,
		      leer, leer,
		      F_B_MENU_CLR,
		      NOBOX_MODE);
      
      ch = make_choose (f, mn->cnt,
			F_B_MENU_CHS,
			VERT_MODE | LET_MODE | AUTO_MODE | BREAK_MODE
			);
      choose_help (ch, hlp_text);
      
      for (i=0, mn_x = 1, mn_y = 0, son = mn->son;
	   i < mn->cnt && son != NULL;
	   i++, son = son->next)
	{
	  f_putstr (f, mn_x, mn_y, son->title, f->color, NOMODE);
	  ch->x   [i] = mn_x,
	  ch->y   [i] = mn_y,
	  ch->len [i] = strlen (son->title);
	  mn_x       += strlen (son->title) + 1;     /* goto current column */
	}
      
      view_frame (f);
      
      for (lk = -1; end == FALSE; lk = -1)
	{
	  ch->mode   = ch->mode & ~ NOWAIT_MODE;
	  if (! mn_down) lk = f_choose (ch);	 /* normal choose */
	  
	  ch->mode   = ch->mode | NOWAIT_MODE;	/* wenn menu down ist true */
	  f_choose (ch);
	  
	  /* current son */
	  for (i=0, son=mn->son; i<ch->pos; son = son->next, i++);
	  
	  if (mn_down && ! end)
	    {
	      if (son->son != NULL)		/* sons are available */
		{
		  lk = menu_bar (f->x + ch->x [ch->pos] - 1,
				 f->y + ch->y [ch->pos] + 1,
				 son,
				 &key);
		}
	      else
		{
		  lk = getkey (-1);	/* ther are no sons for this menu*/
		}	      
	      if (lk == ENTER || lk == CTRL_M)
		{
		  if (son->son != NULL) break;		/* key is stored */
		  if (son->key > 0)			/* menu has a key */
		    {
		      key = son->key;
		      break;
		    }
		  if (son->comm [0] != 0)	/* menu must be executed */
		    {
		      run (son->comm, SAVE_SCREEN);
		      lk = -1;			/* don't terminate */
		    }
		}
	    }
	 
	  if (! end) switch (lk)
	    {
	    case CUR_RIGHT: ch->pos ++; break;
	      
	    case CUR_LEFT:  ch->pos --; break;
	      
	    case CUR_DOWN:  mn_down = TRUE; break ;

	    case CTRL_M:
	    case ENTER:
	      if (mn_down) end     = TRUE;
	      else         mn_down = TRUE;
	      break;
	      
	    case ESC:
	      if (mn_down && menu_usermode)
		{
		  mn_down = FALSE;
		}
	      else end = TRUE; 
	      break;
	    }
	  
	  ch->pos = (mn->cnt + ch->pos) % mn->cnt;
	  if (end || nextfor) break ;
	}
      
      del_choose (ch);
      hide_frame (f, DELETE_MODE | HIDE_MODE);

      if (!nextfor) break;

    }  /* for (;TRUE;*) */
  
  return lk == ENTER || lk == CTRL_M ? key : -1;
}  /* menu */

/*
 * correct the menu width 
 */
int menu_resize (menu_T mn)	
{
  menu_T son, parent;
  if (mn == NULL) return FALSE;
  parent = mn->parent;	 /* parent of the son to correct the entry for it */
  for (parent->width = 0, parent->cnt = 0, son = parent->son;
       son != NULL; son = son->next)
    {
      parent->width = max (parent->width, strlen (son->title));
      parent->cnt   ++;			/* correct number of sons */
    }
					/* current prompt for choose */
  parent->pos = max (0, min (parent->cnt-1, parent->pos));
  return TRUE;
}

#define menu_changekey(mn, k) (mn->key = (k))

int menu_changetext (menu_T mn, char * title, char * comm)
{
  if (title != NULL)
    {
      Free (mn->title);
      mn->title = getmem (strlen(title)+1);
      strcpy (mn->title, title);
    }
  if (comm != NULL)
    {
      Free (mn->comm);
      mn->comm = getmem (strlen(comm)+1);
      strcpy (mn->comm, comm);
    }
  menu_resize (mn);
  return TRUE;
}


/***********************************************************************
 * USER Menu
 ***********************************************************************/

/**********************************************************************
 * read from saved file ".dirmenu" in um_items, ..
 ***********************************************************************/
menu_T make_umenu (void)
{
  menu_T son, mn;
  int i, j;
  /* Initialising
   */
  mn = make_menu (NULL, 0, " USER ", NULL);		/* make top */
					/* user can make only in the pull-down
					 * menu his entries, that we need a
					 * dummy son for the top to have the
					 * pull-down entries
					 */
  son = make_menu (mn, 0, " Edit user menu ", NULL);
  menu_open (son);			/* read entries from file */
  if (son->son == NULL)			/* there are no values readed */
    make_menu (son, 0, " Dummy entry ", NULL);
  return mn;				/* return the new pointer */
}

void umenu_call (void)
{
  menu_usermode = TRUE;			/* enable changing menu entries */
  if (umenu_top == NULL)
    umenu_top  = make_umenu () ;	/* Top header */
  menu_choose  (umenu_top);		/* scroll/edit  */
  menu_close   (umenu_top->son);	/* save in file */
  menu_rdelete (umenu_top);		/* Free memory */
  umenu_top = NULL;			/* remark umenu_top to NULL */
  menu_usermode = FALSE;		/* disable menu user/changemode */
}


menu_T menu_insert (menu_T parent)
{
  menu_T son;
  son = make_menu (parent, 0, NULL, NULL);	/* crate son */
  if (menu_edit (son)) return son;		/* successful edited */
  menu_rdelete (son);				/* delete if not aborted */
  return NULL ;					/* error */
}

					/* edit menu entries
					 */
int menu_edit (menu_T mn)
{
  struct frame_t * f;
  struct input_t * in;
  struct log_t   * lg [2];
  char
    Title [MNTITL_LEN],
    Comm  [MNCOMM_LEN],
    hlp_text [] = " Enter your Title and the command line ... ";
  int
    width = min (max_x, MNTITL_LEN),
    down  = 8,
    x     = (max_x-width)/2,
    y     = (max_y-down)/2,
    have_sons = (mn->son != NULL ? TRUE : FALSE),
    i, j;

  f = make_frame (x, y, width, down,
		  " Insert/Edit user menu ", leer,
		  F_B_UTIL_CLR,
		  NOMODE);
  f_putstr  (f, 1, 1, " Title  :", f->color, NOMODE);
  f_putstr  (f, 1, 2, " Command:", f->color, NOMODE);
  f_putline (f, 3, f->color);
  f_putstr  (f, 1, 4, " Prompt for parameter:", f->color, NOMODE);
  f_putline (f, 5, f->color);
  f_putstr  (f, 1, 6,
	     have_sons
	            ? " There are sons availaible ... it's a menu ... !"
	            : " Type................:",
	     f->color, NOMODE);

  in = make_input (f, 4, CUR_MODE);
  input_help (in, hlp_text);
  for (i=0; i<2; i++)
    {
      lg [i] = make_log   (f, 2, F_B_UTIL_CHS);
    }

  input_string (in, 0, 16, 1, Title, MNTITL_LEN, MNTITL_LEN);
  input_string (in, 1, 16, 2, Comm,  MNCOMM_LEN, MNCOMM_LEN);

  strcpy (Title, mn->title); in->str [0] = Title;
  strcpy (Comm,  mn->comm);  in->str [1] = Comm;	

  log_format (lg [0], "X", " ");
  log_xy     (lg [0], 26, 4, 26, 4);
  log_format (lg [1], "Item", "Menu");
  log_xy     (lg [1], 26, 6, 36, 6);

  input_log (in, 2, lg [0]);
  input_log (in, 3, lg [1]);
				   /* when sons are avail.,
				    * no [item/menu] is prompted for choosing
				    */
  if (have_sons) in->cnt -- ;

  view_frame (f); fflush_();

  i = f_getstr (in);			/* edit menu entry */

  if (i != ESC)			        /* successfull changed */
    {
      menu_changetext (mn, Title, Comm);
					/* for menu entry, create a dummy
					 * barscroll, when there is no sons
					 * available
					 */
      if (lg[1]->pos == 1 && mn->son == NULL)
	{
	  make_menu (mn, 0, "Dummy entry (ls)", "ls");
	}
    }

  del_input (in);
  hide_frame (f, HIDE_MODE|DELETE_MODE);

  return i != ESC ? TRUE : FALSE;
}

char * menu_name (void)
{
  static char buf [MAXPATH];
  sprintf (buf, "%s/.dirmenu", getenv ("HOME"));  
  return buf;
}

int menu_close (menu_T mn)
{
  if (Getyesno (" Do you want to save the new menu changing ? ;;"
		" otherwise all changes well be destroyed know ..! ",
		NULL,
		" Save user menu ",
		0)
      != 0
      ) 
    return FALSE;
  return menu_save (mn);  
}
int menu_save (menu_T mn)
{
  FILE * f;
  int level = 0, ret;
  f    	  = fopen (menu_name(), "w+");
  ret     = menu_write (f, mn, &level);
  fclose (f);
  return ret;
}

int menu_write (FILE * f, menu_T mn, int * level)
{
  menu_T next;
  int ll = level [0];

  level[0] ++;					/* increment to next level */
  for (next=mn->son; next != NULL; next = next->next)
    {
#define LEVEL	  "Level   ["
#define TITEL	  "Title   ["
#define COMMAND	  "Command ["
#define MNU_KEY	  "Key     ["
      fprintf (f, "Level   [%i];\n", level [0] );
      fprintf (f, "Title   [%s];\n", next->title);
      fprintf (f, "Command [%s];\n", next->comm);
      fprintf (f, "Key     [%i];\n", next->key );
      menu_write (f, next, level);	/* recursive */
    }
  level [0] = ll;
  return TRUE;
}

int menu_open (menu_T mn)
{
  int f, level;
  f = open (menu_name (), O_RDONLY);
  if (f <= 0) return FALSE;
  level = menu_readlevel (f);		/* first level       */
  menu_read (f, mn, &level);		/* recursive reading */
  close (f);
  return TRUE;
}

char * menu_readline (int f, char * name)
{
  static char buf [256];
  int len, pos;

  for (; TRUE; )
    {
      len = read_line (f, buf, 256);	  /* next line 	  */
      if (len < 0) return NULL;		  /* end of file  */
      if (begins (buf, name) == 0) break; /* title found [Level,Key,..] */
    }
  pos = chrindex (buf, "];");		  /* index of "];" 		*/
  if (pos > 0) buf [pos] = 0;		  /* delete str. at "];" to end */
  return buf + strlen (name);		  /* cut the name from string   */
}

int menu_readlevel (int f)
{
  char * buf  = menu_readline (f, LEVEL);
  return buf == NULL ? -1 : atoi (buf);
}

int menu_readkey (int f)
{
  char * buf = menu_readline (f, MNU_KEY);
  return buf == NULL ? -1 : atoi (buf);
}

char * menu_readtitle (int f)
{
  return menu_readline (f, TITEL);
}

char * menu_readcomm (int f)
{
  return menu_readline (f, COMMAND);
}

int menu_read (int f, menu_T mn, int * level)
{
  menu_T son = NULL;
  int ll = level [0];

  if (level [0] == -1) return FALSE;  /* file empty or end of file */

  for (ll = level [0]; ll == level [0]; )
    {
      son = make_menu (mn, 0, NULL, NULL);

      menu_changetext (son, menu_readtitle (f), NULL);
      menu_changetext (son, NULL, menu_readcomm (f));
      menu_changekey  (son, menu_readkey  (f));

      level [0] = menu_readlevel (f);

      if (level [0] > ll)
	{
	  menu_read (f, son, level);	        /* read his sons, recursive */
	}
    }
  return TRUE;
}

/*
 * convert string array in menu structure
 */
int menu (int    keys  [][MAXMENU_ENTRY],
	  char * line  [],
	  char * event [][MAXMENU_ENTRY]) 
{
  menu_T top, parent, user;
  int i, j, ret;

  parent = top = make_menu (NULL, 0, NULL, NULL);

  /*
   * Build Menu-Structur from Parameters
   */
  for (i=0; line [i] != NULL; i ++)
    {
      parent = make_menu (top, 0, line [i], NULL); /* rekursive */
      for (j=0; event [i][j] != NULL; j++)
	{
	  make_menu (parent, keys [i][j], event [i][j], NULL);
	}
    }

  /*
   * Add User-Menu on the End of the List
   */
  user = make_menu (top, 0, " USER ", NULL); /* insert user menu in struct. */
  menu_open (user);			     /* load entries from file      */

  ret = menu_choose (top);	/* menu system */
  menu_rdelete (top);		/* recursive delete without asking */
  return ret;
}

