/************************************************************************
 *  (C)opyright 1996/1997 by Khamis Abu El-Komboz                       *
 *  xkhfm 1.0 can be copied and distributed freely for any              *
 *  non-commercial purposes. xkhfm can only be incorporated             *
 *  into commercial software with the permission of the current author. *
 *  Mail: khamis@multix.de                                              *
 ************************************************************************/
/* Funktionsname=%M%, SID=%I%, Date=%E% */

#ifndef DIR_H
#define DIR_H

#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <sys/param.h>
#include <fcntl.h>

/* availiable file types */
enum
{
  TYPE_DIR,
  TYPE_REG,
  TYPE_EXE,
  TYPE_BLK,
  TYPE_CHR,
  TYPE_FIFO,
  TYPE_LNK,
  TYPE_NAME,
  TYPE_SEM,
  TYPE_SHD,
  TYPE_UNK
};
static char TypeChr[] = "/ *bcflnsd?";

#define TREE_DIR	PIXM_tree_dir16
#define TREE_OPENDIR	PIXM_tree_odir16

/* Defines for sorting a directory entry */
#define SORT_NAME		0x0001
#define SORT_DIR_FIRST		0x0002
#define SORT_MOD_TIME		0x0004
#define SORT_CREATE_TIME	0x0008
#define SORT_ACCESS_TIME	0x0010
#define SORT_USER_ID		0x0020
#define SORT_GRP_ID		0x0040
#define SORT_USER_NAME		0x0080
#define SORT_GRP_NAME		0x0100
#define SORT_INODE		0x0200
#define SORT_LINKS		0x0400
#define SORT_UNSORTED		0x0800
#define SORT_SIZE		0x1000

/* defines for reading a direcory entries */
#define READ_DIR_ONLY		0x0001
#define DIR_SELECTED		0x0002
#define DIR_READ_ALL		0x0003
#define DIR_SOFT_SELECTED	0x0010
#define DIR_COPY_SELECTED	0x0020

/* defines for Tree functions */
#define TREE_SELECTED		0x0001
#define TREE_LAST_ICON		0x0002
#define TREE_REFRESH		0x0004
#define TREE_BUILD		0x0008
#define TREE_READ_RECURSIVE	0x0010
#define TREE_HAVE_SUBDIR	0x0020
#define TREE_TESTED_SUBDIR	0x0040
#define TREE_SOFT_SELECTED	0x0080
#define TREE_COPY_SELECTED	0x0100

/* defines for buildind status string "dirMakeChoose" */
#define DIR_STATUS	0x0001
#define DIR_CHOOSE	0x0002
#define DIR_PRINT_NAME	0x0004
#define DIR_DIFFER	0x0008
#define DIR_LINKED_TO	0x0010	/* if file is a link, the linked file is *
				 * viewed at the end of string           */

#ifdef DOS

#	define S_ISDIR(m)  ( m & S_IFDIR)
#	define S_ISREG(m)  ( m & S_IFREG)
#	define S_ISBLK(m)  ( m & S_IFBLK)
#	define S_ISCHR(m)  ( m & S_IFCHR)
#	define S_ISFIFO(m) ( m & S_IFIFO)
#	define S_ISNAM(m)  ( TRUE)
#	define S_ISLNK(m)  ( TRUE)
#else 		/* UNIX */

#if ! defined (S_ISLNK)
#if defined (S_IFLNK)
#define S_ISLNK(m) 	((m & S_IFLNK) == S_IFLNK)
#elif defined (_IFLNK)
#define S_ISLNK(m)	((m & _IFLNK) == _IFLNK)
#else
#define S_ISLNK(m)	((m & 0120000) == 0120000)
#endif

#define O_BINARY	0x0000

#endif

#if ! defined (S_ISNAM)
#if defined (S_IFNAM)
#define S_ISNAM(m)	((m & S_IFNAM) == S_IFNAM)
#else
#define S_ISNAM(m)      FALSE
#endif
#endif

#endif	/* DOS */

#ifndef time_t
#define time_t	long
#endif
#ifndef uid_t
#define uid_t	long
#endif
#ifndef gid_t
#define gid_t	long
#endif
#ifndef ino_t
#define ino_t	long
#endif
#ifndef nlink_t
#define nlink_t	long
#endif

/* known file extensions */
typedef struct
{
  char * StringMatch;
  int Type;
} * DirExtendedTypes_t, dirExtendedTypes_t;

typedef struct
{
  char * Name;
  char * Path;
  struct stat * Stat;
  short  ExtentedType;
  int    Info;
} dir_t, * Dir_t;

typedef struct
{
  char * Name;
  char * PathName;	/* full pathname */
  short	 Level;
  int    Info;		/* selected, .. */
  Dchain_t subTree;
  Dchain_t Parent;
} dirTree_t, * DirTree_t;

Dir_t       dirMake  (char *);
int         dirDest  (Dir_t lst);
int         dirComp  (Dir_t liste, char * Name);
Dchain_t    dirAdd   (Dchain_t top, char * Name);
int         dirCopy  (Dir_t lst1, Dir_t lst2);
int         dirChng  (Dchain_t lst, char * newNr);
int         dirSelect (Dchain_t lst, int mode);
int         dirSoftSelect (Dchain_t lst, int mode);
int         dirCopySelect (Dchain_t lst, int mode);
int         dirDelete (Dchain_t lst);
#define	    dirSelected(lst) \
                       (((Dir_t)UdchClient(lst))->Info & DIR_SELECTED)
#define	    dirSoftSelected(lst) \
			 (((Dir_t)UdchClient(lst))->Info & DIR_SOFT_SELECTED)
#define	    dirCopySelected(lst) \
			 (((Dir_t)UdchClient(lst))->Info & DIR_COPY_SELECTED)

#define dirName(lst) ((Dir_t)UdchClient(lst))->Name
#define dirPath(lst) ((Dir_t)UdchClient(lst))->Path
char * dirValue (Dchain_t lst);

int         dirReadDir (Dchain_t top, char * PathName, char * Filter, int mode);
struct stat * dirStat  (Dchain_t entry);
int	    dirType    (Dchain_t entry);
int	    dirExtendedType (Dchain_t entry);

time_t  dirModTime	(Dchain_t dir);
time_t  dirCreateTime   (Dchain_t dir);
time_t  dirAccessTime   (Dchain_t dir);
uid_t   dirUserId       (Dchain_t dir);
gid_t   dirGrpId        (Dchain_t dir);
char *  dirUserName     (Dchain_t dir);
char *  dirGrpName      (Dchain_t dir);
ino_t   dirINode        (Dchain_t dir);
nlink_t dirLinks        (Dchain_t dir);
dev_t   dirDevice       (Dchain_t);
dev_t   dirRDevice      (Dchain_t);
off_t   dirSize         (Dchain_t);
char *  dirGetName      (char * PathName, int NameNum);
char *  dirMakeStatStr  (Dchain_t top, int mode);
char *  dirMakeChoose   (Dchain_t dir, int mode);
int     dirChngOwnerGrp (Dchain_t, int owner, int grp);
mode_t  dirRights       (Dchain_t entry);
int     dirExists       (Dchain_t entry);
int     dirCreateDir    (Dchain_t dir, int mode);
char *  dirRealPath     (Dchain_t dir);
int     dirTypeOfRealPath (Dchain_t dir);
char *  dirCutPath (char * path, int max_l, int mode);

/** info structur for recursive reading **/
#define DIR_ASK			0x00000001
#define DIR_RECURSIVE		0x00000002
#define DIR_CANCEL		0x00000004
#define DIR_ALL			0x00000008
#define DIR_NO			0x00000010
#define DIR_ERROR		0x00000020
#define DIR_SKIP_EXISTING	0x00000040
#define DIR_SKIP_LINKS		0x00000080
#define DIR_SKIP_FILES		0x00000100
#define DIR_SKIP_DIRS		0x00000200
#define DIR_ERROR_AVAIL		0x00000400
#define DIR_CLIENT		0x00000800
typedef struct
{
  char * Filter;
  char * Enclose;
  char * Ignore;
  int (* ViewProc        ) (int, int, void *);
  int (* ViewFilesProc   ) (Dchain_t, Dchain_t, void *);
  int (* AskProc         ) (Dchain_t, Dchain_t, void *);
  int (* AskOverwriteProc) (Dchain_t, Dchain_t, void *);
  int (* ErrProc         ) (Dchain_t, void *);
  int (* ClProc          ) (Dchain_t, Dchain_t, void *);
  void * ClData;
  unsigned long mode;
} dirInfo_t, * DirInfo_t;
int dirRecursive (Dchain_t src, Dchain_t dst, DirInfo_t info);

/* Tree functions */
DirTree_t dirTreeMake  (char *);
int       dirTreeDest  (DirTree_t lst);
int       dirTreeComp  (DirTree_t liste, char * Name);
Dchain_t  dirTreeAdd   (Dchain_t top, char * Name);
#define   dirTreeDel(lst) UdchDelete (lst, (int(*)(void*))dirTreeDest)
#define   dirTreeRemove(lst) UdchEmptyAndDel (lst, (int(*)(void*))dirTreeDest)
int       dirTreeCopy  (DirTree_t lst1, DirTree_t lst2);
int       dirTreeChng  (Dchain_t lst, char * newNr);
int       dirTreeData  (Dchain_t lst, struct stat *);
char *    dirTreeName  (Dchain_t lst);
char *    dirTreePathName (Dchain_t lst);
int       dirTreeBuild 	(Dchain_t top, char * PathName, int mode);
int       dirTreeRead	(Dchain_t top, char * PathName, char * Filter,
			 int mode);
int      dirTreeRefresh  (Dchain_t top, int mode);
int      dirTreeCount    (Dchain_t tree);
int      dirTreeSelect   (Dchain_t tree, int mode);

Dchain_t dirTreeAddSubTree  (Dchain_t tree);

int	 dirTreeHaveSubDir (Dchain_t tree);
int      dirTreeSetParent  (Dchain_t lst, Dchain_t parent);

#define dirTreeHaveSubTree(tree) (((DirTree_t) UdchClient (tree))->subTree \
				  != NULL)
#define	dirTreeLevel(lst) (((DirTree_t)UdchClient(lst))->Level)
#define	dirTreeSelected(lst) (((DirTree_t)UdchClient(lst))->Info & \
			      TREE_SELECTED)
#define dirTreeSoftSelected(lst) (((DirTree_t)UdchClient(lst))->Info & \
				  TREE_SOFT_SELECTED)
#define dirTreeCopySelected(lst) (((DirTree_t)UdchClient(lst))->Info & \
				  TREE_COPY_SELECTED)
int dirTreeSoftSelect (Dchain_t tree, int mode);
int dirTreeCopySelect (Dchain_t tree, int mode);

#define	dirTreeLastIcon(lst) (((DirTree_t)UdchClient(lst))->Info & \
			      TREE_LAST_ICON)
#define dirTreeParent(tree) (((DirTree_t)UdchClient(tree))->Parent)
#endif



