/************************************************************************
 *  (C)opyright 1996/1997 by Khamis Abu El-Komboz                       *
 *  xkhfm 1.0 can be copied and distributed freely for any              *
 *  non-commercial purposes. xkhfm can only be incorporated             *
 *  into commercial software with the permission of the current author. *
 *  Mail: khamis@multix.de                                              *
 ************************************************************************/
/*
 * $Author: khamis $
 * $Id: cnfutil.c,v 1.1 1996/10/22 09:57:15 khamis Exp khamis $
 * $Source: /home/xkhfm/khlib/khmisc/RCS/cnfutil.c,v $
 */

#define IS_EXTERN
#define EXTERN extern

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>

#include "defines.h"
#include "Ustrutil.h"
#include "Ufileutil.h"
#include "Ucnfutil.h"

int confAddSect (Conf_t cnf, char * section);
int confAddStr (Conf_t cnf, char * str, int pos, int mode);

Conf_t confOpen  (char * confFile, char * title)
{
  Conf_t cnf = gcalloc (sizeof (conf_t), __LINE__, __FILE__);

  if (cnf == NULL)
    return NULL;

  DEBUG2 ("confOpen: Dateiname (%s)\n", confFile);

  strcpy (cnf->Filename, confFile);
  cnf->FileMode   = 0;
  cnf->CurrentPos = 0;
  /******* Defaultmaessig *****************************************************/
  cnf->Trennz [0] = '=';
  cnf->Comment    = '#';
  cnf->LineAppend = '\\';

  /****** Falls datei vorhanden, Inhalt der Datei in den Speicher kopieren ****/
  if (UFileExists (cnf->Filename))
    {
      cnf->FileDesc = open (cnf->Filename,
			    O_RDONLY,
			    UGetUmask (True));
      if (cnf->FileDesc < 0)
	{
	  MFree (cnf, sizeof (conf_t));
	  return NULL;
	}
      
      /****** Speicher fuer die gesamte Konfigurationsdatei *******************/
      cnf->FileSize  = UFileSize (cnf->Filename);
      cnf->FileEntry = gmalloc (cnf->FileSize, __LINE__, __FILE__);

      /****** Inhalt der Datei in den Speicher holen **************************/
      read  (cnf->FileDesc, cnf->FileEntry, cnf->FileSize);
      close (cnf->FileDesc);
    }
  else
    {
      char tmp [MAXPATHLEN];

      sprintf (tmp, "## Filename: %s\n", cnf->Filename);
      /****** Speicher fuer die gesamte Konfigurationsdatei *******************/
      cnf->FileSize  = 0;
      cnf->FileEntry = gmalloc (cnf->FileSize, __LINE__, __FILE__);
      confAddStr (cnf, tmp,   cnf->CurrentPos, False);
      confAddStr (cnf, title, cnf->CurrentPos, False);
    }
    
  DEBUG1 ("confOpen: Ende");

  return cnf;
}

int confClose (Conf_t cnf)
{
  int ret = True;
  if (cnf == NULL)
    return False;
  
  DEBUG2 ("confClose: filename (%s)\n", cnf->Filename);

  /****** Die Datei nur sichern, wenn was geaendert wurde *********************/
  if (cnf->WriteFlag)
    {
      /* Datei erstellen und Rechte setzen ************************************/
      cnf->FileDesc = open (cnf->Filename,
			    O_WRONLY|O_TRUNC|O_CREAT,
			    UGetUmask (True));
      
      /****** Fehler beim Oeffnen *********************************************/
      if (cnf->FileDesc < 0)
	{
	  ERROR2 ("confOpen: Datei (%s) kann nicht geoeffnet werden\n",
		  cnf->Filename);
	  ret = False;
	}
      else
	/****** Speicher in die Datei kopieren ********************************/
	write (cnf->FileDesc, cnf->FileEntry, cnf->FileSize);
      
      /** geoffnete Parameterdatei schliessen *********************************/
      close (cnf->FileDesc);
    }

  if (cnf->FileEntry != NULL)
    MFree (cnf->FileEntry, cnf->FileSize);

  MFree (cnf, sizeof (conf_t));

  DEBUG1 ("confClose: Ende");

  return ret;
}

int confSectPos (Conf_t cnf, char * section)
{
  long pos;
  char * lsection;

  /****** Falls Section leer ist oder gleich NULL dann wird der Anfang der ****/
  /****** Datei als Section verstanden					  *****/
  if (section == NULL || strcmp (section, "") == 0)
    {
      int i, oi;

      /** In diesem Falle werden die ersten Zeilen der Kommentare *************/
      /** ueberschritten ******************************************************/
      for (i=oi=0; i <= cnf->FileSize && cnf->FileEntry [i] == cnf->Comment;)
	{
	  oi=i;

	  /** Bis Zeilenende **************************************************/
	  while (i <= cnf->FileSize && cnf->FileEntry [i] != '\n')
	    i++;
	  i++;
	}
      cnf->CurrentPos = oi;
      return cnf->CurrentPos;
    }

  /****** Section bilden ******************************************************/
  lsection = gmalloc (strlen (section) + 4, __LINE__, __FILE__);
  sprintf (lsection, "\n[%s]", section);
  
  /****** Suchen, 1. Stelle bei dieser Funktion beginnt ab 1 ******************/
  pos = StringFindNStr (cnf->FileEntry, lsection,
			cnf->FileSize,
			strlen (lsection));

#if 0
  DEBUG3 ("confSectPos: section (%s), pos=(%i)\n", lsection, pos);
#endif

  MFree (lsection, strlen (section) + 4);

  if (pos <= 0)
    pos = -1;
  else
    cnf->CurrentPos = pos;

  return pos;
}

int confAddSect (Conf_t cnf, char * section)
{
  long pos;
  char * lsection;

#if 0
  DEBUG2 ("confAddSect: Sektion (%s)\n", section);
#endif

  /****** Falls die Section vorhanden ist, die Position danach zureuckgeben ***/
  pos = confSectPos (cnf, section);

  if (pos >= 0)
    {
      /****** Bis Zeilenende vorruecken ***************************************/
      confReLine (cnf, False);
      return cnf->CurrentPos;
    }

  /****** Aus der uebergebenen Section wird "[" Section "]" gebildet **********/
  lsection = gmalloc (strlen (section) + 5, __LINE__, __FILE__);
  sprintf (lsection, "%s[%s]\n",
	   cnf->FileEntry [cnf->FileSize-1] != '\n'
	   ? "\n"
	   : "",
	   section);

  /****** Am Ende der Datei schreiben *****************************************/
  cnf->FileEntry = realloc (cnf->FileEntry,
			    cnf->FileSize + strlen (lsection));
  memcpy (cnf->FileEntry + cnf->FileSize, lsection, strlen (lsection));
  cnf->FileSize += strlen (lsection); 

  /****** Position an der richtigen Stelle bewegen ****************************/
  cnf->CurrentPos = cnf->FileSize;

  /****** Speicher freigeben **************************************************/
  MFree (lsection, strlen (section) + 5);

#if 0
  DEBUG1 ("confAddSect: Ende");
#endif

  return cnf->FileSize;
}

int confDelLine (Conf_t cnf, int pos)
{
  char * line;
  int i, bis, len;
  
  if (cnf == NULL || pos >= cnf->FileSize)
    return False;

  cnf->CurrentPos = pos;

  /****** an die Naechste Zeile positionieren ********************************/
  line = confReLine (cnf, False);
  len  = cnf->CurrentPos - pos;
  
  
#if 0
  DEBUG2 ("confDelLine: Zeile (%s) loeschen\n", line);
#endif

#if 0
  DEBUG2 ("confDelLine: %s", "(");
  for (i=0; i<cnf->FileSize; i++)
    DEBUG2 ("%c", cnf->FileEntry [i]);
  DEBUG1 (")");
#endif

  /****** Hinteres Ende vorruecken ********************************************/
  for (i=pos; i<cnf->FileSize-len; i++)
    cnf->FileEntry [i] = cnf->FileEntry [i+len];

  /****** Speicher kuerzen ****************************************************/
  cnf->FileSize  = cnf->FileSize - len;
  cnf->FileEntry = realloc (cnf->FileEntry, cnf->FileSize);

#if 0
  DEBUG2 ("confDelLine: neuer Inhalt(", "");
  for (i=0; i<cnf->FileSize; i++)
    DEBUG2 ("%c", cnf->FileEntry [i]);
  DEBUG1 (")");
#endif

  cnf->CurrentPos = pos;

  return True;
}

char * confFormat (Conf_t cnf, char * buf)
{
#define LINE_LEN	100
  int len, i, j, ll;
  char * fStr;
  
  if (buf == NULL)
    return NULL;

  fStr = gmalloc (strlen (buf) * 2, __LINE__, __FILE__);
  len  = strlen (buf);
  for (i=j=ll=0; i<len; i++, ll++)
    {
      /** Neue Zeile **********************************************************/
      if (buf[i] == '\n')
	{
	  fStr[j++] = cnf->LineAppend;
	  fStr[j++] = 'n';
	}

      /** Das Trennzeichen als normales Zeichen *******************************/
      /** BackSlash auch maskieren ********************************************/
      else if (buf[i] == cnf->Trennz[0] || buf[i] == cnf->LineAppend)
	{
	  fStr [j++] = cnf->LineAppend;
	  fStr [j++] = buf[i];
	}

      /** Normale Zeichen *****************************************************/
      else
	fStr[j++] = buf[i];

#if 1
      /** Zeilenumbruch, generieren, falls die Zeile zubreit wird *************/
      if (ll > LINE_LEN && cnf->LineAppend != 0)
	{
	  fStr [j++] = cnf->LineAppend;
	  fStr [j++] = '\n';
	  ll = 0;
	}
#endif
    }
  fStr [j] = 0;

#if 0
  DEBUG3 ("confFormat: (%s) ==> (%s)\n", buf, fStr);
#endif

  return fStr;
}

int confAddStr (Conf_t cnf, char * str, int pos, int mode)
{
  char * fStr;
  int i, len;

  /****** Zeile formatieren, falls erforderlich *******************************/
  if (mode)
    fStr = confFormat (cnf, str);
  else
    fStr = str;

#if 0
  DEBUG2 ("confAddStr: alter Inhalt(", "");
  for (i=0; i<cnf->FileSize; i++)
    DEBUG2 ("%c", cnf->FileEntry [i]);
  DEBUG1 (")");
#endif

  len = strlen (fStr);

  /****** Speicher ausdehnen **********************************************/
  cnf->FileEntry = realloc (cnf->FileEntry, cnf->FileSize + len);
 
  /****** Dateigroesse anpassen **********************************************/ 
  cnf->FileSize += len;

  /****** Platz fuer den neuen Eintrag schaffen *******************************/
  for (i=cnf->FileSize-1; i>=pos; i--)
    cnf->FileEntry [i] = cnf->FileEntry [i-len];
  
  /****** String Kopieren *************************************************/
  memcpy (cnf->FileEntry + pos, fStr, len);

#if 0
  DEBUG2 ("confAddStr: Neuer Inhalt(", "");
  for (i=0; i<cnf->FileSize; i++)
    DEBUG2 ("%c", cnf->FileEntry [i]);
  DEBUG1 (")");
#endif

  cnf->CurrentPos = pos + len;

  if (mode)
    MFree (fStr, strlen (str) * 2);

  return cnf->CurrentPos;
}

char * confReLine (Conf_t cnf, int mode)
{
  int i, j, len = MAXLEN_CNFLINE;
  int comment = False;
  char o_ch = 0;
  static char buf [MAXLEN_CNFLINE];

  /****** Dateiende ***********************************************************/
  if (cnf->CurrentPos >= cnf->FileSize)
    return NULL;
  
  for (j = 0, i = cnf->CurrentPos;
       i<cnf->FileSize;
       i++) 
    {
      /* Ende der Zeile wird genau dann, wenn kein Verbindungszeichen *****/
      /* vorliegt *********************************************************/
      if (cnf->FileEntry [i] == '\n' && o_ch != cnf->LineAppend)
	{
	  i++;
	  break;
	}

      /* Falls ein Kommentarzeichen gefunden wird, vormerken **************/
      if (cnf->FileEntry [i] == cnf->Comment && o_ch != cnf->LineAppend)
	comment = True;

      /****** Falls Kommentarzeichen erreicht, nur bis Ende der Zeile *********/
      /****** einlesen ********************************************************/
      if (! comment && j < len-1)
	{
	  if (mode)
	    {
	      /** Zeile nur dann formatiert einlesen, wenn das mode auf True  */
	      
	      /* Newline **************************************************/
	      if (o_ch == cnf->LineAppend && cnf->FileEntry[i] == 'n')
		buf [j++] = '\n';
	      
	      /* Tab ******************************************************/
	      else if (o_ch == cnf->LineAppend && cnf->FileEntry[i] == 't')
		buf [j++] = '\t';
	      
	      /* Das Zeichen danach ***************************************/
	      else if (o_ch == cnf->LineAppend && cnf->FileEntry[i] != '\n')
		buf [j++] = cnf->FileEntry [i];
	      
	      /* Falls kein '\\' oder '\n' ********************************/
	      else if(cnf->FileEntry [i] != cnf->LineAppend &&
		      cnf->FileEntry [i] != '\n')
		buf [j++] = cnf->FileEntry[i];
	    } /* if */
	  
	  /****** Das Verbindungszeichen und das \n nicht mituebergeben *******/
	  else if (cnf->FileEntry [i] == '\n' && o_ch == cnf->LineAppend)
	    {
	      j--;
	    }
	  else
	    {
	      buf [j++] = cnf->FileEntry [i];
	    }
	}

      /* zuletzt eingelesenes Zeichen vormerken ***************************/
      o_ch = cnf->FileEntry [i];
    } /* for */
  buf [j] = 0;

  cnf->CurrentPos = i;

  return buf;
}

int confWrite (Conf_t cnf, char trennz, char * section, int argc, char * argv[])
{
  char * value;
  int i;
  long pos;
  char * line, arg0 [MAXLEN_CNFLINE];
  int WriteParam = False;	/* Flagge *************************************/
  int cnt;
  if (argc == -1)
  {
    for (cnt=0; argv[cnt] != NULL; cnt++);
  }
  else
  {
    cnt = argc;
  }
      

  if (cnf == NULL)
    return False;
  
  /****** Trennzeichen behalten ***********************************************/
  cnf->Trennz [0] = trennz;

  /****** Beim Schliessen muss die Datei aktualisiert werden ******************/
  cnf->WriteFlag = True;

  /****** Section anlegen, falls nicht vorhanden ******************************/
  pos = confSectPos (cnf, section);
  if (pos < 0)
    {
      pos = confAddSect (cnf, section);
      if (pos < 0)
	{
	  ERROR2 ("confWrite: Kann Sektion (%s) nicht anlegen\n", section);
	  return False;
	}
       cnf->CurrentPos =
	 pos =
	   cnf->FileSize;
    }
  else
    {
      /****** Bis Zeilenende einlesen *****************************************/
      line = confReLine (cnf, False);
#if 0
      DEBUG2 ("confWrite: Sektion (%s)\n", line);
#endif  
    }

  /*** Naechste Zeile einlesen ********************************************/
  for (; True; )
    {
      /****** Alte Position merken ****************************************/
      pos = cnf->CurrentPos;
      
      /* Zeile Lesen ******************************************************/
      line  = confReLine (cnf, False);
      
#if 0
      DEBUG2 ("confWrite: line =(%s)\n", line != NULL ? line : "");
#endif
      
      /* Parameter filtern ************************************************/
      if (line != NULL)
	Ureadarg (line, arg0, 0, trennz);
      else
	arg0 [0] = 0;
      
#if 0
      DEBUG2 ("confWrite: Arg0=(%s)\n", arg0);
#endif
      
      /*** Dateiende, ohne den Parameter zu finden ************************/
      if (line == NULL)
	{
	  WriteParam = True;
	}
      
      /*** Eine neue Section erreicht, ohne den Parameter zu finden *******/
      else if (arg0 [0] == '[')
	{
	  cnf->CurrentPos = pos;
	  
	  /* Parameter wird am Ende der Section geschrieben ***************/
	  WriteParam = True;
	}
      
      /** Parameter gefunden **********************************************/
      else if (strcmp (arg0, argv[0]) == 0)
	{
	  /****** die Zeile loeschen **************************************/
	  confDelLine (cnf, pos);
	  
	  /* Parameter wird ersetzt ***************************************/
	  WriteParam = True;
	}
      
      /** Parameter in die Datei schreiben ********************************/
      if (WriteParam)
	{
	  char wline [MAXLEN_CNFLINE];
	  char *p;
	  
	  p = confFormat (cnf, argv[0]);
	  strcpy (wline, p);
	  MFree (p, strlen(argv[0])*2);
	  
	  /* Werte des Parameters anhaengen *******************************/
	  for (i=1; i<cnt && (value = argv[i]) != NULL; i++)
	    {
	      /** Kleiner Trick, falls das Trennzeichen ein '#', **********/
	      /** dann TAB einf. ******************************************/
	      if (cnf->Trennz[0] == cnf->Comment)
		{
		  strcpy (wline+strlen(wline), "\t");
		}
	      
	      /** Trennzeichen ********************************************/
	      strcpy (wline+strlen(wline), cnf->Trennz);
	      
	      /** auch nach dem Kommentarzeichen ein Leerzeichen einf. ****/
	      if (cnf->Trennz[0] == cnf->Comment)
		{
		  strcpy (wline+strlen(wline), " ");
		}
	      
	      p = confFormat (cnf, value);
	      strcpy (wline+strlen(wline), p);
	      MFree (p, strlen(value)*2);
	    }
	  
#if 0
	  DEBUG2 ("confWrite: wline (%s)\n", wline);
#endif
	  /** ganze Zeile formatiert einfuegen ****************************/
	  confAddStr (cnf, wline, cnf->CurrentPos, False);
	  
	  /* NewLine ******************************************************/
	  confAddStr (cnf, "\n", cnf->CurrentPos, False);
	}
      
      if (WriteParam)
	break;
    } /* for */

#if 0
  DEBUG1 ("confWrite: Ende");
#endif

  return True;
}

char * confRead (Conf_t cnf, char trennz, char * section, char * param)
{
  char * line;
  char arg0 [CNF_MAXPARAMLEN];
  int pos;
  int found = False;

  if (cnf == NULL)
    return NULL;

  /****** Position der Section ermitteln **************************************/
  pos = confSectPos (cnf, section);

  if (pos < 0)
    return NULL;

  /*** Bis Zeilenende auslesen ************************************************/
  confReLine (cnf, False);

  /****** Die Parameter auswerten, bis die entsprechende Zeile gefunden wird **/
  for (; True; )
    {
      
      /* Zeile Lesen ******************************************************/
      line  = confReLine (cnf, False);
      
#if 0
      DEBUG2 ("confRead: line =(%s)\n", line != NULL ? line : "NULL");
#endif

      /* Parameter filtern ************************************************/
      if (line != NULL)
	Ureadarg (line, arg0, 0, trennz);
      else
	arg0 [0] = 0;
      
#if 0
      DEBUG3 ("confRead: Arg0=(%s), trennz=(%c)\n", arg0, trennz);
#endif
      
      /*** Dateiende, ohne den Parameter zu finden ************************/
      if (line == NULL)
	{
	  break;
	}

      /*** Eine neue Section erreicht, ohne den Parameter zu finden *******/
      else if (arg0 [0] == '[')
	{
	  break;
	}

      /****** Parameter gefunden **********************************************/
      else if (strcmp (arg0, param) == 0)
	{
	  found = True;
	  break;
	}
    }

  return found ? line : NULL;
}

int confAppend (Conf_t cnf, char trennz, char * section, int argc, char * argv[])
{
  char * value;
  int i;
  long pos;
  int cnt;
  if (argc == -1)
  {
    for (cnt=0; argv[cnt] != NULL; cnt++);
  }
  else
  {
    cnt = argc;
  }

  if (cnf == NULL)
    return False;
  
  /****** Trennzeichen behalten ***********************************************/
  cnf->Trennz [0] = trennz;

  /****** Beim Schliessen muss die Datei aktualisiert werden ******************/
  cnf->WriteFlag = True;

  /****** Section anlegen, falls nicht vorhanden ******************************/
  pos = confSectPos (cnf, section);

  /****** 1. Fall: Section nicht vorhanden ************************************/
  if (pos < 0)
    {
      pos = confAddSect (cnf, section);

      if (pos < 0)
	{
	  ERROR2 ("confAppend: Kann Sektion (%s) nicht anlegen\n", section);
	  return False;
	}
    }
  else
    /****** Nach der Sektion vorruecken ***************************************/
    confReLine (cnf, False);

  /****** Parameter wird sofort nach Sektionsanfang geschrieben **************/
  confAddStr (cnf, argv[0], cnf->CurrentPos, False);

  DEBUG1 ("confAppend: Add Arguments");

  /****** Werte des Parameters anhaengen **********************************/
  for (i=1; i<cnt && (value = argv[i]) != NULL; i++)
    {
      confAddStr (cnf, cnf->Trennz, cnf->CurrentPos, False);
      confAddStr (cnf, value, cnf->CurrentPos, True);
    }

  DEBUG1 ("confAppend: Add new line");

  /* NewLine ******************************************************************/
  confAddStr (cnf, "\n", cnf->CurrentPos, False);

#if 1
  DEBUG1 ("confAppend: Ende");
#endif

  return True;
}

int confDelSect (Conf_t cnf, char * section)
{
  long pos;
  char * line;

#if 0
  DEBUG2 ("confDelSect: Sektion (%s)\n", section);
#endif

  /****** Falls die Section nicht vorhanden ist, nichts zu machen *************/
  pos = confSectPos (cnf, section);
  if (pos < 0)
    {
      return True;
    }

  /****** Konfigurationsdatei geaendert ***************************************/
  cnf->WriteFlag = True;

  /****** Sektion loeschen ****************************************************/
  confDelLine (cnf, pos);

  /****** Inhalt der Sektion loeschen *****************************************/
  for (;True;)
    {
      /** Naechste Zeile einlesen *********************************************/
      line = confReLine (cnf, False);
      /** Dateiende ***********************************************************/
      if (line == NULL)
	break;
      /** neue Sektion erreicht ***********************************************/
      if (line [0] == '[')
	break;
      /** eingelesene Zeile leoschen ******************************************/
      confDelLine (cnf, pos);
    }

#if 0
  DEBUG1 ("confDelSect: ende");
#endif
  return True;
}

int confDelToEnd (Conf_t cnf, int pos)
{
  if (cnf == NULL || pos < 0 || pos > cnf->FileSize)
    return False;

  cnf->FileEntry[pos] = 0;
  cnf->FileSize       = strlen (cnf->FileEntry);
  cnf->FileEntry      = realloc (cnf->FileEntry, cnf->FileSize + 1);
  cnf->WriteFlag      = True;

  return True;
}
