/************************************************************************
 *  (C)opyright 1996/1997 by Khamis Abu El-Komboz                       *
 *  xkhfm 1.0 can be copied and distributed freely for any              *
 *  non-commercial purposes. xkhfm can only be incorporated             *
 *  into commercial software with the permission of the current author. *
 *  Mail: khamis@multix.de                                              *
 ************************************************************************/
/*
 * $Author: khamis $
 * $Id: dchain.c,v 1.1 1996/10/28 03:37:18 khamis Exp khamis $
 * $Source: /home/xkhfm/khlib/khmisc/RCS/dchain.c,v $
 */

#define IS_EXTERN
#define EXTERN extern

#include <stdlib.h>
#include <stdio.h>

#include "defines.h"
#include "Umem.h"
#include "Udchain.h"

int UdchCopy (Dchain_t item, void * data,
	      int (* ClCopy) (void *, void *))
{
  DEBUG1 ("UdchCopy: Daten kopieren");

  return ClCopy (item->client, data);
}

Dchain_t UdchMake (void * data, void * (* ClMake) (void *))
{
  void * client = NULL;
  Dchain_t l;

#if 0
  DEBUG1 ("IdchMake begin");
#endif
  /****** Falls ClientFkt vorhanden, dann aufrufen ****************************/
  if (ClMake != NULL)
    {
      client = ClMake (data);
      /** Null-Pointer, d.h. Eintrag verweigert *******************************/
      if (client == NULL)
	return NULL;
    }

  /****** Liste erstellen *****************************************************/
  l = gcalloc (sizeof (dchain_t), __LINE__, __FILE__);
  /****** Den gewonnenen Pointer der Liste zuordnen ***************************/
  l->client = client;

  return l;
}

int UdchDestroy (Dchain_t lst, int (* ClDelete) (void *))
{
  if (lst->client != NULL)
    {
      if (! ClDelete (lst->client))
	return False;
    }
    
  MFree (lst, sizeof (dchain_t));

  return True;
}

int UdchDelete (Dchain_t l, int (* ClDelete) (void *))
{
  Dchain_t prev, next;

  if (l == NULL)
    return False;

  /****** 1. Listenelement aus der Liste schneiden ****************************/
  prev = l->prev,
  next = l->next;
  
  /****** 2. Entferntes Element freigeben *************************************/
  if (! UdchDestroy (l, ClDelete))
    return False;

  /****** Teile der Liste miteinander verbinden *******************************/
  /****** Die Teile werden nur dann miteinander verbunden, solange das letzte */
  /****** Element nicht entfernt wurde ****************************************/
  if (prev != NULL)
      prev->next = next;

  if (next != NULL)
    next->prev = prev;  

  return True;
}

int UdchEmpty (Dchain_t top, int (* ClDelete) (void *))
{
  Dchain_t l;
  Dchain_t prev, next;

  if (top == NULL)
    return False;


  /****** Das oberste Element solange entfernen, bis die Liste durchlaufen ****/
  /****** ist *****************************************************************/
  for (l = top->next; l != NULL; l = top->next)
    {
      if (! UdchDelete (l, ClDelete))
	return False;
    }

  return True;
}

int UdchAppend (Dchain_t top,
		Dchain_t newLst,
		void * data,
		int (* ClCompare) (void *, void *))
{
  Dchain_t next, pos;

  if (top == NULL || newLst == NULL)
    return False;

  /****** Position ermitteln **************************************************/
  /****** Die Liste wird solange durchlaufen, bis das Ende gefunden wird oder */
  /****** der ClCompare liefert einen Wert >= 0, dies bedeutet, dass das      */
  /****** Element vor dem Listeneintrag eingetragen wird                      */
  for (pos = top; pos != NULL && pos->next != NULL; pos = pos->next)
    {
      if (ClCompare (pos->next->client, data) >= 0)
	break;
    }

  /****** In Liste einfuegen und die Knoten reseten ***************************/
  if (pos != NULL)
    {
      next = pos->next;
    }

  /****** Dies waere das erste Element in der Liste ***************************/
  else
    {
      next = NULL;
    }

  /****** das neue Listenelement in der Liste einbinden ***********************/
  newLst->next = next;
  newLst->prev = pos;
  
  if (pos  != NULL) pos->next = newLst;
  if (next != NULL) next->prev = newLst;

  return True;
}

Dchain_t  UdchAdd (Dchain_t top,
		   void * data,
		   void * (* ClMake  ) (void *),
		   int    (* ClCompare) (void *, void *))
{
  Dchain_t newLst;

  /****** Listeneintrag erstellen *********************************************/
  newLst = UdchMake (data, ClMake);

  if (newLst == NULL)
    return NULL;

  /****** Listeneintrag an der richtigen Stelle einfuegen *********************/
  UdchAppend (top, newLst, data, ClCompare);

  return newLst;
}

Dchain_t UdchFind (Dchain_t top,
		   void * data,
		   int (* ClCompare) (void *, void *))
{
  Dchain_t p;

  /****** Liste leer **********************************************************/
  if (top == NULL || data == NULL)
    return NULL;

  /****** Liste durchlaufen und den Namen vergleichen, bis der Name gefunden **/
  /****** ist *****************************************************************/
  for (p = top->next; p != NULL; p = p->next)
    {
      /** An die User-Funktion uebergeben *************************************/
      if (ClCompare (p->client, data) == 0)
	return p;
    }

  return NULL;
}

int UdchCount (Dchain_t top)
{
  Dchain_t p;
  int count = 0;

  /****** Liste durchlaufen und Elemente zaehlen ******************************/
  if (top != NULL)
    for (count = 0, p = top->next; p != NULL; p = p->next) count ++;

  return count;
}

int UdchPos (Dchain_t top, Dchain_t lst)
{
  Dchain_t p = NULL;
  int Pos = 0;

  /****** Liste durchlaufen und Elemente zaehlen ******************************/
  if (top != NULL)
    for (Pos = 0, p = top->next; p != NULL && p != lst; p = p->next) Pos ++;

  return Pos;
}

int UdchItemExists (Dchain_t top,
		    void * data,
		    int (* ClCompare) (void *, void *))
{
  /****** ob ein Element vorhanden ist, wird einfach gesucht ******************/
  return UdchFind (top, data, ClCompare) != NULL;
}

int UdchChange (Dchain_t top,
		Dchain_t liste,
		void * newValue,
		int  (* ClChange)  (Dchain_t, void *),
		int  (* ClCompare) (void *, void *)
		)
{
  Dchain_t newLst, prev, next;
  int ret;

  /****** 0. Parameter ueberpruefen *******************************************/
  if (top == NULL || liste == NULL || newValue == NULL)
    return False;

  /****** 1. Falls der Wert sich nicht aendert, dann nichts tun ***************/
  if (ClCompare (liste->client, newValue) == 0)
    return True;

  /****** 2. Liste aendern ****************************************************/
  if (ClChange (liste, newValue) == False)
    return False;

  /****** 3. Listenelement aus der Liste rausnehmen ***************************/
  prev = liste->prev;
  next = liste->next;

  /****** 4. Teile der Liste verbinden ****************************************/
  if (prev != NULL) prev->next = next;
  if (next != NULL) next->prev = prev;

  /****** 5. Listenelement an der richtigen Stelle einfuegen ******************/
  ret = UdchAppend (top, liste, newValue, ClCompare);

  return ret;
}

Dchain_t UdchTop (Dchain_t lst)
{
  Dchain_t top = NULL;

  /****** Liste nach oben durchlaufen und den Kopf ermitteln ******************/
  if (lst != NULL)
    for (top = lst; top != NULL && top->prev != NULL; top = top->prev);

  return top;
}

Dchain_t UdchLast (Dchain_t lst)
{
  Dchain_t last = NULL;

  /****** Liste nach oben durchlaufen und den Kopf ermitteln ******************/
  if (lst != NULL)
    for (last = lst; last != NULL && last->next != NULL; last = last->next);

  return last;
}

int UdchEmptyAndDel (Dchain_t top, int (* ClDelete) (void *))
{
  return
    UdchEmpty  (top, ClDelete) &&
    UdchDelete (top, ClDelete);
}
