/************************************************************************
 *  (C)opyright 1996/1997 by Khamis Abu El-Komboz                       *
 *  xkhfm 1.0 can be copied and distributed freely for any              *
 *  non-commercial purposes. xkhfm can only be incorporated             *
 *  into commercial software with the permission of the current author. *
 *  Mail: khamis@multix.de                                              *
 ************************************************************************/
/*
 * $Author: khamis $
 * $Id: fileutil.c,v 1.1 1997/01/07 08:39:42 khamis Exp $
 * $Source: /home/khamis/source/xkhfm/khlib/khmisc/RCS/fileutil.c,v $
 */

#define IS_EXTERN
#define EXTERN extern

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>

#ifdef DOS
#include <io.h>
#include <dir.h>
#endif

#include "defines.h"
#include "Ustrutil.h"
#include "Ufileutil.h"

int C_X( int u, int g, int m) 
{
  return(
	 (
	  ( C_O_X( m) != 0 && u == UserId ) ||
	  ( C_G_X( m) != 0 && g == GroupId ) ||
	  ( C_A_X( m) != 0 )
	  )  ||
	 ( (
	    ( C_O_X( m) != 0 ) ||
	    ( C_G_X( m) != 0 ) ||
	    ( C_A_X( m) != 0 ) ||
	    ( S_ISDIR( m)    )
	    ) &&
	  UserId == 0
	  )
	 );
}

int C_R( int u, int g, int m) 
{
  return(
	 ( C_O_R( m) != 0 && u == UserId ) ||
	 ( C_G_R( m) != 0 && g == GroupId ) ||
	 ( C_A_R( m) != 0 )              ||
	 UserId == 0
	 );
}

int C_W( int u, int g, int m) 
{
  return(
	 ( C_O_W( m) != 0 && u == UserId ) ||
	 ( C_G_W( m) != 0 && g == GroupId ) ||
	 ( C_A_W( m) != 0              ) ||
	 ( UserId == 0                   )
	 );
}



int UFileExists (char * file)
{
  struct stat st;
  if (lstat( file, &st) != 0) return( FALSE);
  return( TRUE);
}

int UFileIsDir (char * dr)
{
  struct stat st;
  /*
   * mode : DIR_SYMLINK, if file is symbolic link, locks for it's realname and
   *        return its type
   * otherwise for symbolic links will be symbolic type returned
   */
  if( lstat( dr, &st) == -1) return( FALSE);
  
  if (S_ISDIR (st.st_mode)) return TRUE ;
  
  return FALSE ;
}

int UFileIsFile (char *file)
{
  /* mode = TRUE  : directory isn't a LINK SYMBOL   */
  /* mode = FALSE : directory can a LINK to another */
  struct stat st;
  if (lstat( file, &st) != 0) return FALSE ;

  if (S_ISREG (st.st_mode)) return TRUE ;

  return FALSE ;
}

int UFileOwner( char * file)
{
  struct stat st;
  if( lstat( file, &st) == -1) return( FALSE);
  return( st.st_uid);
}

int UFileGroup( char * file)
{
  struct stat st;
  if( lstat( file, &st) == -1) return( FALSE);
  return( st.st_gid);
}

int UFileCanWrite( char * file)
{
  struct stat st;
  if( lstat( file, &st) != 0) return( FALSE);
  return( C_W( st.st_uid, st.st_gid, st.st_mode) );
}

int UFileCanRead( char * file)
{
  struct stat st;
  if( lstat( file, &st) != 0) return( FALSE);
  return( C_R( st.st_uid, st.st_gid, st.st_mode) );
}

int UFileCanExec( char * file)
{
  struct stat st;
  if( lstat( file, &st) != 0) return( FALSE);
  return( C_X( st.st_uid, st.st_gid, st.st_mode) );
}

unsigned short UFileStat( char * file)
{
  struct stat st;
  if( lstat( file, &st) != 0) return( FALSE);
  return st.st_mode;
}

mode_t UGetUmask (int mode) 
{
  mode_t msk = umask (0777) & 0777; /* Only 8 bits */
  umask ((short) msk);              /* set the umask to it's old value */

#ifdef DOS
  return msk;
#else
  msk = (mode
	  ? msk | 0111		/* Ohne Exec-Rechte **/
	  : msk);		/* mit  ........... **/
  return ~msk & 0777;		/* Rechte ableiten */
#endif
}

/***********************************************************************/
/******** test if a file is ascii-text, data or executable file ********/
/***********************************************************************/

#define FILEBUF	512

int UFileType( char * file)
{
  unsigned char buf[ FILEBUF];
  int
    file_h,
    i,
    len,
    not_ascii = 0,
    is__ascii = 0;

  if( UFileCanExec (file))
    return EXEC_FILE;   /* file can execute */
  
  file_h = open (file, O_RDONLY);
  if( file_h < 0) return( FALSE);
  
  len = read( file_h, buf, FILEBUF);
  
  for( i=0; i<len; i++) {
    if( is_ascii( buf[i] & 0x00ff) ||
       buf[i] == '\n' ||
       buf[i] == '\t' ||
       buf[i] == '' )
      is__ascii++;
    else {
      not_ascii++;
      break;
    }
  }
  close( file_h);
  
  return ( not_ascii > 0 ? DATA_FILE : TEXT_FILE);
}

long UFileSizeCount (char * file, int  * files, int  * dirs,
		    int mode)
{
  
  /***** mode = RECURSIVE_MODE : recursive size <for directories> ******/
  /***** files, is the number of access files/directories	    */
  
  struct stat st;
  long size = 0;
  
  if (lstat( file, &st) != 0) return 0;  /* error */
  
  if (S_ISREG( st.st_mode))  		/* only when a regular file */
    {
      (* files) ++;
      return st.st_size;
    }

  if( ! (mode & RECURSIVE_MODE) ) return (0);  /* not recursive size */
  
  if (S_ISDIR( st.st_mode))  /* size of the directory */
    {    
      DIR * dir;
      struct dirent * d_file;
      char tmp[MAXPATHLEN];
      
      if( (dir = opendir( file)) == NULL) return 0; /* error */
      
      dirs[0] ++;				/* increase directories */
      
      while( ( d_file = readdir( dir)) != NULL)
	{
	  
	  if (strcmp( d_file->d_name, "." ) == 0 ||
	      strcmp( d_file->d_name, "..") == 0
	      ) continue;
	  
	  sprintf( tmp, "%s/%s", file, d_file->d_name);
	  size += UFileSizeCount (tmp, files, dirs, mode); /* recursive */
	}
      
      closedir( dir);
    }
  
  return( size);
}

long UFileSize (char * file)
{
  int files = 0, dirs = 0;
  return UFileSizeCount (file, &files, &dirs, RECURSIVE_MODE);
}

int UCreateDir( char * dr, int mode)
{
  int err_;
  if (UFileIsDir (dr)) err_ = FALSE;      /* nothing to do */
  else
    {
      err_ = mkdir (dr, (short) mode);
    }
  return (err_);
}

/* search for file in a given path */
char * FileGetWithPath (char* path, char* file)
{
  static char iconPath[MAXPATHLEN];
  char tmpPath [MAXPATHLEN];
  char * p;
  int i;

#ifdef DOS
#define SPLIT_CHR	';'
#else
#define SPLIT_CHR	':'
#endif
  
  strcpy (iconPath, file);

  if (UFileExists (iconPath))
    return iconPath;

  for (p = path; p != NULL && *p; )
    {
      for (i=0; *p && *p != SPLIT_CHR; p++)
	    {
	      tmpPath[i++] = *p;
	    }
      tmpPath[i] = 0;
      if (*p == SPLIT_CHR) p++;
      
      if (UFileExists (UPathAppend (tmpPath, file)))
	{
	  strcpy (iconPath, UPathAppend (tmpPath, file));
	  break;
	}
    }

  return iconPath;
}

char * UPathAppend (char * path1, char * path2)
{
  static char buf [MAXPATHLEN];
  int pos = max (0,  strlen (path1) - 1);

  /* if the second path is having absolute pathname */
  if (path2[0] == RootDir)
    {
      strcpy (buf, path2);
      return buf;
    }

  strcpy (buf, path1);

  /** delete the Slashes at then end **/
  while (pos >= 0 && buf [pos] == RootDir)
    {
      buf [pos --] = 0;
    }

  /** die zwei Pfade zusammenverbinden **/
  sprintf (buf + strlen (buf), "%c%s", RootDir, path2);

  return buf;
}

char * UDirName (char * path)
{
  static char buf [MAXPATHLEN];
  int pos = max (0,  strlen (path) - 1);

  DEBUG2 ("UFilePath: (%s)\n", path);

  strcpy (buf, path);

  /****** Das root Verzeichnis ************************************************/
  if (strcmp (buf, "/") == 0)
    return buf;

  /****** Die Slashes am Ende entfernen ***************************************/
  /****** wie "/usr/lib/" ==> "/usr/lib" **************************************/
  for (;pos >= 0 && buf [pos] == RootDir; pos --)
    {
      buf [pos] = 0;
    }

  /****** Letztes Wort entfernen **********************************************/
  /****** wie "/usr/lib" ==> "/usr/" ******************************************/
  for (; pos >= 0 && buf [pos] != RootDir; pos --)
    {
      buf [pos] = 0;
    }

  /****** Falls nur ein Wort im Pfad, dann "." einfuegen **********************/
  /****** wie "usr" oder "." oder ".." ==> "" *********************************/
  if (strlen (buf) == 0)
    strcpy (buf, ".");

  return buf;
}

/******************************************************************************/
/* mode: FIND_CASESENSITIVE don't ignore littel and big latters		      */
/*       False:             ignore littel and big letters		      */
/******************************************************************************/
long UFileFindString (int fdesc, char * str, int mode) 
{
  char 
    * tmp,
    gr_chr = '*',
    gr_emp = '-';
  int    
    i,
    value,
    blck,                            /* the readed buffer block */
    len      = strlen(str),          /* strlength of battern to find */
    buf_len  = max (1024*10, len*2),    /* buffer block */
    str_pos  = 0;                    /* ring */
  
  long
    old_pos = lseek (fdesc, 0, SEEK_CUR),  /* save position */
    pos     = old_pos,                      /* graph view */
    size    = lseek (fdesc, 0, SEEK_END);  /* size of file */
  
  if (len <= 0) return -1;                     /* no string */
  
  lseek( fdesc, old_pos, SEEK_SET);      /* to current position */
  
  tmp = gmalloc (buf_len + 1, __LINE__, __FILE__);                 /* buffer length */
  
  for( value=0; (blck=read (fdesc, tmp+str_pos, buf_len-str_pos)) > 0; ) 
    {
      
      if (mode & FIND_CASESENSITIVE)
	{
	  value = StringFindNStr (tmp, str, min (blck+str_pos, buf_len), len);
	}
      else
	{
	  value = StringFindCaseStr (tmp,
				     str,
				     min (blck+str_pos, buf_len),
				     len,
				     mode
				     );
	}
      
      if (value > 0) break;           /* battern found */

      pos += buf_len-str_pos;         /* graph view */
      
      str_pos = len;                  /* slide the ring */
      
      for (i=0; i<str_pos; i++) tmp[i] = tmp[i+buf_len-str_pos];  /* ring */
    }

  MFree (tmp, buf_len + 1);
  
  if( value == 0) 
    {
      lseek (fdesc, old_pos, SEEK_SET);
      return -1;
    }
  
  pos = lseek (fdesc,
	       value-min(blck+str_pos,buf_len)-1,
	       SEEK_CUR
	       ); 			/* current position */

  pos = max (0, pos);
  
  return pos;
}

/******************************************************************************/
/* Eine Zeile aus einer Datei lesen                                           */
/* mode = True: Die Backslashes werden nicht entfernt			      */
/******************************************************************************/
int UFileReadLine (int f, char * buf, int len, int mode)
{
#define READ_BL		256
  char tmp_buf [READ_BL], o_ch = 0;
  int l, i, j, end_ = FALSE, l_pos, comment = False;
  long pos = lseek (f, 0, SEEK_CUR);
  
#if 0
  DEBUG2 ("Zeile=%c", '(');
#endif

  for (l_pos = j = 0; TRUE; ) 
    {
      l = read (f, tmp_buf, READ_BL);
      for (i = 0; i<l; i++, l_pos++) 
	{
	  /* Ende der Zeile wird genau dann, wenn kein Verbindungszeichen *****/
	  /* vorliegt *********************************************************/
	  if (tmp_buf [i] == '\n' && o_ch != '\\')
	    end_ = TRUE;                /* read to end of line */

	  if (end_) break;

#if 0
	  DEBUG2 ("%c", tmp_buf[i]);
#endif

	  if (buf != NULL && ! comment && j < len-1)
	    {
	      /** Zeile nur dann formatiert einlesen, wenn das mode auf True  */
	      /** gesetzt ist.						      */
	      if (mode)
		{
		  /* Newline **************************************************/
		  if (o_ch == '\\' && tmp_buf[i] == 'n')
		    buf [j++] = '\n';
		  
		  /* Tab ******************************************************/
		  else if (o_ch == '\\' && tmp_buf[i] == 't')
		    buf [j++] = '\t';
		  
		  /* Das Zeichen danach ***************************************/
		  else if (o_ch == '\\' && tmp_buf[i] != '\n')
		    buf [j++] = tmp_buf [i];
		  
		  /* Falls ein Kommentarzeichen gefunden wird, vormerken ******/
		  else if (tmp_buf [i] == '#')
		    comment = True;

		  /* Falls kein '\\' oder '\n' ********************************/
		  else if(tmp_buf [i] != '\\' && tmp_buf[i] != '\n')
		    buf [j++] = tmp_buf[i];
		}
	      else
		{
		  if (tmp_buf [i] != '\n')
		    buf [j++] = tmp_buf [i];
		}
	    }

	  /* zuletzt eingelesenes Zeichen vormerken ***************************/
	  o_ch = tmp_buf [i];
	}
      if (l <= 0 || end_) break;
    }

#if 0
  DEBUG3 ("), pos=(%i), l_pos=(%i)\n", pos, l_pos);
#endif

  if (buf != NULL)
    buf [j] = 0;

  /****** Auf Anfang der naechsten Zeile positionieren ************************/
  lseek (f, pos+l_pos+1, SEEK_SET);	/* correct position */

  return (l == 0 && j == 0)		/* end of file reached */
         ? -1				/* mark for end of file */
         : j;				/* string length, was reading */
}

int UFileCopyPart (int fsrc, int fdesc, long beg, long end)
{
#define BUFSIZE		2048
  char * buf [BUFSIZE];
  int
    len,
    size = end - beg + 1;

#if 1
  DEBUG3 ("FileCopyPart: schreibe von (%i) bis (%i)\n", beg, end);
#endif

  /****** Nichts zu kopieren **************************************************/
  if (size <= 0)
    return False;

  /****** An die richtige Stelle positionieren ********************************/
  lseek (fsrc, beg, SEEK_SET);

  /****** Kopieren ************************************************************/
  for (;
       (len = read (fsrc, buf, min (BUFSIZE, size))) > 0; /* Lesen ************/
       size -= len)
    {
#if 0
      DEBUG2 ("FileCopyPart: schreibe (%s)\n", buf);
#endif

      /** In die zweite Datei schreiben ***************************************/
      if (write (fdesc, buf, len) != len)
	return False;
    }

  DEBUG1 ("UFileCopyPart: ende");

  return True;
}

char * UGetTmpFile (char *path, char * prefix)
{
  return (char*) tempnam (path, prefix);
}
