/************************************************************************
 *  (C)opyright 1996/1997 by Khamis Abu El-Komboz                       *
 *  xkhfm 1.0 can be copied and distributed freely for any              *
 *  non-commercial purposes. xkhfm can only be incorporated             *
 *  into commercial software with the permission of the current author. *
 *  Mail: khamis@multix.de                                              *
 ************************************************************************/
/*
 * $Author: khamis $
 * $Id: copy.c,v 1.1 1997/01/19 23:10:36 khamis Exp khamis $
 * $Source: /usr2/xkhfm/khlib/khsrc/copy.c,v $
 */
#define IS_EXTERN
#define EXTERN		extern

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include "Xinc.h"

#include "defines.h"
#include "Umenu.h"
#include "Ustrutil.h"
#include "doregex.h"
#include "dodirent.h"
#include "Ufileutil.h"
#include "util.h"
#include "dir.h"
#include "copy.h"

#define BLOCKSIZE	10240

#ifndef O_BINARY
#define O_BINARY 0x0
#endif
  
int cpyRemove (Dchain_t src, CopyInfo_t cpy_info)
{
  int ret;
  if (! (cpy_info->mode & (CPY_ERROR|CPY_CANCEL)) &&
      (cpy_info->mode & CPY_MOVE))
    {
      ret = remove (dirName (src));
      if (ret != 0)
	{
	  cpy_info->mode |= CPY_ERROR_AVAIL;
	  if (cpy_info->ErrProc != NULL)
	    cpy_info->mode |= cpy_info->ErrProc (src, cpy_info->ClData);
	}
    }

  return cpy_info->mode;
}

/* return 0, when copied file */
int CopyFile (Dchain_t src, Dchain_t dest, CopyInfo_t cpy_info)
{  
  char block   [BLOCKSIZE];
  unsigned long size = 0, copied = 0;
  int Aborted = False, Del_Src = False;
  int src_h, dest_h;
  int len, i, j, ret;
  int new_owner, new_grp;
  int ch_ownergrp = (cpy_info->mode & CPY_SAME_OWNER ||
		     cpy_info->mode & CPY_SAME_GROUP);
  mode_t rights;

  DEBUG2 ("CopyFile: Source=(%s)\n", dirName (src));
  DEBUG2 ("CopyFile: Destin=(%s)\n", dirName (dest));
  
  rights = ((cpy_info->mode & CPY_SAME_RIGHTS)
	    ? dirRights (src) & 0777
	    : ~UGetUmask (True)) & 0777;	/* no exec (--x--x--x) rights */

  /* source and dest. are identical */
  if (strcmp (dirName (dest), dirName (src)) == 0)
    {
      cpy_info->mode |= CPY_ERROR|CPY_ERROR_AVAIL;
      return cpy_info->mode;
    }

  DEBUG1 ("CopyFile: new_owner");

  /* if same user & group id needed */
  new_owner = (cpy_info->mode & CPY_SAME_OWNER) ? dirUserId (src) : UserId;
  new_grp   = (cpy_info->mode & CPY_SAME_GROUP) ? dirGrpId  (src) : GroupId;

  /* 
   * if SKIP_EXISTING return if not a directory
   */
  if (dirExists (dest) &&
      dirType (dest) != TYPE_DIR &&
      (cpy_info->mode & CPY_SKIP_EXISTING))
    {
      return cpy_info->mode;
    }

  /*
   * when overwrite mode disabled, ask for overwriting
   */
  if (! (cpy_info->mode & CPY_OVERWRITE) && dirExists (dest))
    {
      switch (dirType (src))
	{
	case TYPE_REG:
	case TYPE_EXE:
	case TYPE_LNK:
	  /* Client Ask Procedure to answer overwrite */
	  if (cpy_info->AskOverwriteProc != NULL)
	    {
	      int ret = cpy_info->AskOverwriteProc (src, dest,
						    cpy_info->ClData);

	      switch (ret)
		{
		  /* overwrite this file */
		case CPY_YES:
		  /* nothing to do */
		  break;

		  /* overwrite all */
		case CPY_OVERWRITE:
		  cpy_info->mode |= CPY_OVERWRITE;
		  break;

		  /* cancel copying */
		case CPY_CANCEL:
		  cpy_info->mode |= CPY_CANCEL | CPY_ERROR_AVAIL;
		  return cpy_info->mode;
		  break;

		  /* unknown answer */
		default:
		  cpy_info->mode |= CPY_ERROR | CPY_ERROR_AVAIL;
		  return cpy_info->mode;
		  break;
		}
	    }
	  else
	    {
	      cpy_info->mode |= CPY_ERROR | CPY_ERROR_AVAIL;
	      return cpy_info->mode;
	    }
	  break;
	}
    }

  /*
   * copy links (only the regular files)
   */
  if ((cpy_info->mode & CPY_COPY_LINKS) &&
      dirType (src) == TYPE_LNK)
    {
      if (dirTypeOfRealPath (src) == TYPE_DIR)
	return cpy_info->mode;

      DEBUG4 ("CopyFile: copy link (%s) of (%s) --> (%s)\n",
	      dirRealPath (src),
	      dirName (src),
	      dirName (dest));

      src     = dirAdd (NULL, dirRealPath (src));
      Del_Src = True; /* delete the local created source file */
    }

  /* copy file or create directory (supporting list as switching) */
  switch (dirType (src))
    {
      /* create a new directory */
    case TYPE_DIR:

      DEBUG2 ("CopyFile: create directory (%s)\n", dirName (dest));

      if (cpy_info->ViewProc != NULL)
	{
	  cpy_info->mode |= cpy_info->ViewProc (50, 100, cpy_info->ClData);
	  if (cpy_info->mode & CPY_CANCEL)
	    return cpy_info->mode;
	}

      if (! (cpy_info->mode & CPY_SAME_RIGHTS))
	rights = ~ UGetUmask (False) & 0777;	/* with exec rights */

      ret = dirCreateDir (dest, rights);
      if (ret != 0)
	{
	  cpy_info->mode |= CPY_ERROR | CPY_ERROR_AVAIL;
	  break;
	}

      /* new owner/group */      
      if (ch_ownergrp) 
	dirChngOwnerGrp (dest, new_owner, new_grp);

      if (cpy_info->ViewProc != NULL)
	{
	  cpy_info->mode |= cpy_info->ViewProc (100, 100, cpy_info->ClData);
	  if (cpy_info->mode & CPY_CANCEL)
	    return cpy_info->mode;
	}

      break;

      /* copy regular file */
    case TYPE_REG:
    case TYPE_EXE:

      DEBUG2 ("CopyFile: copy file (%s)\n", dirName (dest));

      src_h  = open (dirName (src),  O_RDONLY|O_BINARY);
      if (src_h < 0)
	{
	  cpy_info->mode |= CPY_ERROR | CPY_ERROR_AVAIL;
	  break;
	}
      
      if (! (cpy_info->mode & CPY_SAME_RIGHTS) )
	{
	  rights &= ~0111;
	}
      
      dest_h = open (dirName (dest),
		     O_WRONLY|O_BINARY|O_CREAT|O_TRUNC,
		     rights);
      if (dest_h < 0)
	{
	  cpy_info->mode |= CPY_ERROR | CPY_ERROR_AVAIL;
	  break;
	}
      
      /** file size */
      size   = dirSize (src);

      /** copy file now */
      for (copied = 0; (len = read (src_h, block, BLOCKSIZE)) > 0; )
	{
	  if (write (dest_h, block, len) != len)
	    {
	      close  (dest_h);
	      close  (src_h);
	      remove (dirName (dest));
	      cpy_info->mode |= CPY_DISK_ERROR | CPY_CANCEL;

	      Aborted = True;
	      break;
	    }
	  
	  copied += len;
	  
	  /* view position of coping with help of a client program */
	  if (cpy_info->ViewProc != NULL)
	    {
	      cpy_info->mode |=
		cpy_info->ViewProc (copied, size, cpy_info->ClData);
	      if (cpy_info->mode & CPY_CANCEL)
		return cpy_info->mode;
	    }
	}

      close (dest_h);
      close (src_h);
      
      if (Aborted)
	break;
      else
	{
	  if (cpy_info->mode & CPY_SAME_RIGHTS)
	    chmod (dirName (dest), rights);
	}


      /* change owner and group if needed */
      if( ch_ownergrp) 
	dirChngOwnerGrp (dest, new_owner, new_grp);

      break;

      /* link symbolik links to the same link of the source */
    case TYPE_LNK:
      DEBUG3 ("CopyFile: symlink (%s) --> (%s)\n",
	      dirRealPath (src),
	      dirName (dest));

      if (cpy_info->ViewProc != NULL)
	{
	  cpy_info->mode |= cpy_info->ViewProc (50, 100, cpy_info->ClData);
	  if (cpy_info->mode & CPY_CANCEL)
	    return cpy_info->mode;
	}

      /* remove old link/file */
      if (dirExists (dest))
	remove (dirName (dest));
      
      if (symlink (dirRealPath (src), dirName (dest)) != 0)
	cpy_info->mode |= CPY_ERROR | CPY_ERROR_AVAIL;

      if (cpy_info->ViewProc != NULL)
	{
	  cpy_info->mode |= cpy_info->ViewProc (100, 100, cpy_info->ClData);
	  if (cpy_info->mode & CPY_CANCEL)
	    return cpy_info->mode;
	}

      break;

      /* file type not supported */
    default:

      DEBUG3 ("CopyFile: unknown type of (%s) = 0x%x\n",
	      dirName (src),
	      dirType (src));

      cpy_info->mode |= CPY_ERROR | CPY_ERROR_AVAIL;
      break;
    }

  /* By moving files, delete source file */
  if (dirType (src) != TYPE_DIR)
    cpy_info->mode |= cpyRemove (src, cpy_info);

  /* delete view position of coping with help of a client program */
  if (cpy_info->ViewProc != NULL)
    {
      cpy_info->mode |= cpy_info->ViewProc (0, 100, cpy_info->ClData);
      if (cpy_info->mode & CPY_CANCEL)
	return cpy_info->mode;
    }

  /* delete local created file structure */
  if (Del_Src)
    dirDelete (src);

  DEBUG1 ("CopyFile: end");

  return cpy_info->mode;
}

int RecursiveCopy (Dchain_t src,
		   Dchain_t dst,
		   CopyInfo_t cpy_info)
{
  Dchain_t list, p, dstList, srcList;
  int  ret = FALSE;

  /* check regular expression */
  if (strlen (cpy_info->donotCopy) > 0 &&
      IsRegular (cpy_info->donotCopy, basename (dirName (src))))
    {
      DEBUG2 ("RecusiveCopy: Donot Copy: Skip (%s)\n", dirName (src));
      return cpy_info->mode;
    }
  if (strlen (cpy_info->copy) > 0 &&
      strcmp (cpy_info->copy, "*") != 0 &&
      ! IsRegular (cpy_info->copy, basename (dirName (src))))
    {
      DEBUG2 ("RecusiveCopy: CopyFilter: Skip (%s)\n", dirName (src));
      return cpy_info->mode;
    }

  /* view files with client program , there will be copy now */
  cpy_info->mode |= cpy_info->ViewFilesProc (src, dst, cpy_info->ClData);
  if (cpy_info->mode & CPY_CANCEL)
    return cpy_info->mode;

  /* directories can only be copied by recursive copying flag */
  if (dirType (src) == TYPE_DIR && ! (cpy_info->mode & CPY_RECURSIVE))
    {
      cpy_info->mode |= CPY_ERROR | CPY_ERROR_AVAIL;
      return cpy_info->mode;
    }

  /* if asking mode is enabled, get answer */
  if ((cpy_info->mode & CPY_ASK) && cpy_info->AskProc != NULL)
    {
      cpy_info->mode &= ~ CPY_NO;			/* delete old answers */
      /* ask for coping */
      cpy_info->mode |= cpy_info->AskProc (src, dst, cpy_info->ClData);
      if (cpy_info->mode & (CPY_CANCEL|CPY_NO))
	return cpy_info->mode;
    }
  
  /* try to use the rename procedure for files on one filesystem */
  if (cpy_info->mode & CPY_MOVE)
    {
      int ret;
      ret = rename (dirName (src), dirName (dst));
      if (ret == 0)
	return cpy_info->mode;
    }

  /* call the copying procedure */
  cpy_info->mode |= CopyFile (src, dst, cpy_info);

  /* if an error is expected */
  if (cpy_info->mode & CPY_ERROR)
    {
      cpy_info->mode &= ~ CPY_ERROR;		/* delete error flag */

      /* call user defined message dialog */
      if (cpy_info->ErrProc != NULL)
	{
	  cpy_info->mode |= cpy_info->ErrProc (dst, cpy_info->ClData);
	}
      return cpy_info->mode;
    }
      
  /* if directory and recursive copying */
  if ((dirType (src) == TYPE_DIR) && (cpy_info->mode & CPY_RECURSIVE))
    {
      /* make head for a new directory list */
      list = dirAdd (NULL, "TOP");

      /* read directory entries */
      dirReadDir (list, dirName (src),
		  cpy_info->filter == NULL || strlen (cpy_info->filter) == 0
		  ? "*"
		  : cpy_info->filter,
		  False);
      
      for (p=UdchNext (list);
	   p!=NULL;
	   p=UdchNext (p))
	{
	  /* append path for src. and dest. files */
	  srcList = dirAdd (NULL,
			    UPathAppend (dirName (src), dirName (p)));
	  dstList = dirAdd (NULL,
			    UPathAppend (dirName (dst), dirName (p)));
	  
	  /* call recursive copying for every entry */
	  cpy_info->mode |= RecursiveCopy (srcList,
					   dstList,
					   cpy_info);

	  /* delete created list entries */
	  dirDelete (srcList);
	  dirDelete (dstList);
	  
	  /* if break mode is seted, break copying  */
	  if (cpy_info->mode & CPY_CANCEL)
	    break;
	}
      
      /* delete created list of directories */
      dirDelete (list);

      /* Delete directories by moved flag */
      cpy_info->mode |= cpyRemove (src, cpy_info);
      
      return cpy_info->mode;
    }  /* if CPY_RECURSIVE  */

  return cpy_info->mode;
}  /* rec */

