/************************************************************************
 *  (C)opyright 1996/1997 by Khamis Abu El-Komboz                       *
 *  xkhfm 1.0 can be copied and distributed freely for any              *
 *  non-commercial purposes. xkhfm can only be incorporated             *
 *  into commercial software with the permission of the current author. *
 *  Mail: khamis@multix.de                                              *
 ************************************************************************/
/*
 * $Author: khamis $
 * $Id: dirTree.c,v 1.1 1997/01/07 08:39:42 khamis Exp $
 * $Source: /home/khamis/source/xkhfm/khlib/khsrc/RCS/dirTree.c,v $
 */
#define IS_EXTERN
#define EXTERN		extern

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "Xinc.h"

#include "defines.h"
#include "Umenu.h"
#include "Ustrutil.h"
#include "doregex.h"
#include "dodirent.h"
#include "dir.h"

DirTree_t dirTreeMake (char * Name)
{
  DirTree_t l = gcalloc (sizeof (dirTree_t), __LINE__, __FILE__);

#if 0
  DEBUG2 ("dirTreeMake: Name=(%s)\n", Name);
#endif

  /****** Strukturfelder auf Defaultwerte festlegen **************************/
  l->Name     = StrReplace (l->Name, Name != NULL ? Name : ".");
  l->PathName = StrReplace (l->PathName, l->Name);
  l->Level    = 0;
  l->subTree  = NULL;
  return l;
}

int dirTreeDest (DirTree_t lst)
{
#if 0
  DEBUG3 ("dirTreeDest: (%s) PN=(%s)\n", lst->Name, lst->PathName);
#endif

  StrReplace (lst->Name,     NULL);
  StrReplace (lst->PathName, NULL);

  /* delete Subtrees */
  if (lst->subTree != NULL)
    {
      UdchEmptyAndDel (lst->subTree, (int (*)(void*)) dirTreeDest);
    }

  MFree (lst, sizeof (dirTree_t));
  return True;
}

int dirTreeComp (DirTree_t liste, char * Name)
{
  int ret;

  /****** Liste leer **********************************************************/
  if (liste == NULL || Name == NULL)
    return 0;

  /****** Namen vergleichen *************************************************/
  ret = strcmp (liste->Name, Name);

  return ret;
}

Dchain_t dirTreeAdd   (Dchain_t top, char * Name)
{
  Dchain_t newLst;

#if 0
  DEBUG2 ("dirTreeAdd: Name (%s) in Liste eintragen\n", Name);
#endif

  /****** Es kann sein, dass dies das erste Element ist, ansonsten darf *******/
  /****** pos kein NULL Pointer sein ******************************************/
  /****** keine double-deutige Nummern ****************************************/
  if (UdchItemExists (top, Name, (int (*) (void*, void*)) dirTreeComp))
    {
#if 0
      ERROR2 ("dirTreeAdd: Name [%s] is already existing\n",
	      Name);
#endif
      return NULL;
    }

  /****** In die Liste eintragen **********************************************/
  newLst = UdchAdd (top, Name,
		    (void * (*) (void *)) dirTreeMake,
		    (int    (*) (void *, void *)) dirTreeComp);

  /* set level to the new created list*/
  if (newLst != NULL && top != NULL)
    {
      dirTreeSetLevel  (newLst, dirTreeLevel (top));
    }

  return newLst;
}

int dirTreeCopy (DirTree_t lst1, DirTree_t lst2)
{

  /****** Namen kopieren ******************************************************/
  memcpy (lst2, lst1, sizeof (dirTree_t));
  
  return True;
}

int dirTreeSaveName (Dchain_t lst, char * Name)
{
  DirTree_t dirTree;

  if (lst == NULL || Name == NULL || strlen (Name) == 0)
    return False;

  dirTree = (DirTree_t) UdchClient (lst);

  dirTree->Name = StrReplace (dirTree->Name, Name);

  return True;
}

char * dirTreeName (Dchain_t lst)
{
  return ((DirTree_t) UdchClient (lst))->Name;
}

char * dirTreePathName (Dchain_t lst)
{
  return ((DirTree_t) UdchClient (lst))->PathName;
}

#if 0
int dirTreeLevel (Dchain_t lst)
{
  return ((DirTree_t) UdchClient (lst))->Level;
}
#endif

int dirTreeSetLevel (Dchain_t lst, int num)
{
  ((DirTree_t) UdchClient (lst))->Level = num;
  return True;
}

int dirTreeSelect (Dchain_t tree, int mode)
{
  if (mode )
    ((DirTree_t)UdchClient(tree))->Info |= TREE_SELECTED;
  else
    ((DirTree_t)UdchClient(tree))->Info &= ~TREE_SELECTED;

  return True;
}

int dirTreeSoftSelect (Dchain_t tree, int mode)
{
  if (mode )
    ((DirTree_t)UdchClient(tree))->Info |= TREE_SOFT_SELECTED;
  else
    ((DirTree_t)UdchClient(tree))->Info &= ~TREE_SOFT_SELECTED;

  return True;
}

int dirTreeCopySelect (Dchain_t tree, int mode)
{
  if (mode )
    ((DirTree_t)UdchClient(tree))->Info |= TREE_COPY_SELECTED;
  else
    ((DirTree_t)UdchClient(tree))->Info &= ~TREE_COPY_SELECTED;

  return True;
}

/* make a subtree and increment the parent level to the subtree */
Dchain_t dirTreeAddSubTree (Dchain_t tree)
{
  DirTree_t trCl;

  if (tree == NULL)
    return NULL;
  
  trCl = UdchClient (tree);
  if (trCl->subTree == NULL)
    {
      /* Make head of the new sublist */
      trCl->subTree = dirTreeAdd (NULL, "TOP");

      /* Set Parent to the Children */
      dirTreeSetParent (trCl->subTree, tree);

      /* Set Level */
      dirTreeSetLevel  (trCl->subTree, dirTreeLevel (tree) + 1);
    }

  return trCl->subTree;
}

int dirTreeChng (Dchain_t lst, char * newName)
{
  DirTree_t tl;

#if 0
  DEBUG2 ("dirTreeChng: newName (%s)", newName != NULL ? newName : "NULL");
#endif

  if (newName == NULL || strlen (newName) == 0)
    return False;

  tl = (DirTree_t) UdchClient (lst);

  /****** ueberpruefen, ob Aendern moeglich ist *******************************/
  /* NOTHING TO DO */

  /****** die neue Nummer in die Liste kopieren *******************************/
  tl->Name = StrReplace (tl->Name, newName);

#if 0
  DEBUG1 ("dirTreeChng: ende");
#endif

  return True;
}

/* building the tree to the last name, so we kann view the fullpathname as */
/* a tree */
int dirTreeBuild (Dchain_t top, char * PathName, int mode)
{
  static char Filter [] = "*";
  Dchain_t actTree = top, saveTree;
  DirTree_t t;
  char MyPN     [MAXPATHLEN];
  char actualPN [MAXPATHLEN];
  char * pName;
  int i, j;

  /* save the directory name */
  strcpy (MyPN, PathName);
  
  /* get the Current Work Directory */
  if (MyPN [0] != '/')
    {
#if 0
      if (chdir (PathName) != 0)
	return False;
#endif
      if (getcwd (MyPN, sizeof (MyPN)) == NULL)
	{
	  perror ("dirTreeBuild");
	  ERROR1 ("dirTreeBuild: can not get current working directory");
	  exit (1);
	}
      DEBUG2 ("dirTreeBuild: dir(%s)\n", MyPN);
    }
  
  actualPN [0] = 0;
  for (i = 0; dirGetName (MyPN, i) != NULL && actTree != NULL; i++)
    {
      pName = dirGetName (MyPN, i);

      if (strlen (actualPN) > 1)
	strcat (actualPN, "/");
      strcat (actualPN, pName);

      DEBUG3 ("dirTreeBuild: pName(%s), i(%i)\n", pName, i);

      /* find out, if the file already exists */
      saveTree = UdchFind (actTree,
			   pName,
			   (int (*)(void*,void*))dirTreeComp);

      /* create a new tree entry */
      if (saveTree == NULL)
	{
	  saveTree = dirTreeAdd (actTree, pName);

	  /* Set Parent to the new Child */
	  if (saveTree != NULL)
	    {
	      t = UdchClient (saveTree);
	      t->PathName = StrReplace (t->PathName, actualPN);
	      dirTreeSetParent (saveTree,
				(actTree != NULL
				 ? dirTreeParent (actTree)
				 : NULL));
	    }
	}

      /* set the subitem to the parent */
      actTree = saveTree;

      /* An error expected */
      if (actTree == NULL)
	break;

      /* create subtree if not availiable */
      if (((DirTree_t) UdchClient (actTree))->subTree == NULL)
	{
	  ((DirTree_t) UdchClient (actTree))->subTree =	    
	    dirTreeAddSubTree (actTree);
	}
      actTree = ((DirTree_t) UdchClient (actTree))->subTree;

      dirTreeRead (actTree, actualPN, Filter, mode);
    }

  return True;
}

int dirTreeRefrRec (Dchain_t top, int mode)
{
  Dchain_t dirTr, nTr;

  if (top == NULL)
    return False;

  for (dirTr = UdchNext (top);
       dirTr != NULL;
       )
    {
      nTr = UdchNext (dirTr);

      /* directory doesn't more exist */
      if (! UFileIsDir (dirTreePathName (dirTr)))
	{
	  UdchDelete (dirTr, (int (*)(void*)) dirTreeDest);
	}
      /* refresh recursive */
      else
	{
	  dirTreeRefrRec (((DirTree_t)UdchClient (dirTr))->subTree,
			  mode);
	}
      
      dirTr = nTr;
    }

  return True;
}
  
int dirTreeRefresh (Dchain_t top, int mode)
{

  
  DEBUG2 ("dirTreeRefresh: (%s)\n", dirTreePathName (top));

  /* directory doesn't more exist */
  if (! UFileIsDir (dirTreePathName (top)))
    {
      UdchDelete (top, (int (*)(void*)) dirTreeDest);
    }
  else
    {
      dirTreeRefrRec (((DirTree_t)UdchClient (top))->subTree, mode);
    }
  return True;
}

int dirTreeRead (Dchain_t top, char * PathName, char * Filter, int mode)
{
  Dchain_t dir, pdir;
  Dchain_t tree;
  char PN [MAXPATHLEN];
  int ret = False;

  /* read directories in the actual directory */
  dir = dirAdd (NULL, "TOP");
  ret = dirReadDir (dir, PathName, Filter, READ_DIR_ONLY);
  if (!ret)
    return ret;

#if 0
  DEBUG2 ("dirTreeRead; PathName (%s)\n", PathName);
#endif

  for (pdir = UdchNext (UdchNext (UdchNext(dir)));
       pdir != NULL;
       pdir = UdchNext (pdir))
    {
      DirTree_t dirTree;


      tree = dirTreeAdd (top, dirName (pdir));
      
      /* name doesn't exist */
      if (tree != NULL)
	{
	  dirTree = UdchClient (tree);

	  /* Set Parent to the new Child */
	  dirTreeSetParent (tree, (top != NULL ? dirTreeParent (top) : NULL));

	  /* store the fullpath of the directory */
	  strcpy (PN, PathName);
	  if (strcmp (PathName, "/") != 0) strcat (PN, "/");
	  strcat (PN, dirTreeName (tree));
	  dirTree->PathName = StrReplace (dirTree->PathName, PN);

	  /* the last icon on the tree list */
	  if (UdchNext (pdir) == NULL)
	    dirTree->Info |= TREE_LAST_ICON;
	}
      
      if (mode & TREE_READ_RECURSIVE)
	{
	  Dchain_t sTree;
	  if (tree == NULL)
	    {
	      tree = UdchFind (top,
			       dirName (pdir),
			       (int (*)(void*,void*)) dirTreeComp);
	      if (tree == NULL)
		continue;
	    }
	  sTree = ((DirTree_t)UdchClient (tree))->subTree;
	  if (sTree == NULL)
	    {
	      sTree =
		((DirTree_t)UdchClient(tree))->subTree = 
		  dirTreeAddSubTree (tree);
	    }
	  if (sTree != NULL)
	    ret = dirTreeRead (sTree,
			       dirTreePathName (tree),
			       Filter,
			       mode);
	}
    }

#if 0
  DEBUG4 ("dirTreeRead; PathName (%s), entry=(%i), dircount (%i)\n",
	  PathName,
	  UdchCount (top),
	  UdchCount (dir));
#endif

  /* Delete the temporary Direcory List */
  UdchEmptyAndDel (dir, (int (*)(void*))dirDest);

  return ret;
}

int dirTreeCount (Dchain_t tree)
{
  Dchain_t tt;
  int count = 0;

  if (tree == NULL)
    return count;

#if 0
  DEBUG2 ("dirTreeCount: count (%s)\n", dirTreeName (UdchNext (tree)));
#endif

  for (tt = UdchNext (tree);
       tt != NULL;
       tt = UdchNext (tt))
    {
      count ++;
      count += dirTreeCount ( ((DirTree_t)UdchClient (tt))->subTree);
    }

  return count;
}

int dirTreeHaveSubDir (Dchain_t tree)
{
  DirTree_t dTree = (DirTree_t) UdchClient (tree);

  /* Test if the directory have subdirectories */
  if ((dTree->Info & TREE_TESTED_SUBDIR) == False)
    {
      Dchain_t dir;
      int ret;

      /* Have already subdirectories */
      if (dTree->subTree != NULL && UdchNext (dTree->subTree) != NULL)
	dTree->Info |= TREE_HAVE_SUBDIR;
      else
	{
	  dir = dirAdd (NULL, "TOP");
	  ret = dirReadDir (dir, dTree->PathName, "*", READ_DIR_ONLY);
	  if (!ret)
	    return False;
	  
	  if (UdchNext (UdchNext (UdchNext(dir))) != NULL)
	    dTree->Info |= TREE_HAVE_SUBDIR;
	  else
	    dTree->Info &= ~ TREE_HAVE_SUBDIR;
	  
	  /* Delete the temporary Direcory List */
	  UdchEmptyAndDel (dir, (int (*)(void*))dirDest);
	}

      dTree->Info |= TREE_TESTED_SUBDIR;
    }

  return dTree->Info & TREE_HAVE_SUBDIR;
}

int dirTreeSetParent  (Dchain_t tree, Dchain_t parent)
{
  ((DirTree_t)UdchClient(tree))->Parent = parent;
}

int dirTreeVLineAvail (Dchain_t tree, int level)
{
  Dchain_t parent;
  int i;

  if (level < 1)
    return False;

  for (parent = tree, i=0;
       parent != NULL && i < level;
       i++, parent = ((DirTree_t) UdchClient(parent))->Parent)
    {
    }

  
  /* Error expected */
  if (parent == NULL)
    {
      WARNING3 ("dirTreeVLineAvail: can't get Information for level (%i) "
		"on Path [%s]\n",
		level,
		((DirTree_t) UdchClient(tree))->PathName);
      return True;
    }

  /* if no items follow the Parent */
  else if  (UdchNext(parent) == NULL)
    return False;

  /* there are some items following in the same list */
  else
    return True;
}
