/************************************************************************
 *  (C)opyright 1996/1997 by Khamis Abu El-Komboz                       *
 *  xkhfm 1.0 can be copied and distributed freely for any              *
 *  non-commercial purposes. xkhfm can only be incorporated             *
 *  into commercial software with the permission of the current author. *
 *  Mail: khamis@multix.de                                              *
 ************************************************************************/
/*
 * $Author: khamis $
 * $Id: remove.c,v 1.1 1996/10/16 06:52:58 khamis Exp $
 * $Source: /home/xkhfm/khlib/khsource/RCS/remove.c,v $
 */
#define IS_EXTERN
#define EXTERN		extern

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include "Xinc.h"

#include "defines.h"
#include "Umenu.h"
#include "Ustrutil.h"
#include "doregex.h"
#include "dodirent.h"
#include "Ufileutil.h"
#include "util.h"
#include "remove.h"

/* mode = RM_RECURSIVE : delete directory recursive */  
int RecursiveRemove (char * file,
		     char * RemFilter,
		     char * DonotRemFilter,
		     unsigned long (* ViewState) (char*, void*),
		     unsigned long (* AskProc) (char *, void*),
		     unsigned long (* ErrProc) (char *, void*),
		     unsigned long mode,
		     void * ClData)
{
  struct stat st;
  DIR * dir;
  struct dirent * d;
  char filename  [MAXPATHLEN];   /* to give file with there path */
  int ret;

  /* removing canceled */
  if (mode & REM_CANCEL)
    return mode;

  DEBUG2 ("RecusiveRemove: (%s)\n", file);

  if (ViewState != NULL)
    {
      mode |= ViewState (file, ClData);
      if (mode & REM_CANCEL)
	return mode;
    }

  /* Cannot get the file stat */
  if (MY_stat( file, &st) != 0)
    {
      mode |= REM_ERROR_AVAIL;

      if (ErrProc == NULL)
	return mode;

      /* call the client program to handle the error */
      mode |= ErrProc (file, ClData);

      if (mode & (REM_CANCEL|REM_NO))
	return mode;
    }

  /* remove a directory */
  if (S_ISDIR (st.st_mode)) 
    {
      if (mode & REM_RECURSIVE) 
	{

	  DEBUG2 ("remove: recursive remove (%s)\n", file);

	  if( ( dir = X_opendir( file)) == NULL) 
	    {
	      mode |= REM_ERROR_AVAIL;

	      /* call the client program to handle the error */
	      if (ErrProc != NULL)
		mode |= ErrProc (file, ClData);

	      return mode;
	    }
	  
	  while( ( d = X_readdir( dir)) != NULL) 
	    {
	      
	      if (strcmp( d->d_name, "." ) == 0 ||
		  strcmp( d->d_name, "..") == 0)
		continue;

	      /* file with path */
	      sprintf( filename, "%s/%s", file, d->d_name);

	      /* recursive call of remove */
	      mode |= RecursiveRemove (filename,
				       RemFilter,
				       DonotRemFilter,
				       ViewState,
				       AskProc,
				       ErrProc,
				       mode,
				       ClData);
	      
	      if (mode & REM_CANCEL)
		break;
	    }
	  
	  X_closedir( dir); 
	}  /* if mode */
      
#ifndef XENIX

#if 0	      
      if (! (mode & REM_ALL) && (mode & REM_ASK) && AskProc != NULL)
	{
	  mode |= AskProc (file, ClData);
	  
	  if (mode & REM_CANCEL)
	    return mode;
	  if (mode & REM_NO)
	    {
	      mode &= ~ REM_NO;
	      return mode;
	    }
	}
#endif

      if (mode & REM_EMPTY_DIR)
	{
	  ret = rmdir (file);
	  
	  if (ret != 0)
	    {
	      mode |= REM_ERROR_AVAIL;

	      /* Error, when no filter active and removing
	       * directories requered
	       */
	      if ((mode & REM_EMPTY_DIR) &&
		  ErrProc != NULL &&
		  strcmp (RemFilter, "*") == 0 &&
		  strlen (DonotRemFilter) == 0)
		mode |= ErrProc (file, ClData);
	    }
	}
#else
#endif
    }  /* if S_ISDIR */

  /* remove file */
  else
    {

      if (strlen (DonotRemFilter) > 0 &&
	  IsRegular (DonotRemFilter, basename (file)))
	{
	  DEBUG2 ("RecusiveRemove: Donot RemFilter: Skip (%s)\n", file);
	  return mode;
	}
      
      if (strlen (RemFilter) > 0 &&
	  strcmp (RemFilter, "*") != 0 &&
	  ! IsRegular (RemFilter, basename (file)))
	{
	  DEBUG2 ("RecusiveRemove: RemFilter: Skip (%s)\n", file);
	  return mode;
	}
      
      /* asking for remove files only */
      if (! (mode & REM_ALL) && (mode & REM_ASK) && AskProc != NULL)
	{
	  mode |= AskProc (file, ClData);
	  
	  if (mode & (REM_CANCEL|REM_NO))
	    {
	      mode &= ~ REM_NO;
	      return mode;
	    }
	}
      
      /* remove file */
      if (remove (file) != 0)
	{
	  mode |= REM_ERROR_AVAIL;

	  if (ErrProc != NULL)
	    mode |= ErrProc (file, ClData);
	}
    }
  
  return mode;
}  /* rec */

