/************************************************************************
 *  (C)opyright 1996/1997 by Khamis Abu El-Komboz                       *
 *  xkhfm 1.0 can be copied and distributed freely for any              *
 *  non-commercial purposes. xkhfm can only be incorporated             *
 *  into commercial software with the permission of the current author. *
 *  Mail: khamis@multix.de                                              *
 ************************************************************************/
/*
 * $Author: khamis $
 * $Id: sort.c,v 1.1 1997/01/07 08:39:42 khamis Exp $
 * $Source: /home/khamis/source/xkhfm/khlib/khsrc/RCS/sort.c,v $
 */

#define EXTERN extern
#define IS_EXTERN

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#ifdef DOS
#include <dos.h>
#endif

#include "defines.h"
#include "Umem.h"
#include "Ustrutil.h"
#include "sort.h"

#ifndef max
#	define max( x, y) ((x) > (y) ? (x) : (y))
#endif

#ifndef TRUE
#	define TRUE 1
#endif

void OTT_qsort (void * ClData,
		int l, int r,
		int (* compare) ( const int, const int, void *),
		int (* swap   ) ( const int, const int, void *))
{
  int i, j, m, v;
  
  while( r > l) {
    i = l-1;
    j = v = r; /* pivot element */
    
    m = (r+l) / 2;				/* median of three elements */
    
    if( compare( l, r, ClData) > 0) swap( l, r, ClData);
    if( compare( l, m, ClData) > 0) swap( l, m, ClData);
    if( compare( r, m, ClData) < 0) swap( m, r, ClData);
    
    /* merge two parts together */
    while( TRUE) {
      do { i++; } while( compare( i, v, ClData) < 0); 
      do { j--; } while( compare( j, v, ClData) > 0);
      if( j < l || i > r) {
	printf("Error in qsort: i=[%i] j=[%i] r=[%i] l=[%i]\n",
	       i, j, r, l);
	sleep (1);
      }
      if( i >= j) break;
      swap( i, j, ClData);
    } /* while TRUE */
    
    swap( i, r, ClData);
    
    if( (i - 1 - l) <= ( r - i - 1)) {
      OTT_qsort (ClData, l, i-1, compare, swap);
      l = i+1;
    }
    else {
      OTT_qsort(ClData, i+1, r, compare, swap);
      r = i-1;
    }
  } /* while */
  
}  /* OTT_qsort */

#if 0
int bin_srch( void * v,     /* the element to search */
	      void * kette, /* pointer on the head of the kette */
	      int cnt,      /* count of elements */
	      int len,      /* length of a element */
	      int ( * compare) ( const void *, const void *)) {
    int x, l = 0, r = cnt - 1, value;
    char * tmp;

    do {
	x = ( l + r) / 2;
	tmp = (char *) kette+len*x;
	value = compare( (void *) v, (void *) tmp);
	if( value < 0) r = x - 1; else l = x + 1;
        } while( value != 0 && l <= r );

    if( value == 0) return( x);

    return( -1);
    }

int KHA_search( void * v,     /* the element to search */
		int beg,      /* count of elements */
		int end,      /* length of a element */
		int ( * compare) ( const void *, const int)) {
    int x, l = beg, r = end - 1, value;
    char * tmp;

    do {
	x = ( l + r) / 2;
	value = compare( (void *) v, (int) x);
	if( value < 0) r = x - 1; else l = x + 1;
	} while( value != 0 && l <= r );

    if( value == 0) return( x);

    return( -1);
    }

long power( long a, int n) {
	long ret = 1;
	if( n <= 0) return( 0);
	while( n>0) {
		if( n>2 && n % 2 == 0)
			 { n /= 2; ret = ret * power( a, n); }
		else { n--; ret = ret * a; }
		}
	return( ret);
	}

/* String Matching */
/* Alg. from Knuth, Morris und Pratt */
/* With this alg. it is possible to test the string matching in only */
/* O( n + m) time */
/* n-Faktor is the length of the string in it to search */
/* m-Faktor to calculate the suffix and prefix in the string to find */

int strfind( char * s, char * p)
{
  return( nstrfind (s, p, strlen(s), strlen(p)) );
}

void F( char * s, int f[], int m)
{
  int
    i, j;
  f [0] = 0;
  f [1] = 1;
  i = 3;
  while( i <= m) 
    {
      j = f[i-2];
      while( j != 0 && s[j-1] != s[i-2]) j = f[j-1];
      f[i-1] = j + 1;
      i++;
    }
}

int nstrfind (char * s, char * p, int n, int m)
{
  int
    i = 1,
    j = 1,
    * f = (int *) gmalloc ( max(m,2) * sizeof(int), __LINE__, __FILE__),
    found = 0;

  F (p, f, m);			/* prefix and suffix values */
  
  while (i <= n)
    {
      while (j != 0 && p[j-1] != s[i-1]) j = f[j-1];
      if( j == m)
	{
	  found = 1; break;
	}
      else i++, j++;
    }
  MFree (f, max(m,2) * sizeof(int));
  return (found ? i - j + 1 : 0);
}

int case_strfind (char * s, char * p, int n, int m, int mode)
{
  int
    i = 1,
    j = 1,
    * f = (int *) gmalloc ( max(m,2) * sizeof(int), __LINE__, __FILE__),
    found = 0;

  F( p, f, m);			/* prefix and suffix values */

  while( i <= n)
    {
      if (mode & CASES_MODE)
	{
	  while( j != 0 && p[j-1] != s[i-1]) j = f[j-1];
	}
      else
	{
	  while (j != 0
		 && ch_lowercase (p[j-1]) != ch_lowercase(s[i-1])
		 )
	    j = f[j-1];
	}

      if (j == m)
	{
	  found = 1;
	  break;
	}
      else i++, j++;
    }
  MFree (f, max(m,2) * sizeof(int));
  return( found ? i - j + 1 : 0);
}

#ifdef TEST_F
#include "dodefs.h"
#include "doglobal.h"
#include "docfg.h"
#include "doscr.h"

void test_F (void)
{
  struct frame_t * f;
  struct input_t * in ;
  char
    str [65],
    tmp [10];
  int
    width = max_x,
    down  = 7,
    x     = 0,
    y     = (max_y-down)/2,
    num [44],
    i;

  f = make_frame (x, y, width, down, " Test F-Function ", leer,
		  F_B_NORMAL,
		  NOMODE) ;
  in = make_input (f, 1, CUR_MODE);
  f_putstr (f, 1, 2, "Text: ", f->color, NOMODE);

  in->x       [0] = 7;
  in->y       [0] = 2;
  in->str     [0] = str;
  str	      [0] = 0;
  in->str_len [0] = 39;
  in->inp_len [0] = 39;

  view_frame (f);

  while (f_getstr (in) != ESC)
    {

      F (str, num, strlen (str)) ;

      f_putstr (f, 1, 4, leer, f->color, FILL_MODE);
      f_putstr (f, 1, 5, leer, f->color, FILL_MODE);
      for (i=0; str [i] != 0; i++)
	{
	  f_putch  (f, 2 + i * 2, 4, str [i], f->color, NOMODE);
	  sprintf  (tmp, "%i", num [i]);
	  f_putstr (f, 2 + i * 2, 5, tmp, f->color, NOMODE);
	}
      view_frame (f);
    }
  del_input  (in);
  hide_frame (f, HIDE_MODE|DELETE_MODE);
}
#endif

#endif
