/************************************************************************
 *  (C)opyright 1996/1997 by Khamis Abu El-Komboz                       *
 *  xkhfm 1.0 can be copied and distributed freely for any              *
 *  non-commercial purposes. xkhfm can only be incorporated             *
 *  into commercial software with the permission of the current author. *
 *  Mail: khamis@multix.de                                              *
 ************************************************************************/
/*
 * $Author: khamis $
 * $Id: list.c,v 1.1 1996/12/05 20:41:42 khamis Exp khamis $
 * $Source: /usr2/xkhfm/xkhlib/xkhmisc/RCS/list.c,v $
 */
#define IS_EXTERN
#define EXTERN		extern

#include <stdio.h>
#include <stdlib.h>
#include "Xinc.h"

#include "defines.h"
#include "about.h"
#include "quit.h"
#include "Umenu.h"
#include "Uxutil.h"

#ifndef XmFONTLIST_DEFAULT_TAG
#define XmFONTLIST_DEFAULT_TAG XmSTRING_DEFAULT_CHARSET
#endif

int listAddItem (Widget lst, char * item, int mode)
{
  int ret;
  char * value;
  int count, i;
  XmString * items;
  XmString NewItem;

  /****** Sortiert einfuegen, dies erfolgt in mehrere Schritte ****************/
  if (mode & LIST_SORTED)
    {
      count = listCount (lst);

      /****** Liste der Elemente ermitteln ************************************/
      XtVaGetValues (lst,
		     XmNitems,     &items,
		     NULL);
      
      /****** Passende Stelle suchen ******************************************/
      for (i=0; i<count; i++)
	{
	  XmStringGetLtoR (items[i], XmFONTLIST_DEFAULT_TAG, &value);
	  ret = strcmp (item, value);	  
	  XtFree (value);

	  /****** Stelle gefunden *********************************************/
	  if (ret <= 0) break;
	}
      i ++;
    }
  else
    {
      i = 0; /* Am ende der Liste einfuegen ***********************************/
    }

#if 0
  DEBUG3 ("listAddItem: Item (%s) an Stelle (%i) einfuegen\n", item, i);
#endif

  /****** Element vorbereiten *************************************************/
  NewItem = XmStringCreateSimple (item);

  /****** Element in der Liste einfuegen **************************************/
  XmListAddItem    (lst, NewItem, i);

#if 0
  DEBUG1 ("listAddItem: Item schon eingetragen");
#endif

  /****** Dies hat zufolge, das die Callback-Prozedur der Selektion aufgerufen*/
  /****** wird. Also VORSICHT *************************************************/
  if (mode & LIST_SELECTED)
    {
      XmListSelectItem (lst, NewItem, True);
      XmListSetPos     (lst, max (0, i - 2));
    }

  XmStringFree (NewItem);

  return True;
}

int listAddPos (Widget lst, char * item, int Pos, int mode)
{
  int ret;
  char * value;
  int count, i;
  XmString * items;
  XmString NewItem;

  /****** Element vorbereiten *************************************************/
  NewItem = XmStringCreateSimple (item);

  count = listCount (lst);

  /****** Element in der Liste einfuegen **************************************/
  XmListAddItem (lst, NewItem, min (count + 1, Pos));

  /****** Dies hat zufolge, das die Callback-Prozedur der Selektion aufgerufen*/
  /****** wird. Also VORSICHT *************************************************/
  if (mode & LIST_SELECTED)
    {
      XmListSelectItem (lst, NewItem, True);
      XmListSetPos     (lst, max (0, Pos - 2));
    }

  XmStringFree (NewItem);

  return True;
}

int listAddTo (Widget list, Menu_t mnuList)
{
  int i;

  /******* Falls keine Liste vorhanden ****************************************/
  if (mnuList == NULL || mnuList[0].menu_item == NULL)
    return True;

  /******* Liste durchlaufen **************************************************/
  for (i=0; mnuList[0].menu_item[i].label != NULL; i++)
    {
      /******* Listeneintrag in einen XmString umwandeln  *********************/
      XmString list_item = XmStringCreateSimple (mnuList[0].menu_item[i].label);

      /******* Element am Anfang der Liste einfuegen **************************/
      XmListAddItem  (list, list_item, 0);

      /******* String freigeben ***********************************************/
      XmStringFree   (list_item);

      /******* Falls Callbackaufrufe vorhanden ********************************/
      if (mnuList[0].menu_item[i].prog != NULL)
	XtAddCallback (list, 
		       XmNbrowseSelectionCallback /*XmNdefaultActionCallback*/ ,
		       (XtCallbackProc) mnuList[0].menu_item[i].prog,
		       (XtPointer) i);
    }
  
  if (mnuList[0].prog != NULL)
    XtAddCallback (list, 
		   XmNbrowseSelectionCallback /*XmNdefaultActionCallback*/ ,
		   (XtCallbackProc) mnuList[0].prog,
		   (XtPointer) i);
  return True;
}

int listCount (Widget lst)
{
  int count;

  if (lst == NULL)
    return 0;

  XtVaGetValues (lst,
		 XmNitemCount, &count,
		 NULL);

#if 0
  DEBUG2 ("listCount: ListCount (%i)\n", count);
#endif

  return count;
}

Widget listCreate (Widget parent, Menu_t mnuList, int mode)
{
  Widget list;
  Arg args[20];
  int i, n, j;

#ifdef DEBUG
  if (mnuList != NULL)
    {
      DEBUG2 ("listCreate: %s\n", mnuList[0].menu_bar);
    }
  else
    {
      DEBUG1 ("listCreate: keine Eintraege");
    }
#endif

  /******* Anzahl der sichtbaren Zeilen ***************************************/
  n = 0;
  XtSetArg (args[n], XmNvisibleItemCount, mode); n++;

  /******* Liste erstellen ****************************************************/
  list = XmCreateScrolledList (parent,
			       mnuList != NULL ? mnuList[0].menu_bar : NULL,
			       args, n);

  /******* Falls Listeneintr. vorhanden, dann eintragen ***********************/
  listAddTo (list, mnuList);

/*
  XtManageChild (list);
*/

  return list;
}

int listDelPos (Widget lst, int pos, int count)
{
  int ItemCount;

  /****** Liste ist noch nicht erzeugt ****************************************/
  if (lst == NULL || count == 0)
    return True;

  /****** Anzahl der Listeneintraege ermitteln ********************************/
  ItemCount = listCount (lst);

  /****** Listeneintraege loeschen ********************************************/
  XmListDeleteItemsPos (lst, min (count, ItemCount - pos + 1), max (1, pos));

  return True;
}

int listSelItem (Widget lst, char * item)
{
  /****** Erweiterten String erzeuegen ****************************************/
  XmString xmItem = XmStringCreateSimple (item);

  /****** Element in der Liste selektieren ************************************/
  XmListSelectItem (lst, xmItem, True);
  /****** Element im Fenster zeigen *******************************************/
  XmListSetItem    (lst, xmItem);

  /****** Element aus der Liste entfernen *************************************/
  XmStringFree (xmItem);

  return True;
}
