/************************************************************************
 *  (C)opyright 1996/1997 by Khamis Abu El-Komboz                       *
 *  xkhfm 1.0 can be copied and distributed freely for any              *
 *  non-commercial purposes. xkhfm can only be incorporated             *
 *  into commercial software with the permission of the current author. *
 *  Mail: khamis@multix.de                                              *
 ************************************************************************/
/*
 * $Author: khamis $
 * $Id: message.c,v 1.1 1996/10/20 22:32:33 khamis Exp khamis $
 * $Source: /home/xkhfm/xkhlib/xkhmisc/RCS/message.c,v $
 */

#define IS_EXTERN
#define EXTERN			extern

#include <stdio.h>
#include <stdlib.h>

#include "Xinc.h"		/* XIncludes mit enthalten */
#include <Xm/Protocols.h>
#include "defines.h"
#include "Umessage.h"

typedef struct
{
  void (* ClOk)     (void);
  void (* ClCancel) (void);
  void (* ClHelp)   (void);

  Widget dialog;
  int    Return;
  int    Mode;
} message_t, * Message_t;

/******************************************************************************/
/* die folgenden keinen Prozeduren sind ein Teil von der Funktion MessageBox **/
/******************************************************************************/
void OkPressed (Widget dialog, Message_t msg)
{
  XtDestroyWidget (dialog);
  msg->dialog = NULL;
  msg->Return = 0;

  if (msg->ClOk != NULL) msg->ClOk ();
}

void CancelPressed (Widget dialog, Message_t msg)
{
  XtDestroyWidget (dialog);
  msg->dialog = NULL;
  msg->Return = 1;
  if (msg->ClCancel != NULL) msg->ClCancel ();
}

void HelpPressed (Widget dialog, Message_t msg)
{
  msg->Return = 2;

  /****** Falls Dialog bei Hilfe geschlossen werden soll **********************/
  if (msg->Mode & CLOSE_ON_HELP)
    {
      XtDestroyWidget (dialog);
      msg->dialog = NULL;
    }
}

void WaitForCloseMessageBox (Message_t msg)
{
  XEvent event;

#if 0
  DEBUG1 ("Wait for MessageBox");
#endif
  
#ifdef HAVE_THREAD
  if (mainloop == NULL)
    {
      thr_suspend (*mainloop);
    }
#endif

  for (;;)
    {
      if (msg->dialog == NULL)
	break;

      XtAppNextEvent (app_context, &event);
      XtDispatchEvent(&event);
    }

#ifdef HAVE_THREAD
  if (mainloop == NULL)
    {
      thr_continue (*mainloop);
    }
#endif

#if 0
  DEBUG1 ("MessageBox closed, leaving the SubEventLoop");
#endif
}

/******************************************************************************/
/********* Hauptfunktion ******************************************************/
/******************************************************************************/
int MessageBox (Widget parent,
		char * title,
		char * text,
		int type,
		int mode,
		void (* ok_prog    ) (void),
		void (* cancel_prog) (void))
{
  Message_t msg;
  Widget
    btnCancel = NULL,
    btnOk = NULL,
    btnHelp = NULL;
  Atom WM_DELETE_WINDOW;
  Arg args[10];
  XmString str;
  XmString titel;
  int n, ReturnValue;

  msg = gcalloc (sizeof (message_t), __LINE__, __FILE__);
  msg->ClOk     = ok_prog;
  msg->ClCancel = cancel_prog;
  msg->Return   = -1;
  msg->Mode     = mode;

  str   = XmStringCreateLtoR (text,  XmSTRING_DEFAULT_CHARSET);
  titel = XmStringCreateLtoR (title, XmSTRING_DEFAULT_CHARSET);

#if 0
  DEBUG1 ("MessageBox");
#endif

  n = 0;
  XtSetArg (args [n], XmNautoUnmanage,   False); n++;
  XtSetArg (args [n], XmNdialogType,     type); n++;
  XtSetArg (args [n], XmNdialogStyle,    XmDIALOG_APPLICATION_MODAL); n++;
  XtSetArg (args [n], XmNdeleteResponse, XmDO_NOTHING); n++;

  /****** Messagetext und Titel nur nach Wunsch setzen, da sie dann ueber die */
  /****** Resourcedateien nicht mehr gestzt werden koennen ********************/
  if (! (mode & DONOT_SETTITLE))
    {
      XtSetArg (args [n], XmNdialogTitle,   titel); n++;
    }
  if (! (mode & DONOT_SETMESSAGE))
    {
      XtSetArg (args [n], XmNmessageString, str); n++;
    }

  msg->dialog =
    XmCreateMessageDialog (parent != NULL ? parent : topShell,
			   title,
			   args,
			   n);

  XmStringFree (str);
  XmStringFree (titel);

  if (msg->dialog == NULL)
    {
      MFree (msg, sizeof (message_t));
      return -1;
    }

#if 0
  DEBUG1 ("MessageBox: get buttons");
#endif

#if 1
  btnOk     = XmMessageBoxGetChild (msg->dialog, XmDIALOG_OK_BUTTON);
  btnCancel = XmMessageBoxGetChild (msg->dialog, XmDIALOG_CANCEL_BUTTON);
  btnHelp   = XmMessageBoxGetChild (msg->dialog, XmDIALOG_HELP_BUTTON);
#endif

#if 0
#ifdef DEBUG
  {
  XmString ok;
  char * txt;
  XtVaGetValues (btnOk, XmNlabelString, &ok, NULL);
  XmStringGetLtoR (ok, XmFONTLIST_DEFAULT_TAG, &txt);
  DEBUG2 ("okLabel (%s)\n", txt);
  XmStringFree (ok); XtFree (txt);

  XtVaGetValues (btnCancel, XmNlabelString, &ok, NULL);
  XmStringGetLtoR (ok, XmFONTLIST_DEFAULT_TAG, &txt);
  DEBUG2 ("CancelLabel (%s)\n", txt);
  XmStringFree (ok); XtFree (txt);

  XtVaGetValues (btnHelp, XmNlabelString, &ok, NULL);
  XmStringGetLtoR (ok, XmFONTLIST_DEFAULT_TAG, &txt);
  DEBUG2 ("HelpLabel (%s)\n", txt);
  XmStringFree (ok); XtFree (txt);
  }
#endif
#endif

  if (mode & CANCEL_BUTTON)		     /* Falls Cancel-Button vorhanden */
    {
      XtAddCallback (msg->dialog, XmNcancelCallback,
		     (XtCallbackProc) CancelPressed,
		     msg);
    }
  else					     /* ansonsten hidden	      */
    {
      if (btnCancel != NULL) XtUnmanageChild (btnCancel);
    }

  if (mode & OK_BUTTON)			     /* falls OK-Button vorhanden     */
    {
      XtAddCallback (msg->dialog, XmNokCallback,
		     (XtCallbackProc) OkPressed,
		     msg);
    }
  else					     /* ansonsten hidden	      */
    {
      if (btnOk != NULL) XtUnmanageChild (btnOk);
    }

  if (mode & HELP_BUTTON)
    {
      XtAddCallback (msg->dialog, XmNhelpCallback,
		     (XtCallbackProc) HelpPressed,
		     msg);
    }
  else
    {
      if (btnHelp != NULL) XtSetSensitive (btnHelp, False);
    }

  /****** Die Close Funktion der Shell abfangen *******************************/
  WM_DELETE_WINDOW =
    XmInternAtom (XtDisplay (msg->dialog), "WM_DELETE_WINDOW",
		  False);
  XmAddWMProtocolCallback (GetTopShell (msg->dialog),
			   WM_DELETE_WINDOW,
			   (XtCallbackProc) CancelPressed,
			   (XtPointer) msg);

#if 0
  DEBUG1 ("MessageBox: manage dialog");
#endif

  XtManageChild (msg->dialog);
  XtPopup (XtParent (msg->dialog), XtGrabNone);

  if (msg->Mode & DOPPLE_SIZE)
    {
      int screen_w, screen_h;
      int width, height;

      screen_w = WidthOfScreen  (XtScreen (msg->dialog));
      screen_h = HeightOfScreen (XtScreen (msg->dialog));

      XtVaGetValues (GetTopShell (msg->dialog),
		     XmNwidth,  &width,
		     XmNheight, &height,
		     NULL);

#if 0
      DEBUG5 ("Dopplte Groesse screen h=(%i), w=(%i)\n"
	      "dialog h=(%i), w=(%i)\n",
	      screen_h, screen_w,
	      height, width);
#endif

      XtVaSetValues (GetTopShell (msg->dialog),
#if 0
		     XmNx,      max (0, (screen_w - width   * 2) / 2),
		     XmNwidth,  min (screen_w, width  * 2),
#endif
		     XmNy,      max (0, (screen_h - height  * 2) / 2),
		     XmNheight, min (screen_h, height * 2),
		     NULL);
    }

#if 0
  DEBUG1 ("MessageBox: wait for closing message dialog");
#endif

  /****** Solange warten, bis das Fenster geschlossen wird, damit man den *****/
  /****** Rueckgabewert abfaengt **********************************************/
  WaitForCloseMessageBox (msg);
  ReturnValue = msg->Return;

  MFree (msg, sizeof (message_t));

#if 0
  DEBUG1 ("Message Ende");
#endif

  return ReturnValue;
}
