/************************************************************************
 *  (C)opyright 1996/1997 by Khamis Abu El-Komboz                       *
 *  xkhfm 1.0 can be copied and distributed freely for any              *
 *  non-commercial purposes. xkhfm can only be incorporated             *
 *  into commercial software with the permission of the current author. *
 *  Mail: khamis@multix.de                                              *
 ************************************************************************/
/*
 * $Author: khamis $
 * $Id: mwinDisp.c,v 1.4 1997/01/02 00:55:29 khamis Exp khamis $
 * $Source: /usr2/xkhfm/xkhlib/xkhsrc/RCS/mwinDisp.c,v $
 */
#define IS_EXTERN
#define EXTERN		extern

#include <stdio.h>
#include <stdlib.h>
#include <sys/param.h>
#include <Xm/DrawnB.h>
#include "Xinc.h"
#include <Xm/CutPaste.h>
#include <X11/keysym.h>

#include "defines.h"
#include "Umem.h"
#include "Uxutil.h"
#include "Umessage.h"
#include "Umenu.h"
#include "Ustrutil.h"
#include "Ufileutil.h"
#include "help.h"
#include "mainwin.h"

#define DefFileSpaceBetwColumns		15
#define DefFileSmallSpBetwColumns	7
#define DefFileDetailsSpBetwColumns	7
#define DefFileSpaceBetwIconAndName	5

#define WDG_IS_LOWER(mwin, dirWdg) (min (dirWdg->iconX, dirWdg->nameX) > \
	    mwin->dimFilesWidth - mwin->FilesViewX || \
	    min (dirWdg->iconY, dirWdg->nameY) > \
	    mwin->dimFilesHeight - mwin->FilesViewY)
#define WDG_IS_UPPER(mwin, dirWdg) (max (dirWdg->iconX + dirWdg->iconWidth, \
	         dirWdg->nameX + dirWdg->nameWidth) \
                 < - mwin->FilesViewX || \
            max (dirWdg->iconY + dirWdg->iconHeight, \
	         dirWdg->nameY + dirWdg->nameHeight) \
                 < - mwin->FilesViewY)

#define WDG_NOT_IN_VIEW_AREA(mwin, dirWdg) (min (dirWdg->iconX, dirWdg->nameX) > \
	    mwin->dimFilesWidth - mwin->FilesViewX || \
	    min (dirWdg->iconY, dirWdg->nameY) > \
	    mwin->dimFilesHeight - mwin->FilesViewY || \
            max (dirWdg->iconX + dirWdg->iconWidth, \
	         dirWdg->nameX + dirWdg->nameWidth) \
                 < - mwin->FilesViewX || \
            max (dirWdg->iconY + dirWdg->iconHeight, \
	         dirWdg->nameY + dirWdg->nameHeight) \
                 < - mwin->FilesViewY)

int mwinSelectOrCopy (Mainwin_t mwin, XEvent *event, int mode);
int mwinLoadColormap (Mainwin_t mwin);

/* static variables */
static int DefaultColormap_Loaded = False;
static Colormap lastColormap = 0;
static int lastSelected = 0;

int mwinRedisp (Mainwin_t mwin)
{
  if (mwin->fomFiles == NULL || mwin->FilesWindow == 0)
    return False;

  DEBUG1 ("mwinRedisp:\n");
  
  mwinLoadColormap (mwin);

  mwin->dimFilesWidth  = wdgWidth (mwin->fomFiles);
  mwin->dimFilesHeight = wdgHeight(mwin->fomFiles);
  mwin->FilesWindow    = XtWindow (mwin->fomFiles);

  /*
   * Clear old Entries
   */
  if (mwin->FilesClearIcons)
    {
      DEBUG1 ("mwinRedisp: Clear Area");
      XClearArea (display,
		  mwin->FilesWindow,
		  abs (mwin->FilesViewX), abs (mwin->FilesViewY),
		  mwin->dimFilesWidth, mwin->dimFilesHeight,
		  False);
      mwin->FilesClearIcons = False;
    }
  
  /*
   * Fill the DrawingArea with the Directory entries
   */
  mwinDisp (mwin);

  DEBUG1 ("mwinRedisp: End");

  return True;
}

int mwinRedraw (Mainwin_t mwin, int mode)
{
  if (mode)
    {
      mwin->FilesClearIcons = True;
      mwinCalc   (mwin);
    }
  mwinRedisp (mwin);
  return True;
}

/** calculate x-, y-coordinates, width and height of the items in the */
/** file list */
int mwinCalc (Mainwin_t mwin)
{
  DirWidgets_t dirWdg;
  Dchain_t dirList;
  int i, j, ll, width;
  int iconsProLine;
  int wX, wY;
  char * p;

  if (mwin->FilesWindow == 0)
    return False;

  mwin->FilesMaxXPaint = 0;
  mwin->FilesMaxYPaint = 0;

  /** different view types have different calculating methodes **/
  switch (mwin->ViewMode)
    {
    case VIEW_LRG_ICONS:
      mwin->PixmWidth     = DefaultPixmWidth;
      mwin->PixmHeight    = DefaultPixmHeight;
      mwin->PixmBoxWidth  = mwin->PixmWidth * 2;
      mwin->PixmBoxHeight = (mwin->PixmHeight
			     + mwin->NameHeight
			     + mwin->SpaceBetweenIcons);
      for (i=0;
	   i<mwin->dirEntryCount;
	   i++)
	{
	  dirWdg  = mwin->dirWidgets[i];
	  if (dirWdg->iconPixm > 0)
	  {
	    unsigned int bw;
	    int x, y;
	    Window root;
	    if (!XGetGeometry (display, dirWdg->iconPixm,
			       &root,
			       &x,
			       &y,
			       &dirWdg->iconWidth,
			       &dirWdg->iconHeight,
			       &bw,
			       &dirWdg->iconDepth))
	    {
	      WARNING2 ("mwinCalc: can't get pixmap geometry of [%s]\n",
	                dirPath (dirWdg->dirListEntry));
	      XFreePixmap (display, dirWdg->iconPixm);
	      dirWdg->iconPixm  = 0;
	      dirWdg->iconDepth = 0;
	    }
	  }
	  else
	  {
	    mwin->dirWidgets[i]->iconWidth = 0;
	    mwin->dirWidgets[i]->iconHeight = 0;
	  }
	}
      iconsProLine = max (1, mwin->dimFilesWidth / mwin->PixmBoxWidth);
      break;

    case VIEW_SML_ICONS:
      mwin->PixmWidth     = 16;
      mwin->PixmHeight    = 16;
      mwin->PixmBoxWidth  = 100;
      mwin->PixmBoxHeight = (max (mwin->PixmHeight, mwin->NameHeight)
			     + DefFileSmallSpBetwColumns);

      /* calculate max. width of an icon */
      for (i=0;
	   i<mwin->dirEntryCount;
	   i++)
	{
	  dirWdg  = mwin->dirWidgets[i];
	  dirList = dirWdg->dirListEntry;
	  dirWdg->nameWidth  = XTextWidth (mwin->fntMainwin,
					   dirName (dirList),
					   strlen (dirName (dirList)));
	  mwin->PixmBoxWidth = max (mwin->PixmBoxWidth,
				    mwin->PixmWidth
				    + dirWdg->nameWidth
				    + mwin->SpaceBetweenIcons);
	}

      iconsProLine = max (1, mwin->dimFilesHeight / mwin->PixmBoxHeight);
      break;

    case VIEW_DETAILS:
      mwin->PixmWidth     = 16;
      mwin->PixmHeight    = 16;
      mwin->PixmBoxWidth  = 80;
      mwin->PixmBoxHeight = (max (mwin->PixmHeight, mwin->NameHeight)
			     + DefFileDetailsSpBetwColumns);
      memset (mwin->NameWidths, 0, sizeof (mwin->NameWidths));
      mwin->NameWidth     = 0;
      break;
    }

  DEBUG2 ("mwinCalc: iconsProLine=(%il)\n", iconsProLine);

  /** calculating x-, y-coordinates and width and height */
  wX = mwin->PixmXBegin, wY = mwin->PixmYBegin;
  for (i=j=0;
       i<mwin->dirEntryCount;
       i++,j++)
    {
      
      dirWdg = mwin->dirWidgets[i];
      dirList = dirWdg->dirListEntry;

      /** different view modes */
      switch (mwin->ViewMode)
	{
	case VIEW_LRG_ICONS:

	  /* icon coordinates */
	  if (dirWdg->iconPixm <= 0 || dirWdg->iconWidth == 0)
	  {
	    dirWdg->iconWidth  = mwin->PixmWidth;
	  }
	  if (dirWdg->iconPixm <= 0 || dirWdg->iconHeight == 0)
	  {
	    dirWdg->iconHeight = mwin->PixmHeight;
	  }

	  /* name coordinates */
	  dirWdg->nameWidth  = XTextWidth (mwin->fntMainwin,
					   dirName (dirList),
					   strlen (dirName
						   (dirList)));
	  dirWdg->nameHeight = (mwin->fntMainwin->ascent
				+ mwin->fntMainwin->descent);
	  
	  width = max (mwin->PixmBoxWidth, max (dirWdg->nameWidth, dirWdg->iconWidth));
	  
          if (wX + width > mwin->dimFilesWidth)
          {
            wY = mwin->FilesMaxYPaint + mwin->SpaceBetweenIcons;
	    wX = mwin->PixmXBegin;
          }
	  
	  dirWdg->iconX = (int)(wX + (width - (int)dirWdg->iconWidth) / 2);
	  dirWdg->iconY = wY;
	     
	  dirWdg->nameX = (int)(wX + (width - (int) dirWdg->nameWidth) / 2);
	  dirWdg->nameY = (wY
	                   + (int) dirWdg->iconHeight
			   + DefFileSpaceBetwIconAndName);
          
          wX = max (dirWdg->nameX + max (mwin->PixmBoxWidth, dirWdg->nameWidth),
	            dirWdg->iconX + max (mwin->PixmBoxWidth, dirWdg->iconWidth))
	     + mwin->SpaceBetweenIcons;
	  break;

	case VIEW_SML_ICONS:
	  
	  /* icon coordinates */
	  dirWdg->iconWidth  = mwin->PixmWidth;
	  dirWdg->iconHeight = mwin->PixmHeight;
	  
	  /* name coordinates */
	  dirWdg->nameHeight = (  mwin->fntMainwin->ascent
				+ mwin->fntMainwin->descent);

	  wX = mwin->PixmXBegin + mwin->PixmBoxWidth  * (i / iconsProLine);
	  wY = mwin->PixmYBegin + mwin->PixmBoxHeight * (i % iconsProLine);
	  
	  dirWdg->iconX = wX;
	  dirWdg->iconY = (wY + ((max (mwin->NameHeight, dirWdg->iconHeight)
				  - dirWdg->iconHeight)
				 / 2));
	  dirWdg->nameX = (wX
			   + dirWdg->iconWidth
			   + DefFileSpaceBetwIconAndName);
	  dirWdg->nameY = (wY + ((max (dirWdg->nameHeight, dirWdg->iconHeight)
				  - dirWdg->nameHeight)
				 / 2));
	  break;

	case VIEW_DETAILS:
	  {
	    char * str;
	    int ii, jj, width, pos;
	    /* icon coordinates */
	    dirWdg->iconWidth  = mwin->PixmWidth;
	    dirWdg->iconHeight = mwin->PixmHeight;
	    
	    str = dirMakeChoose (dirWdg->dirListEntry,
				 DIR_CHOOSE |
				 DIR_PRINT_NAME |
				 DIR_DIFFER |
				 DIR_LINKED_TO);
	    for (width=pos=ii=0, p = str; True; ii++)
	      {
		pos = StrChrIndex (p, '\n');
		if (pos < 0) break;
		mwin->NameWidths[ii] =
		  max (mwin->NameWidths[ii],
		       XTextWidth(mwin->fntMainwin, p, pos)
		       + DefFileSpaceBetwColumns);
		width += mwin->NameWidths[ii];
		p += pos + 1;
	      }
	    width -= DefFileSpaceBetwColumns;
	    mwin->NameWidth    = max (mwin->NameWidth, width);
	    dirWdg->nameWidth  = width;
	    dirWdg->nameHeight = (mwin->fntMainwin->ascent
				  + mwin->fntMainwin->descent);
	    
	    wX = mwin->PixmXBegin;
	    wY = mwin->PixmYBegin + mwin->PixmBoxHeight * j;
	    
	    dirWdg->iconX = wX;
	    dirWdg->iconY = wY + ((max (mwin->NameHeight, dirWdg->iconHeight)
				   - dirWdg->iconHeight)
				  / 2);
	    dirWdg->nameX = wX + dirWdg->iconWidth + 5;
	    dirWdg->nameY = wY + ((max (dirWdg->nameHeight, dirWdg->iconHeight)
				   - dirWdg->nameHeight)
				/ 2);
	    break;
	  } /* case VIEW_DETAILS */
	} /* switch */
      
      mwin->FilesMaxXPaint = max (mwin->FilesMaxXPaint,
				  max (dirWdg->nameX + dirWdg->nameWidth,
				       dirWdg->iconX + dirWdg->iconWidth)
				  + 1);
      mwin->FilesMaxYPaint = max (mwin->FilesMaxYPaint,
				  max (dirWdg->nameY + dirWdg->nameHeight,
				       dirWdg->iconY + dirWdg->iconHeight)
				  + 1);
    } /* for */
    
  /** set the largest width for all icons in the details mode */
  if (mwin->ViewMode == VIEW_DETAILS)
    for (i=0;
	 i<mwin->dirEntryCount;
	 i++)
      {
	mwin->dirWidgets[i]->nameWidth = mwin->NameWidth;
      }
 
  DEBUG2 ("mwinCalc: mwin->FilesMaxXPaint(%i)\n", mwin->FilesMaxXPaint); 
 
  /** resize the window of drawing area for the new calculated dimensions **/
  XResizeWindow (display,
		 XtWindow (mwin->fomFiles),
		 max (mwin->FilesMaxXPaint, mwin->dimFilesWidth),
		 max (mwin->FilesMaxYPaint, mwin->dimFilesHeight));

  /* scollbar sizes */
  mwinSetFilesScrollBars (mwin);

  /* move viewed part to the seen site of window */
  XMoveWindow (display, XtWindow (mwin->fomFiles),
	       mwin->FilesViewX,
	       mwin->FilesViewY);

  return True;
}

int mwinDisp (Mainwin_t mwin)
{
  DirWidgets_t dirWdg;
  int i, j, Recalc = False, x = -1, y = -1;

  if (mwin->FilesWindow == 0)
    return False;

  DEBUG1 ("mwinDisp");

  for (i=0;
       i<mwin->dirEntryCount;
       i++)
    {
      dirWdg = mwin->dirWidgets[i];
      if (WDG_IS_LOWER (mwin, dirWdg))
      {
	break;
      }
      
      /**************************************/
      /* view icon files with there pixmaps */
      for (j=i;
           j<mwin->dirEntryCount && 
           mwin->ViewMode == VIEW_LRG_ICONS &&
	   ! mwin->dirWidgets[j]->iconTested;
	   j++)
      {
        dirWdg = mwin->dirWidgets[j];
        if (WDG_NOT_IN_VIEW_AREA (mwin, dirWdg))
        {
	  break;
        }
	
        mwinLoadIcon (mwin, j);
	if (mwin->dirWidgets[j]->iconPixm > 0)
	  Recalc = True;
	  
	if (x == -1 && y == -1)
	{
	  x = min (dirWdg->iconX, dirWdg->nameX);
	  y = min (dirWdg->iconY, dirWdg->nameY);
	}
      }
      
      if (Recalc)
      {
	mwinCalc (mwin);
	Recalc = False;
	XClearArea (XtDisplay (mwin->shlMainwin),
	            XtWindow (mwin->fomFiles),
		    x, y,
		    mwin->dimFilesWidth,
		    mwin->dimFilesHeight,
		    False);
        x = y = -1;
      }
      /**************************************/

      mwinDispIcon (mwin, i, False);
    }

  DEBUG1 ("mwinDisp: End");

  return True;
}

int mwinLoadIcon (Mainwin_t mwin, int num)
{
  Dchain_t iconDch;
  DirWidgets_t dirWdg = mwin->dirWidgets[num];
  
  /* don't try again */
  dirWdg->iconTested = True;

  if (dirType (dirWdg->dirListEntry) != TYPE_REG)
    {
      return False;
    }
  
  for (iconDch = UdchNext (iconExtensionList);
       iconDch != NULL;
       iconDch = UdchNext (iconDch))
  {
    if (IsRegular (assListPattern (iconDch), dirName (dirWdg->dirListEntry)))
      break;
  }
  
  if (iconDch == NULL)
    return False;
  
  dirWdg->iconPixm =
    LoadPixmap (mwin->fomFiles,
                dirPath (dirWdg->dirListEntry),
		&dirWdg->iconColormap);
#if 1
  if (dirWdg->iconPixm > 0)
  {
    unsigned int bw;
    int x, y;
    Window root;
    if (!XGetGeometry (display, dirWdg->iconPixm,
		       &root,
		       &x,
		       &y,
		       &dirWdg->iconPixmWidth,
		       &dirWdg->iconPixmHeight,
		       &bw,
		       &dirWdg->iconPixmDepth))
    {
      DEBUG2 ("mwinLoadIcon: can't get pixmap geometry of [%s], ignored !\n",
                dirPath (dirWdg->dirListEntry));
      XFreePixmap (display, dirWdg->iconPixm);
      dirWdg->iconPixm  = 0;
      dirWdg->iconPixmDepth = 0;
      dirWdg->iconPixmWidth = 0;
      dirWdg->iconPixmHeight = 0;
      dirWdg->iconColormap = 0;
    }
  }
#endif

  return dirWdg->iconPixm > 0;
}

int mwinDispIcon (Mainwin_t mwin, int num, int mode)
{
  Pixmap pxm = 0;
  char * str = "";
  DirWidgets_t dirWdg;
  int sel;
  int type;
  int px, py, pw, ph, tx, ty, tw, th;

#ifdef DEBUG
  if (XtWindow (mwin->fomFiles) != mwin->FilesWindow)
    {
      ERROR3 ("mwinDispIcon: XtWindow(mwin->fomFiles)[%li] != "
	      "mwin->FilesWindow[%li]\n",
	      XtWindow (mwin->fomFiles),
	      mwin->FilesWindow);
      exit (1);
    }
#endif

  DEBUG2 ("clock-1=[%li]\n", clock());

  dirWdg = mwin->dirWidgets[num];
  sel    = (dirSelected (dirWdg->dirListEntry) ||
	    dirSoftSelected (dirWdg->dirListEntry) ||
	    dirCopySelected (dirWdg->dirListEntry));

  /* not viewed */
  if (WDG_NOT_IN_VIEW_AREA (mwin, dirWdg))
  {
    return True;
  }
  
  DEBUG2 ("clock-2=[%li]\n", clock());

  switch (mwin->ViewMode)
    {
    case VIEW_LRG_ICONS:
      type = dirExtendedType (mwin->dirWidgets[num]->dirListEntry);
      if (mwin->pixmList [type] == (Pixmap) 0)
	type = PIXM_file32;
      str = dirName (mwin->dirWidgets[num]->dirListEntry);
      break;

    case VIEW_DETAILS:
    case VIEW_SML_ICONS:
      type = dirType (mwin->dirWidgets[num]->dirListEntry);
      switch (type)
	{
	case TYPE_LNK:  type = PIXM_file_link16; break;	  
	case TYPE_DIR:  type = PIXM_fldr16;    break;	  
	case TYPE_REG:  type = PIXM_file16; break;	  
	case TYPE_EXE:  type = PIXM_file_exec16; break;
	case TYPE_BLK:  type = PIXM_file_devblk16; break;	  
	case TYPE_CHR:  type = PIXM_file_devchr16; break;	  
	case TYPE_FIFO: type = PIXM_file_pipe16; break;	  
	case TYPE_NAME: type = PIXM_file_unknown16; break;	  
	default:        type = PIXM_file_unknown16; break;
	}
      if (mwin->pixmList [type] == (Pixmap) 0)
	type = PIXM_file16;

      if (mwin->ViewMode == VIEW_SML_ICONS)
	str = dirName (mwin->dirWidgets[num]->dirListEntry);
      else
	{
	  str = dirMakeChoose (mwin->dirWidgets[num]->dirListEntry,
			       DIR_CHOOSE | DIR_PRINT_NAME | DIR_DIFFER |
			       DIR_LINKED_TO);
	}
      break;
    }

  DEBUG2 ("clock-3=[%li]\n", clock());

#if 1
  /* delete selection box */
  SET_FG (display, mwin->gc, mwin->filesBg);
#if 0
  XDrawRectangle (display,
		  mwin->FilesWindow,
		  mwin->gc,
		  dirWdg->iconX-1,
		  dirWdg->iconY-1,
		  dirWdg->iconWidth+1,
		  dirWdg->iconHeight+1);
#endif
  XDrawRectangle (display,
		  mwin->FilesWindow,
		  mwin->gc,
		  dirWdg->nameX-1,
		  dirWdg->nameY-1,
		  dirWdg->nameWidth+1,
		  dirWdg->nameHeight+1);
#endif

  DEBUG2 ("clock-4=[%li]\n", clock());

  /* if having focus, draw first */
  if ( /* True || /**/ sel && mwin->FilesFocus)
    {
      SET_FG (display, mwin->gc, mwin->filesHighLight);
#if 0
      XDrawRectangle (display,
		      mwin->FilesWindow,
		      mwin->gc,
		      dirWdg->iconX-1,
		      dirWdg->iconY-1,
		      dirWdg->iconWidth+1,
		      dirWdg->iconHeight+1);
#endif
      XDrawRectangle (display,
		      mwin->FilesWindow,
		      mwin->gc,
		      dirWdg->nameX-1,
		      dirWdg->nameY-1,
		      dirWdg->nameWidth+1,
		      dirWdg->nameHeight+1);
    }


  SET_FONT (display, mwin->gc, mwin->fntMainwin->fid);

  SET_FG (display, mwin->gc, sel ? mwin->filesBg : mwin->filesFg);
  SET_BG (display, mwin->gc, sel ? mwin->filesFg : mwin->filesBg);

  DEBUG2 ("clock-5=[%li]\n", clock());

  if (mwin->ViewMode == VIEW_LRG_ICONS &&
      mwin->dirWidgets[num]->iconPixm > 0 &&
      mwin->dirWidgets[num]->iconDepth > 0)
  {
    if (mwin->dirWidgets[num]->iconDepth == 1)
      XCopyPlane (display,
		mwin->dirWidgets[num]->iconPixm,
	        mwin->FilesWindow,
	        mwin->gc,
	        0, 0,
	        mwin->dirWidgets[num]->iconWidth,
	        mwin->dirWidgets[num]->iconHeight,
	        mwin->dirWidgets[num]->iconX,
	        mwin->dirWidgets[num]->iconY,
	        mwin->dirWidgets[num]->iconDepth);
    else
      XCopyArea (display,
		mwin->dirWidgets[num]->iconPixm,
		mwin->FilesWindow,
		mwin->gc,
		0, 0,
		mwin->dirWidgets[num]->iconWidth,
		mwin->dirWidgets[num]->iconHeight,
		mwin->dirWidgets[num]->iconX,
		mwin->dirWidgets[num]->iconY);
  }
  else
  {
#ifdef HAVE_XPM
    XCopyArea (display,
	     mwin->pixmList [type],
	     mwin->FilesWindow,
	     mwin->gc,
	     0, 0,
	     mwin->dirWidgets[num]->iconWidth,
	     mwin->dirWidgets[num]->iconHeight,
	     mwin->dirWidgets[num]->iconX,
	     mwin->dirWidgets[num]->iconY);
#else
    XCopyPlane (display,
	      mwin->pixmList [type],
	      mwin->FilesWindow,
	      mwin->gc,
	      0, 0,
	      mwin->dirWidgets[num]->iconWidth,
	      mwin->dirWidgets[num]->iconHeight,
	      mwin->dirWidgets[num]->iconX,
	      mwin->dirWidgets[num]->iconY,
	      1L); /* Depth = 1 Bit */
#endif
  }

  DEBUG2 ("clock-6=[%li]\n", clock());

  if (mwin->ViewMode != VIEW_DETAILS)
    XDrawImageString
      (display,
       mwin->FilesWindow,
       mwin->gc,
       mwin->dirWidgets[num]->nameX,
       mwin->dirWidgets[num]->nameY + mwin->fntMainwin->ascent,
       str,
       strlen (str));
  else
    {
      char * p;
      int i, pos, j;
      mwin->dirWidgets[num]->nameWidth = mwin->NameWidth;
      pos=i=j=0; p = str;
      while (True)
	{
	  pos = StrChrIndex (p, '\n');
	  if (pos < 0) break;

	  XDrawImageString
	    (display,
	     mwin->FilesWindow,
	     mwin->gc,
	     mwin->dirWidgets[num]->nameX + j,
	     mwin->dirWidgets[num]->nameY + mwin->fntMainwin->ascent,
	     p,
	     pos);
	  p += pos + 1;
	  j += mwin->NameWidths [i];
	  i ++;
	}
    }

  DEBUG2 ("clock-7=[%li]\n", clock());

  return True;
}

/** called by motion pressed button or after release the button */
int mwinSelect (Mainwin_t mwin, XEvent *event)
{
  DEBUG1 ("mwinSelect:");

  mwinSelectOrCopy (mwin, event, False);

  DEBUG1 ("mwinSelect: After mwinSelectOrCopy");

  /* Cursor do not points to any icon */
  mwinSoftSelectIcon (mwin, mwin->FilesSoftSelNum, False);
  mwinCopySelectIcon (mwin, mwin->FilesCopySelNum, False);

  DEBUG1 ("mwinSelect: end");
}

/** called by motion pressed button or after release the button */
int mwinRelease (Mainwin_t mwin, XEvent *event)
{
  DEBUG1 ("mwinSelect:");

  mwinSelectOrCopy (mwin, event, SELECT_BTN_RELEASE);

  DEBUG1 ("mwinSelect: After mwinSelectOrCopy");

  /* Cursor do not points to any icon */
  mwinSoftSelectIcon (mwin, mwin->FilesSoftSelNum, False);
  mwinCopySelectIcon (mwin, mwin->FilesCopySelNum, False);

  DEBUG1 ("mwinSelect: end");
}

/** called, when button pressed to verify if the next action a select or */
/** copy/move action */
int mwinPress (Mainwin_t mwin, XEvent *event)
{
  return mwinSelectOrCopy (mwin, event, SELECT_BTN_PRESSED);
}

int mwinPointsToWidget (Mainwin_t mwin, int i,
			int xx1, int yy1, int xx2, int yy2)
{
#if 1
  int
    x1 = min (xx1, xx2),
    y1 = min (yy1, yy2),
    x2 = max (xx1, xx2),
    y2 = max (yy1, yy2);
#endif

  int
    wx1 = mwin->dirWidgets[i]->iconX,
    wy1 = mwin->dirWidgets[i]->iconY,
    wx2 = mwin->dirWidgets[i]->iconX + mwin->dirWidgets[i]->iconWidth,
    wy2 = mwin->dirWidgets[i]->iconY + mwin->dirWidgets[i]->iconHeight;
  int
    nx1 = mwin->dirWidgets[i]->nameX,
    ny1 = mwin->dirWidgets[i]->nameY,
    nx2 = mwin->dirWidgets[i]->nameX + mwin->dirWidgets[i]->nameWidth,
    ny2 = mwin->dirWidgets[i]->nameY + mwin->dirWidgets[i]->nameHeight;
  
  return (
      /* widget inner the selection box */
      wx1 >= x1 && wy1 >= y1 && wx2 <= x2 && wy2 <= y2 ||
      /* half button of icon in the selection box */
      wx1 <= x1 && wx2 >= x1 && wy1 <= y1 && wy2 >= y1 ||
      wx1 <= x2 && wx2 >= x2 && wy1 <= y1 && wy2 >= y1 ||
      wx1 <= x1 && wx2 >= x1 && wy1 <= y2 && wy2 >= y2 ||
      wx1 <= x2 && wx2 >= x2 && wy1 <= y2 && wy2 >= y2 ||
	  /* inside */
      wx1 >= x1 && wx2 <= x2 && wy1 <= y1 && wy2 >= y1 ||
      wx1 <= x2 && wx2 >= x2 && wy1 >= y1 && wy2 <= y2 ||
	  wx1 >= x1 && wx2 <= x2 && wy1 <= y2 && wy2 >= y2 ||
      wx1 <= x1 && wx2 >= x1 && wy1 >= y1 && wy2 <= y2 ||
      
      /* names, no selection box */
      (
       /*
	 (x2 - x1) <= CLICK_REGION &&
	 (y2 - y1) <= CLICK_REGION &&
       */
       (
	/* widget inner the selection box */
	nx1 >= x1 && ny1 >= y1 && nx2 <= x2 && ny2 <= y2 ||
	/* half button of icon in the selection box */
	nx1 <= x1 && nx2 >= x1 && ny1 <= y1 && ny2 >= y1 ||
	nx1 <= x2 && nx2 >= x2 && ny1 <= y1 && ny2 >= y1 ||
	nx1 <= x1 && nx2 >= x1 && ny1 <= y2 && ny2 >= y2 ||
	nx1 <= x2 && nx2 >= x2 && ny1 <= y2 && ny2 >= y2 ||
	/* inside */
	nx1 >= x1 && nx2 <= x2 && ny1 <= y1 && ny2 >= y1 ||
	nx1 <= x2 && nx2 >= x2 && ny1 >= y1 && ny2 <= y2 ||
	nx1 >= x1 && nx2 <= x2 && ny1 <= y2 && ny2 >= y2 ||
	nx1 <= x1 && nx2 >= x1 && ny1 >= y1 && ny2 <= y2
	))
      );
}
int mwinSelectOrCopy (Mainwin_t mwin, XEvent *event, int mode)
{
  int
    x1 = min (mwin->btnPressX, mwin->btnReleaseX),
    y1 = min (mwin->btnPressY, mwin->btnReleaseY),
    x2 = max (mwin->btnPressX, mwin->btnReleaseX),
    y2 = max (mwin->btnPressY, mwin->btnReleaseY);
  int i;

  /** default, select motion */
  if (mode & SELECT_BTN_PRESSED)
    mwin->MotionType = MOTION_SELECT;

  if (! (mode & SELECT_BTN_PRESSED))
    {
      int value, slidersize, incrreturn, pageincrret, maxvalue;

      /** move viewed part to button position */
      if (mwin->btnReleaseY < - mwin->FilesViewY)
	{
	  XmScrollBarGetValues (mwin->scrVertFiles,
				&value,
				&slidersize,
				&incrreturn,
				&pageincrret);
	  XmScrollBarSetValues (mwin->scrVertFiles,
				max (0, value - 1),
				slidersize,
				incrreturn,
				pageincrret,
				True);
	}
      else if (mwin->btnReleaseY > mwin->dimFilesHeight - mwin->FilesViewY)
	{
	  XtVaGetValues (mwin->scrVertFiles,
			 XmNmaximum,       &maxvalue,
			 NULL);
	  XmScrollBarGetValues (mwin->scrVertFiles,
				&value,
				&slidersize,
				&incrreturn,
				&pageincrret);
	  XmScrollBarSetValues (mwin->scrVertFiles,
				min (maxvalue-1, value + 1),
				slidersize,
				incrreturn,
				pageincrret,
				True);
	}

      /* New Selection */
      mwin->SelectedIcons = 0;
    }

  for (i=0; i<mwin->dirEntryCount; i++)
    {
      if (mwinPointsToWidget (mwin, i, x1, y1, x2, y2))
	{

	  /** if called, when button pressed, mark for the copy/move action **/
	  if (mode & SELECT_BTN_PRESSED)
	    {

	      mwinCopySelectIcon (mwin, i, True);

	      mwin->MotionType = MOTION_COPY;
	      return MOTION_COPY;
	    }
	  else
	    {
	      /* by selection only one file toggle selection */
	      if ((event->xbutton.state & ShiftMask) &&
		  (x2 - x1) <= CLICK_REGION &&
		  (y2 - y1) <= CLICK_REGION)
		{
		  if (mwinSelected (mwin, i))
		    {
		      mwinSelectIcon (mwin, i, False);
		      mwin->SelectedIcons --;
		    }
		  else
		    {
		      mwinSelectIcon (mwin, i, True);
		      mwin->SelectedIcons ++;
		      lastSelected = i;
		    }
		}
	      else
		{
		  mwin->SelectedIcons ++;
		  mwinSelectIcon (mwin, i, True);
		  lastSelected = i;
		}
	    }
	}
      else if (! (mode & SELECT_BTN_PRESSED))
	{
	  /* when shift not pressed */
	  if (! (event->xbutton.state & ShiftMask))
	    mwinSelectIcon (mwin, i, False);
	}
    }

#if 1
  if (mode & SELECT_BTN_RELEASE)
    {
      mwinRedisp (mwin);
    }
  else if (! (mode & SELECT_BTN_PRESSED))
    mwinDispStatus (mwin);
#endif

  /* load colormap if needed */
  if (mwin->SelectedIcons == 1 || ! DefaultColormap_Loaded)
    mwinLoadColormap (mwin);
  
  /** redisplay new selected or deselected icons */
  if (mode & SELECT_BTN_PRESSED)
    return MOTION_SELECT;

  return True;
}

int mwinExec (Mainwin_t mwin, XEvent *event)
{
  int i, ret = False;

#if 0
  DEBUG2 ("mwinExec: mwin->SelectedIcons=(%i)\n", mwin->SelectedIcons);
#endif

  if (mwin->SelectedIcons <= 0)
    return False;

  BaisyCursor (mwin->shlMainwin);

  for (i=0; i<mwin->dirEntryCount; i++)
    {
      Dchain_t dir = mwin->dirWidgets[i]->dirListEntry;
      int sel = dirSelected (dir);

      /* verify all selected files and execute or edit it */
      if (sel)
	{
	  int type = dirType (dir);

	  if (type == TYPE_LNK)
	    {
	      type = dirTypeOfRealPath (dir);
	    }
	  
	  switch (type)
	    {
	    case TYPE_DIR:
	      mwinChangePathName
		(mwin,
		 UPathAppend (mwin->DirName,
			      dirName (mwin->dirWidgets[i]->dirListEntry)));
	      break;

	    case TYPE_EXE:
	      cmdExecCmd (mwin->DirName,
			  dirPath (mwin->dirWidgets[i]->dirListEntry),
			  EXEC_IN_XTERM);
	      break;

	      /* edit regular file */
	    case TYPE_REG:
	      cmdExecCmd (mwin->DirName,
			   dirPath (mwin->dirWidgets[i]->dirListEntry),
			   EXEC_ASSOCIATED);
	      break;

	    default:
	      break;
	    }
	  ret = True;
	}
    }

  NormalCursor (mwin->shlMainwin);

  return ret;
}

int mwinSelectIcon (Mainwin_t mwin, int num, int mode)
{
  if (num >= mwin->dirEntryCount)
    return False;

  /* select or unselect an icon */
  if (mwinSelected (mwin, num) != mode)
    {
      dirSelect (mwin->dirWidgets[num]->dirListEntry, mode);
      mwinDispIcon (mwin, num, False);
    }
  
  return True;
}

int mwinSoftSelectIcon (Mainwin_t mwin, int num, int mode)
{
  int i;

  if (num >= mwin->dirEntryCount)
    return False;

  /* soft select or unselect an icon, maximal on Item can be selected */
  if (dirSoftSelected (mwin->dirWidgets[num]->dirListEntry) != mode)
    {
      dirSoftSelect (mwin->dirWidgets[num]->dirListEntry, mode);

      /* unselect other selected items */
      if (mode && mwin->FilesSoftSelNum != num)
	{
	  if (mwin->FilesSoftSelNum < mwin->dirEntryCount)
	    {
	      dirSoftSelect (mwin->dirWidgets
			     [mwin->FilesSoftSelNum]->dirListEntry,
			     False);
	      mwinDispIcon  (mwin, mwin->FilesSoftSelNum, False);
	    }
	  mwin->FilesSoftSelNum = num;
	}

      /* Display the soft selected item */
      mwinDispIcon  (mwin, num, False);
    }
  
  return True;
}

int mwinCopySelectIcon (Mainwin_t mwin, int num, int mode)
{
  int i;

  if (num >= mwin->dirEntryCount)
    return False;

  /* soft select or unselect an icon, maximal on Item can be selected */
  if (dirCopySelected (mwin->dirWidgets[num]->dirListEntry) != mode)
    {
      dirCopySelect (mwin->dirWidgets[num]->dirListEntry, mode);

      /* unselect other selected items */
      if (mode && mwin->FilesCopySelNum != num)
	{
	  if (mwin->FilesCopySelNum < mwin->dirEntryCount)
	    {
	      dirCopySelect (mwin->dirWidgets
			     [mwin->FilesCopySelNum]->dirListEntry,
			     False);
	      mwinDispIcon  (mwin, mwin->FilesCopySelNum, False);
	    }
	  mwin->FilesCopySelNum = num;
	}

      /* Display the soft selected item */
      mwinDispIcon  (mwin, num, False);
    }
  
  return True;
}

int mwinDispStatus (Mainwin_t mwin)
{
  char StatStr [512], * p = StatStr;
  
#if 1
  p = dirMakeStatStr (mwin->dirEntry, DIR_STATUS|DIR_LINKED_TO);
#else
  p[0] = 0;
  sprintf (p, "%i Files, ", mwin->dirEntryCount);
  
  /* if more than one file selected */
  if (mwin->SelectedIcons > 1)
    {
      sprintf (p+strlen(p), "%i selected, ", mwin->SelectedIcons);
    }
  else if (mwin->SelectedIcons == 1)
    {
      int i;
      for (i=0; i<mwin->dirEntryCount; i++)
	{
	  if dirSelected (mwin->dirWidgets[i]->dirListEntry)
	    {
	      /* file stat */
	      strcat (p,
		      dirMakeStatStr (mwin->dirWidgets[i]->dirListEntry,
				      False));
	      break;
	    }
	}
    }
#endif
  
  /* display the status line on the window */
  winMesgArea (mwin->winMainwin, p);

  return True;
}

int dirGetType (char * path);

int mwinCanCopy_x (Mainwin_t mwin, int x, int y, int mode)
{
  static char * data = NULL;
  static int len = 0;
  char * p;
  int i, j, cnt;

  DEBUG1 ("mwinCanCopy");

  if (data != NULL)
    {
      MFree (data, len);
      data = NULL;
    }

  if (mode && mwin->Copy == NULL)
    mwin->Copy = cpyMake ();

  /* get file list */
  len = GetDataFromClipboard (&data);

  if (len < 1)
    return False;

  /* count the files in the clipboard */
  for (cnt=i=0, p = data; p[0] != 0; )
    {
      i = StrChrIndex (p, ' ');
      if (p[0] != ' ') cnt ++;
      if (i<0) break;
      p += i + 1;
    }
  if (cnt == 1 && data[len-1] == ' ')
    data [len-1] = 0;

  if (cnt < 1)
    return False;

  DEBUG3 ("mwinCanCopy: len(%i), cnt(%i)\n", len, cnt);

  /* if x or y point outside the window */
  if (x < 0 || y < 0)
    return False;

  for (i=0; i<mwin->dirEntryCount; i++)
    {
      if (mwinPointsToWidget (mwin, i, x, y, x, y))
	{
	  mwinSoftSelectIcon (mwin, i, True);

	  /* 1. Fall: the File is already selected */
	  if (mwinSelected (mwin, i))
	    return False;

	  /* 2. Fall: Copy many files to none Directory */
	  if (cnt > 1 && /*mwin->SelectedIcons > 1 &&*/
	      dirTypeOfRealPath (mwin->dirWidgets[i]->dirListEntry) != TYPE_DIR)
	    return False;

	  /* 3. Fall: Copy files or directories to a directory */
	  if (dirTypeOfRealPath (mwin->dirWidgets[i]->dirListEntry) == TYPE_DIR)
	    {
	      if (mode)
		{
		  sprintf (mwin->Copy->Dest, "%s/%s",
			   /* Do not forget the root directory */
			   strcmp (mwin->DirName, "/")
			   ? mwin->DirName
			   : "",
			   dirName (mwin->dirWidgets[i]->dirListEntry));
		  /* if there is only one file to copy, it well copy */
		  /* the file/directory to destination with the same */
		  /* name */
		  if (cnt == 1)
		    {
		      strcpy (mwin->Copy->Dest+strlen(mwin->Copy->Dest),
			      "/");
		      strcpy (mwin->Copy->Dest+strlen(mwin->Copy->Dest),
			      basename (data));
		    }
		}
	      return True;
	    }

	  /* 4. Fall: Copy one file to another file (overwrite) */
	  if (cnt == 1 /*mwin->SelectedIcons == 1*/ )
	    {
	      int type1 = dirType (mwin->dirWidgets[i]->dirListEntry);
	      int type2 = dirGetType (data);

	      if (type1 == type2 ||
		  type1 == TYPE_EXE && type2 == TYPE_REG ||
		  type1 == TYPE_DIR && type2 == TYPE_EXE)
		{

		  /* Copy the Destination String to the Copy Window */
		  if (mode)
		    {
		      sprintf (mwin->Copy->Dest, "%s/%s",
			       /* Do not forget the root directory */
			       strcmp (mwin->DirName, "/")
			       ? mwin->DirName
			       : "",
			       dirName (mwin->dirWidgets[i]->dirListEntry));
		    }
		  return True;
		}
	      return False;
	    }
	     
	  return False;
	}
    }

  /* Cursor do not points to any icon */
  mwinSoftSelectIcon (mwin, mwin->FilesSoftSelNum, False);

  /* Special: Move the Widget to another x-y-coordinates */
  if (cnt == 1 /*mwin->SelectedIcons == 1*/ )
    {
      if (mode)
	{
	  strcpy (mwin->Copy->Dest, mwin->DirName);
	  /* if there is only one file to copy, it well copy */
	  /* the file/directory to destination with the same */
	  /* name */
	  strcpy (mwin->Copy->Dest+strlen(mwin->Copy->Dest),
		  "/");
	  strcpy (mwin->Copy->Dest+strlen(mwin->Copy->Dest),
		  basename (data));
	}
      return True;
    }
  
  /* another situations */
  return False;
}

int mwinCanCopy (Mainwin_t mwin, int x, int y)
{
  return mwinCanCopy_x (mwin, x, y, False);
}

int mwinCopy (Mainwin_t mwin, int x, int y)
{
  static char * files = NULL;
  static int len = 0;
  char * data;
  int ret, i;

  DEBUG1 ("mwinCopy:");

  if (files != NULL)
    {
      MFree (files, len);
      files = NULL;
    }

  /* view copy window */
  ret = mwinCanCopy_x (mwin, x, y, True);

  /* copy can not proceed */
  if (!ret)
    return False;

  /* get filelist from Clipboard */
  len = GetDataFromClipboard (&files);
  
  if (len <= 1) return False;

  /* Remove old values */
  mwin->Copy->Source = StrReplace (mwin->Copy->Source, NULL);

  /* Make filelist */
  data = gmalloc (len, __LINE__, __FILE__);
  for (data[0] = 0, i=0; i<len-1; i++)
    {
      if (files[i] == 0)
	{
	  strcpy (data+strlen(data), " ");
	  strcpy (data+strlen(data), &files [i+1]);
	}
      else if (i == 0)
	strcpy (data+strlen(data), &files [i]);
    }

  /* give the filelist to the copy procedure */
  mwin->Copy->Source = StrReplace (mwin->Copy->Source, data);

  MFree (data, len);
  
  ret = cpyView (mwin->shlMainwin, mwin->Copy);

  return ret;
}

#ifdef ABS
#undef ABS
#endif
#define ABS(x) ((x) < 0 ? -(x) : (x))

int mwinKey (Mainwin_t mwin, KeySym key, unsigned int state)
{
  Dchain_t dir;
  int value, slidersize, incrreturn, pageincrret, maxvalue;
  int num = 0, first = 0, last = 0;
  int ChangeDir = False;
  int i;

  DEBUG2 ("mwinKey: key=(%i)\n", key);

  for (i=0; i<mwin->dirEntryCount; i++)
    {
      if (mwinSelected (mwin, i))
	{
	  last = i;
	  if (first == 0)
	    first = i;
	}
    }

  switch (key)
    {
    case XK_Up:
      if (first == 0 && last == 0) return False;
      if (first == 0)
	{
	  ChangeDir = True;
	  num = 0;
	}
      else
	{
	  for (i=first-1; i>=0; i--)
	    {
	      if (ABS (mwin->dirWidgets[i]->iconX - mwin->dirWidgets[first]->iconX)
	          <= mwin->PixmBoxWidth)
		{
		  num = i;
		  ChangeDir = True;
		  break;
		}
	    }
	}
      break;

    case XK_Down:
      if (last == first && last == (mwin->dirEntryCount - 1))
	return False;
      if (last == (mwin->dirEntryCount - 1))
	{
	  num = mwin->dirEntryCount - 1;
	  ChangeDir = True;
	}
      else
	{
	  num = last;
	  for (i=last+1; i<mwin->dirEntryCount; i++)
	    {
	      if (ABS (mwin->dirWidgets[i]->iconX - mwin->dirWidgets[last]->iconX)
	          <= mwin->PixmBoxWidth)
		{
		  ChangeDir = True;
		  num = i;
		  break;
		}
	    }
	}
      break;

      /* one level upper */
    case XK_Left:
      if (first == 0 && last == 0) return False;
      if (first == 0)
	{
	  ChangeDir = True;
	  num = 0;
	}
      else
	{
	  for (i=first-1; i>=0; i--)
	    {
	      if (ABS (mwin->dirWidgets[i]->iconY - mwin->dirWidgets[first]->iconY)
	          <= mwin->PixmBoxHeight)
		{
		  num = i;
		  ChangeDir = True;
		  break;
		}
	    }
	}
      break;

      /* one level right */
    case XK_Right:
      if (first == last && last == (mwin->dirEntryCount - 1))
	return False;
      if (last == (mwin->dirEntryCount - 1))
	{
	  num = mwin->dirEntryCount - 1;
	  ChangeDir = True;
	}
      else
	{
	  num = last;
	  for (i=last+1; i<mwin->dirEntryCount; i++)
	    {
	      if (ABS (mwin->dirWidgets[i]->iconY - mwin->dirWidgets[last]->iconY)
	          <= mwin->PixmBoxHeight)
		{
		  ChangeDir = True;
		  num = i;
		  break;
		}
	    }
	}
      break;

    case XK_Prior:
      XmScrollBarGetValues (mwin->scrVertFiles,
			    &value,
			    &slidersize,
			    &incrreturn,
			    &pageincrret);
#if 1
      scrAddValue (mwin->scrVertFiles,
		   - pageincrret / incrreturn,
		   SCR_DOWN|SCR_REDRAW);
#else
      value = min (value - 1, (value - pageincrret / incrreturn) + 1);
      XmScrollBarSetValues (mwin->scrVertFiles,
			    min (maxvalue-slidersize, max (0, value - 1)),
			    slidersize,
			    incrreturn,
			    pageincrret,
			    True);
#endif
      break;

    case XK_Next:
      XmScrollBarGetValues (mwin->scrVertFiles,
			    &value,
			    &slidersize,
			    &incrreturn,
			    &pageincrret);
#if 1
      scrAddValue (mwin->scrVertFiles, pageincrret / incrreturn,
		   SCR_DOWN|SCR_REDRAW);
#else
      XtVaGetValues (mwin->scrVertFiles,
		     XmNmaximum,       &maxvalue,
		     NULL);
      XmScrollBarGetValues (mwin->scrVertFiles,
			    &value,
			    &slidersize,
			    &incrreturn,
			    &pageincrret);
      value = max (value + 1, (value + pageincrret / incrreturn) - 1);
      XmScrollBarSetValues (mwin->scrVertFiles,
			    min (maxvalue-slidersize, value + 1),
			    slidersize,
			    incrreturn,
			    pageincrret,
			    True);
#endif
      break;

    case XK_Home:
      num = 0;
      ChangeDir = True;
      break;

    case XK_End:
      num = mwin->dirEntryCount - 1;
      ChangeDir = True;
      break;

    case XK_Return:
      mwinExec (mwin, NULL);
      break;
    }

  /* printable key pressed, search for entry with this character */
  if (! (state & ControlMask) && key >= 32 && key <= 256)
    {
      int i;
      for (i=last+1;
	   i<mwin->dirEntryCount;
	   i++)
	{
	  char * p;
	  p = dirName (mwin->dirWidgets[i]->dirListEntry);
	  if ((int) p[0] == (int) key)
	    {
	      num = i;
	      ChangeDir = True;
	      break;
	    }
	}
      /* search from beginning */
      if (i==mwin->dirEntryCount)
	{
	  for (i=0;
	       i<last;
	       i++)
	    {
	      char * p;
	      p = dirName (mwin->dirWidgets[i]->dirListEntry);
	      if ((int) p[0] == (int) key)
		{
		  num = i;
		  ChangeDir = True;
		  break;
		}
	    }
	}
      if (! ChangeDir)
	XBell (display, 0);
	    
    }

  if (ChangeDir)
    {
      mwinShowIcon (mwin, num);

      mwin->SelectedIcons = 1;
      for (i=0; i<mwin->dirEntryCount; i++)
	{
	  if (mwinSelected (mwin, i))
	    {
	      mwinSelectIcon (mwin, i, False);
	    }
	}
      mwinSelectIcon (mwin, num, True);
      mwinDispStatus (mwin);
    }

  return True;
}

int mwinShowIcon (Mainwin_t mwin, int num)
{
  int x, y, x2, y2;
  y  = min (mwin->dirWidgets[num]->iconY,
	    mwin->dirWidgets[num]->nameY);
  y2 = max (mwin->dirWidgets[num]->iconY + mwin->dirWidgets[num]->iconHeight
	    ,
	    mwin->dirWidgets[num]->nameY + mwin->dirWidgets[num]->nameHeight
	    );
  x  = min (mwin->dirWidgets[num]->iconX,
	    mwin->dirWidgets[num]->nameX);
  x2 = max (mwin->dirWidgets[num]->iconX+mwin->dirWidgets[num]->iconWidth,
	    mwin->dirWidgets[num]->nameX+mwin->dirWidgets[num]->nameWidth);
  
  while (y < - mwin->FilesViewY)
    {
      scrAddValue (mwin->scrVertFiles,
		   - (- mwin->FilesViewY - y + (10 - 1)) / 10,
		   SCR_DOWN|SCR_REDRAW);
      break;
    }
  while (y2 > mwin->dimFilesHeight - mwin->FilesViewY)
    {
      scrAddValue (mwin->scrVertFiles,
		   (y2 -
		    (mwin->dimFilesHeight - mwin->FilesViewY)
		    + (10 - 1))
		   / 10,
		   SCR_DOWN|SCR_REDRAW);
      break;
    }
  
  while (x2 > mwin->dimFilesWidth - mwin->FilesViewX &&
	 x > - mwin->FilesViewX + 10)
    {
      scrAddValue (mwin->scrHorizFiles,
		   (x2 -
		    (mwin->dimFilesWidth - mwin->FilesViewX)
		    + (10 - 1))
		   / 10,
		       SCR_DOWN|SCR_REDRAW);
      break;
    }
  while (x < - mwin->FilesViewX)
    {
      scrAddValue (mwin->scrHorizFiles,
		       - (- mwin->FilesViewX - x + (10 - 1)) / 10,
		   SCR_DOWN|SCR_REDRAW);
      break;
    }

  return True;
}

int mwinSetFilesScrollBars (Mainwin_t mwin)
{
  /* set scollbar sizes for the new value */
  if (mwin->scrVertFiles != NULL && mwin->dimFilesWidth > 0)
    {
      scrSetValues (mwin->scrVertFiles,
		    &mwin->FilesViewY,
		    mwin->FilesMaxYPaint,
		    mwin->dimFilesHeight,
		    10 /*mwin->PixmHeight*/ );
      scrSetValues (mwin->scrHorizFiles,
		    &mwin->FilesViewX,
		    mwin->FilesMaxXPaint,
		    mwin->dimFilesWidth,
		    10 /*mwin->PixmWidth*/ );
    }
  return True;
}

int mwinLoadColormap (Mainwin_t mwin)
{
  if (mwin->SelectedIcons == 1 &&
      mwin->dirWidgets[lastSelected]->iconColormap > 0 &&
      lastColormap != mwin->dirWidgets[lastSelected]->iconColormap)
  {
    lastColormap = mwin->dirWidgets[lastSelected]->iconColormap;
    XSetWindowColormap (XtDisplay (mwin->shlMainwin),
    			XtWindow (mwin->shlMainwin),
			mwin->dirWidgets[lastSelected]->iconColormap);
			
    DEBUG3 ("mwinSel*OrCopy: load colormap %i for file[%s]\n",
              lastColormap,
	      dirName (mwin->dirWidgets[lastSelected]->dirListEntry));
    
    DefaultColormap_Loaded = False;
  }
  else if (! DefaultColormap_Loaded)
  {
    lastColormap = XDefaultColormapOfScreen (XtScreen(mwin->shlMainwin));
    XSetWindowColormap (display, XtWindow (mwin->shlMainwin), lastColormap);

    DEBUG3 ("mwinSel*OrCopy: load default colormap %i for file[%s]\n",
              lastColormap,
	      dirName (mwin->dirWidgets[lastSelected]->dirListEntry));

    DefaultColormap_Loaded = True;
  }
  
  return True;
}

int mwinFilesDest (Mainwin_t mwin)
{
  int i;
  
  mwin->SelectedIcons = 0;
  mwinLoadColormap (mwin);
  
  /* free memory */
  if (mwin->dirWidgets != NULL)
    {
      for (i=0; i<mwin->dirEntryCount; i++)
	{
	  /* free colormap */
	  if (mwin->dirWidgets[i]->iconColormap > 0)
	    XFreeColormap (display, mwin->dirWidgets[i]->iconColormap);
	    
	  /* free loaded pixmaps */
	  if (mwin->dirWidgets[i]->iconPixm > 0)
	    XFreePixmap (display, mwin->dirWidgets[i]->iconPixm);
	    
	  MFree (mwin->dirWidgets[i], sizeof (dirWidgets_t));
	}
      MFree (mwin->dirWidgets, sizeof (DirWidgets_t) * mwin->dirEntryCount);
      mwin->dirEntryCount = 0;
    }

  /* delete old directory entries */
  UdchEmpty  (mwin->dirEntry,    (int (*)(void*)) dirDest);
  
  /* remove the control directory entry from memory */
  dirDelete (mwin->dirPathName);

  return True;
}


