/************************************************************************
 *  (C)opyright 1996/1997 by Khamis Abu El-Komboz                       *
 *  xkhfm 1.0 can be copied and distributed freely for any              *
 *  non-commercial purposes. xkhfm can only be incorporated             *
 *  into commercial software with the permission of the current author. *
 *  Mail: khamis@multix.de                                              *
 ************************************************************************/
/*
 * $Author: khamis $
 * $Id: mwinMisc.c,v 1.1 1997/01/07 08:38:27 khamis Exp $
 * $Source: /home/khamis/source/xkhfm/xkhlib/xkhsrc/RCS/mwinMisc.c,v $
 */
#define IS_EXTERN
#define EXTERN		extern

#include <stdio.h>
#include <stdlib.h>
#include <sys/param.h>
#include <string.h>
#include <fcntl.h>
#include "Xinc.h"
#include <X11/Xaw/Paned.h>
#include <Xm/DrawingA.h>
#include <Xm/ScrollBar.h>
#include <Xm/Command.h>

#include "defines.h"
#include "Umem.h"
#include "Uxutil.h"
#include "Umessage.h"
#include "Umenu.h"
#include "Urcnfvoid.h"
#include "Ureadarg.h"
#include "Ufileutil.h"
#include "Ustrutil.h"
#include "about.h"
#include "quit.h"
#include "help.h"
#include "xmisc.h"
#include "window.h"
#include "mainwin.h"

int mwinGetSelectedFiles (Mainwin_t mwin, char ** files)
{
  int i, Found = False;
  
  *files = StrReplace (*files, NULL);

  /* add selected list */
  for (i=0; i<mwin->dirEntryCount; i++)
    {
      if (dirSelected (mwin->dirWidgets[i]->dirListEntry))
	{
	  *files =
	    StrAppend (*files,
		       UPathAppend
		       (mwin->DirName,
			dirName (mwin->dirWidgets[i]->dirListEntry)));
	  *files = StrAppend (*files, " ");
	  Found = True;
	}
    }

  /* display selected list in the file text */
  if (! Found)
    {
      *files =
	StrReplace (*files,
		    dirTreePathName
		    (mwin->dirTreeWidgets[mwin->CurrentTreeNum]->dirTreeEntry));
    }
  return True;
}

int mwinChangeTo (Mainwin_t mwin, char * path)
{
  int ret, i;
  Dchain_t dir = dirAdd (NULL, path);

  if (dirType (dir) != TYPE_DIR)
    {
      char * bn;
      ret = mwinChangePathName (mwin, dirname (path));
      if (!ret)
	return False;
      bn = basename (path);
      for (i=0; i<mwin->dirEntryCount; i++)
	{
	  if (strcmp (dirName (mwin->dirWidgets[i]->dirListEntry), bn) == 0)
	    {
	      mwinSelectIcon (mwin, i, True);
	      mwinShowIcon   (mwin, i);
	      break;
	    }
	}
    }
  else
    ret = mwinChangePathName (mwin, path);

  XMapRaised (XtDisplay (mwin->shlMainwin), XtWindow (mwin->shlMainwin));

  return ret;
}

int mwinProp (Mainwin_t mwin)
{
  int ret = False;
  Dchain_t propList;

  if (mwin->Prop == NULL)
  {
    propList = winListAddProp(winList, "Properities");
    mwin->Prop = winObj (propList);
  }
  if (mwin->Prop == NULL)
  {
    XBell (display, 0);
    return False;
  }

  ret = mwinGetSelectedFiles (mwin, &mwin->Prop->Files);
  if (!ret)
    return False;
    
  /* view the dialog */
  return propView (mwin->shlMainwin, mwin->Prop);
}

int mwinFind (Mainwin_t mwin)
{
  int ret = False;
  Dchain_t fndList;

  if (mwin->Find == NULL)
    {
      fndList = winListAddFind(winList, "Find");
      mwin->Find = winObj (fndList);
      if (mwin->Find == NULL)
      {
        XBell (display, 0);
        return False;
      }
      mwin->Find->ClProc = (void (*)(void*,char*)) mwinChangeTo;
      mwin->Find->ClData = mwin;
    }
  if (mwin->Find == NULL)
    return False;

  mwin->Find->Files = StrReplace (mwin->Find->Files, mwin->DirName);
    
  /* view the dialog */
  return fndView (mwin->shlMainwin, mwin->Find);
}

int mwinCopyMove (Mainwin_t mwin, int mode)
{
  int i, ret;
  Dchain_t list;

  if (mwin->Copy == NULL)
  {
    list = winListAddCopy(winList, "Copy - Move");
    mwin->Copy = winObj (list);
  }
  
  if (mwin->Copy == NULL)
  {
    XBell (display, 0);
    return False;
  }

  ret = mwinGetSelectedFiles (mwin, &mwin->Copy->Source);
  if (!ret) return False;

  if (mode)
    mwin->Copy->info.mode |= CPY_MOVE;
  else
    mwin->Copy->info.mode &= ~CPY_MOVE;

  /* view copy window */
  return cpyView (mwin->shlMainwin, mwin->Copy);
}

int mwinDel (Mainwin_t mwin, int mode)
{
  int len, ret;
  Dchain_t list;
  
  if (mwin->Remove == NULL)
  {
    list = winListAddRemove (winList, "Remove");
    mwin->Remove = winObj (list);
  }
  
  if (mwin->Remove == NULL)
  {
    XBell (display, 0);
    return False;
  }

  /* Remove old values in the remove-window */
  mwin->Remove->Files = StrReplace (mwin->Remove->Files, NULL);

  if (mode != MWIN_READ_NOTHING)
  {
    /* get filelist from Clipboard */
    if (mode == MWIN_READ_FROM_CLIPBOARD)
      len = GetDataFromClipboard (&mwin->Remove->Files);
    else 
      ret = mwinGetSelectedFiles (mwin, &mwin->Remove->Files);

    if (len <= 0)
      return False;
  }
  
  /* view remove window */
  return remView (mwin->shlMainwin, mwin->Remove);
}


int mwinMkDir (Mainwin_t mwin)
{
  Dchain_t list;
  int i;
  if (mwin->MkDir == NULL)
  {
    list = winListAddMkDir (winList, "Make a new directory");
    mwin->MkDir = winObj (list);
  }

  StrReplace (mwin->MkDir->DirName, NULL);

  /* change to current path */
  dirChangeDir (mwin->DirName);

  /* view remove window */
  return mkdirView (mwin->shlMainwin, mwin->MkDir);
}

