/************************************************************************
 *  (C)opyright 1996/1997 by Khamis Abu El-Komboz                       *
 *  xkhfm 1.0 can be copied and distributed freely for any              *
 *  non-commercial purposes. xkhfm can only be incorporated             *
 *  into commercial software with the permission of the current author. *
 *  Mail: khamis@multix.de                                              *
 ************************************************************************/
/*
 * $Author: khamis $
 * $Id: mwinReadDir.c,v 1.1 1996/10/28 03:37:25 khamis Exp khamis $
 * $Source: /home/xkhfm/xkhlib/xkhsrc/RCS/mwinReadDir.c,v $
 */
#define IS_EXTERN
#define EXTERN		extern

#include <stdio.h>
#include <stdlib.h>
#include <sys/param.h>
#include "Xinc.h"

#include "defines.h"
#include "Umem.h"
#include "Ustrutil.h"
#include "sort.h"
#include "xmisc.h"
#include "Umessage.h"
#include "window.h"
#include "mainwin.h"

int mwinSetTree (Mainwin_t mwin, Dchain_t tree, int pos)
{
  Dchain_t tt;
  int i = pos;

  if (tree == NULL)
    return pos;

  ASSERT (pos <= mwin->TreeCount);

  for (i=pos, tt = UdchNext (tree);
       tt != NULL && i<mwin->TreeCount;
       tt = UdchNext (tt))
    {
      mwin->dirTreeWidgets[i] = gcalloc (sizeof (dirTreeWidgets_t), __LINE__, __FILE__);
      mwin->dirTreeWidgets[i]->dirTreeEntry = tt;

      i++;

      i = mwinSetTree (mwin, ((DirTree_t) UdchClient(tt))->subTree, i);
    }

  return i;
}

int mwinLock (Mainwin_t mwin, int num)
{

#if 0
  DEBUG2 ("mwinLock: Locked=(%i)\n", mwin->Locked [num]);
#endif

  while (1)
    {
      /* wait to unlock the procedure */
      while (mwin->Locked[num]) {};
      mwin->Locked[num] ++;
      /* procedure locked successfully */
      if (mwin->Locked[num] == 1)
	break;
      /* other people was faster to lock the procedure */
      mwin->Locked[num] --;
    }

#if 0
  DEBUG2 ("mwinLock: End Locked=(%i)\n", mwin->Locked[num]);
#endif

  return True;
}

int mwinUnLock (Mainwin_t mwin, int num)
{
  mwin->Locked[num] --;
  return True;
}

int mwinReadDir (Mainwin_t mwin)
{
  Dchain_t dirList;
  int i, ret;

  DEBUG2 ("mwinReadDir: (%s)\n", mwin->DirName);

  /* lock the Procedure */
  mwinLock (mwin, 0);

  /* rebuild the Tree if the path not found*/
  if (mwinTreeSelPath (mwin) == False)
    {
      mwinBuildTree (mwin, TREE_BUILD);
    }

  /* destroy filelist */
  mwinFilesDest (mwin);
  
  /****** get and save state of the new directory pathname ***/
  mwin->dirPathName = dirAdd (NULL, mwin->DirName);
  dirStat (mwin->dirPathName);

  /****** Read Directroy Entries ****/
  mwin->dirEntryCount =
    dirReadDir (mwin->dirEntry, mwin->DirName, mwin->DirFilter, False);

  /* initial memory for the icon widgets */
  mwin->dirWidgets = gcalloc (sizeof (DirWidgets_t) * mwin->dirEntryCount, __LINE__, __FILE__);
  for (i=0, dirList=UdchNext (mwin->dirEntry);
       dirList != NULL;
       dirList = UdchNext (dirList), i++)
    {
      mwin->dirWidgets[i] = gcalloc (sizeof (dirWidgets_t), __LINE__, __FILE__);
      mwin->dirWidgets[i]->dirListEntry = dirList;
    }

  /* set new title of window on the new directory */
  mwinTitle (mwin);

  /* sort files as requiered */
  mwinSort (mwin, mwin->SortMethode);

  /* Clear old entry of the drawing area */
  mwin->FilesClearIcons = True;

  mwinUnLock (mwin, 0);

  DEBUG1 ("mwinReadDir: end");
  return True;
}

#define CMP_DEC(a, b) ((a > b) ? 1 : ((a < b) ? -1 : 0));


int mwinComp (const int pos1, const int pos2, Mainwin_t mwin)
{
  Dchain_t d1 = mwin->dirWidgets [pos1]->dirListEntry;
  Dchain_t d2 = mwin->dirWidgets [pos2]->dirListEntry;
  char * str1 = NULL, * str2 = NULL;
  int ret = 0, t1 = 1, t2 = 1, v1, v2;

  /* first criterium */
  switch (mwin->SortMethode)
    {
    case SORT_NAME:
      str1 = dirName (d1);
      str2 = dirName (d2);
      break;

    case SORT_SIZE:
      v1 = dirSize (d1);
      v2 = dirSize (d2);
      break;

    case SORT_MOD_TIME:
      v1 = dirModTime (d1);
      v2 = dirModTime (d2);
      break;

    case SORT_CREATE_TIME:
      v1 = dirCreateTime (d1);
      v2 = dirCreateTime (d2);
      break;

    case SORT_ACCESS_TIME:
      v1 = dirAccessTime (d1);
      v2 = dirAccessTime (d2);
      break;

    case SORT_USER_ID:
      v1 = dirUserId (d1);
      v2 = dirUserId (d2);
      break;

    case SORT_GRP_ID:
      v1 = dirGrpId (d1);
      v2 = dirGrpId (d2);
      break;

    case SORT_USER_NAME:
      str1 = dirUserName (d1);
      str2 = dirUserName (d2);
      break;

    case SORT_GRP_NAME:
      str1 = dirGrpName (d1);
      str2 = dirGrpName (d2);
      break;

    case SORT_INODE:
      v1 = dirINode (d1);
      v2 = dirINode (d2);
      break;

    case SORT_LINKS:
      v1 = dirLinks (d1);
      v2 = dirLinks (d2);
      break;

    case SORT_UNSORTED:
      v1 = v2 = 0;
      return 0;

    default:
      v1 = v2 = 0;
      break;
    }

  /* post directories as first */
  if (mwin->DirFirst)
    {
      t1  = (dirType (d1) == TYPE_DIR) ? 0 : 1;
      t2  = (dirType (d2) == TYPE_DIR) ? 0 : 1;
      ret = CMP_DEC (t1, t2);
    }

  /* compare, if not directories and not "DirFirst" */
  if (ret == 0 && t1 == 1)
    {
      if (mwin->SortMethode & (SORT_NAME|SORT_GRP_NAME|SORT_USER_NAME))
	ret = strcmp (str1, str2);
      else
	ret = CMP_DEC (v1, v2);

      /* if descending, then toggle the return value */
      if (mwin->Descending)
	  ret = - ret;
    }

  /* second criterium,if the icons are aequevalent */
  if (ret == 0)
    ret = strcmp (dirName (d1), dirName (d2));

  return ret;
}

int mwinSwap (const int pos1, const int pos2, Mainwin_t mwin)
{
  Dchain_t d1 = mwin->dirWidgets [pos1]->dirListEntry;
  Dchain_t d2 = mwin->dirWidgets [pos2]->dirListEntry;
  int ret = 0, r1, r2;

  mwin->dirWidgets [pos1]->dirListEntry = d2;
  mwin->dirWidgets [pos2]->dirListEntry = d1;

  return TRUE;
}

int mwinCurrentSort (Mainwin_t mwin, Widget w)
{
  if (w == NULL)
    return False;

  /* disable the button with the current sorting methode */
  if (mwin->LastSortWidget != NULL)
    XtSetSensitive (mwin->LastSortWidget, True);
  
  XtSetSensitive (w, False);

  mwin->LastSortWidget = w;

  return True;
}
int mwinSort (Mainwin_t mwin, int sortMethode)
{
  DEBUG1 ("mwinSort:");

  mwin->SortMethode = sortMethode;

  OTT_qsort (mwin,
	     mwin->BeginSortingAt,
	     mwin->dirEntryCount - 1,
	     (int (*) (const int, const int, void*)) mwinComp,
	     (int (*) (const int, const int, void*)) mwinSwap);
  dirSort (mwin->dirEntry, sortMethode);

  DEBUG1 ("mwinSort: Ende");

  return True;
}

int mwinTreeSelPath (Mainwin_t mwin)
{
  int ret = False, i;

  DEBUG2 ("mwinTreeSelPath: select (%s)\n", mwin->DirName);

  for (i=0; i<mwin->TreeCount; i++)
    {
      if (strcmp (dirTreePathName (mwin->dirTreeWidgets[i]->dirTreeEntry),
		  mwin->DirName) == 0)
	{
	  mwinTreeSelectIcon (mwin, i, True);
	  mwinViewSelectedIcon (mwin);
	  ret = True;
	  break;
	}
    }
  
  DEBUG2 ("mwinTreeSelPath: end ret=(%i)", ret);

  return ret;
}

int mwinAddToHistory (Mainwin_t mwin, char * PathName)
{
  Dchain_t dch;
  XmString strPathName;
  char *str = NULL;
  int i, j;

  /* find out if this directory already availiable */
  for (i=0; i<HIST_MAX_ITEMS; i++)
    {
      if (histList[i] != NULL && strcmp (histList[i], PathName) == 0)
	break;
    }

  /* already availiable, move to top */
  if (i<HIST_MAX_ITEMS && histList[i] != NULL)
    {
      for (j=i; j>0; j--)
	{
	  histList[j] = StrReplace (histList[j], histList[j-1]);
	}
      histList[0] = StrReplace (histList[0], PathName);
    }
  else
    {
      /* wrap the history items down */
      for (i=HIST_MAX_ITEMS-2; i>=0; i--)
	{
	  histList[i+1] = StrReplace (histList[i+1], histList[i]);
	}
      histList[0] = StrReplace (histList[0], PathName);
    }

  for (dch = UdchNext (winList);
       dch != NULL;
       dch = UdchNext (dch))
    {
      WindowList_t win = UdchClient (dch);
      if (win->Type != WIN_REGULAR)
	continue;

      /* wrap the history items down */
      for (i=0; i<HIST_MAX_ITEMS && histList[i] != NULL; i++)
	{
	  strPathName = XmStringCreateSimple (histList[i]);
	  XtVaSetValues (win->wins.mainwin->HistButtons[i],
			 XmNlabelString, strPathName,
			 NULL);
	  XmStringFree (strPathName);
	}
    }
  return True;
}

int mwinChangePathName (Mainwin_t mwin, char * PathName)
{
  int ret = False;

  BaisyCursor (mwin->shlMainwin);

  ret = dirChangeDir (PathName);
  if (ret)
    {
      /* add the current (old) pathname to History */
      mwinAddToHistory (mwin, mwin->DirName);

      strcpy (mwin->DirName, PathName);

#if 0
      /* global Pwd */
      strcpy (CurrentPath, mwin->DirName);
      cmdViewCurrentPath (cmdHist);
#endif

      mwinReadDir           (mwin);
      mwinViewSelectedIcon  (mwin);
      mwinTreeRedisp        (mwin);
      mwinCalc	 (mwin);
      mwin->FilesViewX = mwin->FilesViewY = 0;
      XMoveWindow (display, XtWindow (mwin->fomFiles), 0, 0);
      scrAddValue (mwin->scrVertFiles,  0, SCR_SET);
      scrAddValue (mwin->scrHorizFiles, 0, SCR_SET);
      mwinRedisp     (mwin);
      mwinDispStatus (mwin);

      /* correct entries in the Window-pulldown menu */
      winBuildWinMenu (NULL, NULL);
    }

  NormalCursor (mwin->shlMainwin);

  if (!ret)
    {
      static char MessText[] =
	"Cannot change to\n"
	"%s\n"
	"Permission denied";
      char * Mess;
      int len = strlen (PathName) + strlen (MessText) + 5;
      Mess = gmalloc (len, __LINE__, __FILE__);
      sprintf (Mess, MessText, PathName);
      ErrorBox (mwin->shlMainwin,
		PathName,
		Mess);
      MFree (Mess, len);
    }

  return ret;
}

mwinBuildTree (Mainwin_t mwin, int mode)
{
  int i, ret;

  DEBUG2 ("mwinBuildTree: PathName(%s)", mwin->DirName);

  mwinLock (mwin, 1);

  mwinTreeDest (mwin, mode);
  
  /* build the tree to the name of diepest directory */
  if (mode & TREE_BUILD)
    {
      dirTreeBuild (mwin->dirTree, mwin->DirName, False);
    }

  mwin->TreeCount = dirTreeCount (mwin->dirTree);
  
  mwin->dirTreeWidgets = gcalloc (mwin->TreeCount *
				  sizeof (DirTreeWidgets_t), __LINE__, __FILE__);
  ret = mwinSetTree (mwin, mwin->dirTree, 0);

  /* calculate positions for Icons */
  mwinCalcTree (mwin);

  /* Change to the Current Path if equiered */
  if (! (mode & BUILD_DONOT_CHANGE_PATH))
    {
      mwinTreeSelPath (mwin);
    }

  mwin->TreeClearIcons = True;

  mwinUnLock (mwin, 1);
  
  DEBUG2 ("mwinBuildTree: (%i) Trees declared\n", ret);
  return ret;
}


int mwinViewSelectedIcon (Mainwin_t mwin)
{
  int i = mwin->CurrentTreeNum;
  int x, y, x2, y2, step;

  if (mwin->scrVertTree == NULL || mwin->dimTreeWidth == 0)
    return False;

  y  = min (mwin->dirTreeWidgets[i]->iconY,
	    mwin->dirTreeWidgets[i]->nameY);
  y2 = max (mwin->dirTreeWidgets[i]->iconY
	    + mwin->dirTreeWidgets[i]->iconHeight
	    ,
	    mwin->dirTreeWidgets[i]->nameY
	    + mwin->dirTreeWidgets[i]->nameHeight
	    );
  x  = min (mwin->dirTreeWidgets[i]->iconX,
	    mwin->dirTreeWidgets[i]->nameX);
  x2 = max (mwin->dirTreeWidgets[i]->iconX
	    + mwin->dirTreeWidgets[i]->iconWidth,
	    mwin->dirTreeWidgets[i]->nameX
	    + mwin->dirTreeWidgets[i]->nameWidth);
  step = mwin->PixmTreeHeight;
  /* view selected icon */
  if (y < - mwin->TreeViewY)
    {
#if 1
      scrAddValue (mwin->scrVertTree,
		   - (- mwin->TreeViewY
		      - y
		      + (mwin->PixmTreeHeight - 1)
		      ) / mwin->PixmTreeHeight,
		   SCR_DOWN|SCR_REDRAW);
#else
      scrSetValue (mwin->scrVertTree,
		 &mwin->TreeViewY,
		 mwin->TreeCount * (mwin->PixmTreeHeight+2),
		 mwin->dimTreeHeight,
		 mwin->PixmTreeHeight+2,
		 mwin->dirTreeWidgets[i]->iconY);
#endif
    }
  else if (y2 > mwin->dimTreeHeight - mwin->TreeViewY)
    {
#if 1
      scrAddValue (mwin->scrVertTree,
		   (y2 -
		    (mwin->dimTreeHeight - mwin->TreeViewY)
		    + (step - 1))
		   / step,
		   SCR_DOWN|SCR_REDRAW);
#else
      scrSetValue (mwin->scrVertTree,
		   &mwin->TreeViewY,
		   mwin->TreeCount * (mwin->PixmTreeHeight+2),
		   mwin->dimTreeHeight,
		   mwin->PixmTreeHeight+2,
		   mwin->dirTreeWidgets[i]->iconY +
		   mwin->dirTreeWidgets[i]->iconHeight);
#endif
    }
      
  step = 1;
  if (x < - mwin->TreeViewX)
    {
#if 1
      scrAddValue (mwin->scrHorizTree,
		   - (- mwin->TreeViewX - x + (step - 1)) / step,
		   SCR_DOWN|SCR_REDRAW);

#else
      scrSetValue (mwin->scrHorizTree,
		   &mwin->TreeViewX,
		   mwin->TreeMaxXPaint,
		   mwin->dimTreeWidth,
		   1,
		   mwin->dirTreeWidgets[i]->iconX);
#endif
    }
  else if (x2 > mwin->dimTreeWidth - mwin->TreeViewX)
    {
#if 1
      scrAddValue (mwin->scrHorizTree,
		   (x2 -
		    (mwin->dimTreeWidth - mwin->TreeViewX)
		    + (step - 1))
		   / step,
		   SCR_DOWN|SCR_REDRAW);
#else
      scrSetValue (mwin->scrHorizTree,
		   &mwin->TreeViewX,
		 mwin->TreeMaxXPaint,
		 mwin->dimTreeWidth,
		 1,
		 mwin->dirTreeWidgets[i]->nameX +
		 mwin->dirTreeWidgets[i]->nameWidth);
#endif
    }

  return True;
}
