/************************************************************************
 *  (C)opyright 1996/1997 by Khamis Abu El-Komboz                       *
 *  xkhfm 1.0 can be copied and distributed freely for any              *
 *  non-commercial purposes. xkhfm can only be incorporated             *
 *  into commercial software with the permission of the current author. *
 *  Mail: khamis@multix.de                                              *
 ************************************************************************/
/*
 * $Author: khamis $
 * $Id: mwinSave.c,v 1.1 1996/10/28 03:36:45 khamis Exp khamis $
 * $Source: /home/xkhfm/xkhlib/xkhsrc/RCS/mwinSave.c,v $
 */
#define IS_EXTERN
#define EXTERN		extern

#include <stdio.h>
#include <stdlib.h>
#include <sys/param.h>
#include "Xinc.h"

#include "defines.h"
#include "Umem.h"
#include "Uxutil.h"
#include "Umessage.h"
#include "Urcnfvoid.h"
#include "Ureadarg.h"
#include "Ufileutil.h"
#include "Ustrutil.h"
#include "Ucnfutil.h"
#include "Udchain.h"
#include "window.h"
#include "mainwin.h"

static char RessTitle[] =
"##\n"
"##Khamis Utilities Version 1.0\n"
"##Setup file to restore the last configuration\n"
"\n";

int mwinSaveSetup (void)
{
  Dchain_t winDch;
  WindowList_t win;
  Conf_t cnf;
  char ressfile [MAXPATHLEN];
  char line [1024];
  int ret = False;
  int i;

  DEBUG1 ("mwinSaveSetup:");

  sprintf (ressfile, "%s/.xkhfmrc", getenv ("HOME"));

  cnf = confOpen (ressfile, RessTitle);
  
  if (cnf == NULL)
    return False;

  DEBUG1 ("mwinSaveSetup: Delete Section");

  /* delete old entries */
  confDelSect (cnf, WINDOW_SECTION);

  DEBUG1 ("mwinSaveSetup: Save Coordinates");

  for (i=0,winDch=UdchNext (winList);
       winDch != NULL;
       winDch=UdchNext (winDch), i++)
    {
      win = (WindowList_t)UdchClient (winDch);

      if (win->Type == WIN_REGULAR)
	{
	  char* argv[1];
	  /* if window hidden, don't save it */
	  if (win->wins.mainwin->shlMainwin == NULL)
	    continue;
	  /* format:      window:n :x :y :w :h :t :f :vm:s :st:path*/
	  sprintf (line, "window:%i:%i:%i:%i:%i:%i:%i:%i:%i:%i:%s",
		   i,
		   wdgX      (win->wins.mainwin->shlMainwin),
		   wdgY      (win->wins.mainwin->shlMainwin),
		   wdgWidth  (win->wins.mainwin->shlMainwin),
		   wdgHeight (win->wins.mainwin->shlMainwin),
		   wdgWidth  (win->wins.mainwin->fomTree),
		   wdgWidth  (win->wins.mainwin->fomFiles),
		   win->wins.mainwin->ViewMode,
		   win->wins.mainwin->Style,
		   win->State,
		   win->wins.mainwin->DirName);
	  argv[0] = line;
	  confAppend (cnf, ':', WINDOW_SECTION, 1, argv);
	}
      else if (win->Type == WIN_COMMAND)
	{
	  char *argv[1];
	  /* if window hidden, don't save it */
	  if (win->wins.cmd->dlgCmd == NULL)
	    continue;
	  sprintf (line, "command:%i:%i:%i:%i:%i:%i",
		   -1,
		   wdgX      (win->wins.cmd->dlgCmd),
		   wdgY      (win->wins.cmd->dlgCmd),
		   wdgWidth  (win->wins.cmd->dlgCmd),
		   wdgHeight (win->wins.cmd->dlgCmd),
		   win->State);
	  argv[0] = line;
	  confAppend (cnf, ':', WINDOW_SECTION, 1, argv);
	}
      else if (win->Type == WIN_FIND)
	{
	  char*argv[1];
	  /* if window hidden, don't save it */
	  if (win->wins.fnd->dlgFnd == NULL)
	    continue;
	  sprintf (line, "Find:%i:%i:%i:%i:%i:%i",
		   -1,
		   wdgX      (win->wins.fnd->dlgFnd),
		   wdgY      (win->wins.fnd->dlgFnd),
		   wdgWidth  (win->wins.fnd->dlgFnd),
		   wdgHeight (win->wins.fnd->dlgFnd),
		   win->State);
	  argv[0]=line;
	  confAppend (cnf, ':', WINDOW_SECTION, 1, argv);
	}
    }

  /* Save History for Working Direcories **/
  /* delete old entries */
  confDelSect (cnf, "History");
  for (i=0; i<HIST_MAX_ITEMS; i++)
    {
      char*argv[1];
      if (histList[i] != NULL)
      {
        argv[0] = histList[i];
	confAppend (cnf, ':', "History", 1, argv);
      }
    }

  /** save association list */
  assSave (assList, cnf);

  /** save icon extensions */
  assSave (iconExtensionList, cnf);

  /** read toolbar list */
  tbarSave (toolbarList, cnf);

  DEBUG1 ("mwinSaveSetup: Close cnfFile");

  /* close the resource file */
  confClose (cnf);

  cmdSaveHist (cmdList);
  
  DEBUG1 ("mwinSaveSetup: End");

  return ret;
}

int mwinReadSetup (void)
{
  Dchain_t winDch;
  WindowList_t win;
  Conf_t cnf;
  char ressfile [MAXPATHLEN];
  char line [CNF_MAXPARAMLEN];
  int ret = False;
  int i, pos, j;
  char * p;

  DEBUG1 ("mwinReadSetup:");

  /* init history list */
  for (i=0; i<HIST_MAX_ITEMS; i++)
    {
      histList[i] = NULL;
    }

  sprintf (ressfile, "%s/.xkhfmrc", getenv ("HOME"));

  cnf = confOpen (ressfile, RessTitle);
  
  if (cnf == NULL)
    return False;

  /* Read history section */
  pos = confSectPos (cnf, "History");
  if (pos >= 0)
    {
      /* Skip section name */
      confReLine (cnf, True);
      for (p = confReLine (cnf, True), i = 0;
	   p != NULL && p[0] != '[' && i < HIST_MAX_ITEMS;
	   p = confReLine (cnf, True), i++)
	{
	  histList[i] = StrReplace (histList[i], p);
	}
    }

  /* section position */
  pos = confSectPos (cnf, WINDOW_SECTION);

  if (pos >= 0)
    {
      /* Skip section name */
      confReLine (cnf, True);
      for (p = confReLine (cnf, True), i = 0;
	   p != NULL && p[0] != '[';
	   p = confReLine (cnf, True), i++)
	{
	  int args [9] = {0, 0, 0, 0, 0, 0, 0, 0, 0};
	  int x = 0, y = 0, w = 0, h = 0, tw = 0, fw = 0;
	  int State = 0;
	  
	  if (! Ureadarg (p, line, 0, ':'))
	    continue;
	  
	  /* read the arguments for x,y coordinates */
	  for (j=2; j < 11 && Ureadarg (p, line, j, ':'); j++)
	    {
	      args [j-2] = atoi (line);
	    }
	  Ureadarg (p, line, 0, ':');
	  
	  if (strcmp (line, "window") == 0)
	    {
	      winDch = winListAdd (winList, "mainwin");
	      
	      if (winDch == NULL)
		continue;
	      
	      win = UdchClient (winDch);
	      win->x          = args[0];
	      win->y          = args[1];
	      win->width      = args[2];
	      win->height     = args[3];
	      win->treeWidth  = args[4];
	      win->filesWidth = args[5];
	      win->wins.mainwin->ViewMode = args[6];
	      win->wins.mainwin->Style    = args[7];
	      State = args[8];
	      
	      /* saved Pathname */
	      if (Ureadarg (p, line, 11, ':'))
		strcpy (win->wins.mainwin->DirName, line);

	      /* view the window */
	      mwinView (win->wins.mainwin);
	    }
	  else if (strcmp (line, "command") == 0)
	    {
	      /* perhaps it's already added */
	      if (winListFindFirst (winList, WIN_COMMAND) != NULL)
		continue;
	      winDch = winListAddCmd (winList, "command");
	      win = UdchClient (winDch);
	      win->x      = args[0];
	      win->y      = args[1];
	      win->width  = args[2];
	      win->height = args[3];
	      State       = args[4];

	      cmdView (NULL, win->wins.cmd);

	    }
	  else if (strcmp (line, "Find") == 0)
	    {
#if 0 /* don't restore find windows */
	      winDch = winListAddFind (winList, "Find");
	      win = UdchClient (winDch);
	      win->x      = args[0];
	      win->y      = args[1];
	      win->width  = args[2];
	      win->height = args[3];
	      State       = args[4];

	      fndView (NULL, win->wins.fnd);
#else
	      continue;
#endif
	    }
	  else
	    {
	      continue;
	    }
	    
	  /* if window was iconized, iconize it */
	  if (State == WIN_ICONIZED)
	  {
	    XFlush (display);
	    xiconize (winGetShellFromList (winDch));
	    XFlush (display);
	    XSync  (display, 0);
	    XFlush (display);
	  }
	}
    } /* if pos > 0 */

  /** read association list */
  assRead (assList, cnf);
  
  /** read icon extension */
  assRead (iconExtensionList, cnf);
  
  /** read toolbar list */
  tbarRead (toolbarList, cnf);
  
  /* close the resource file */
  confClose (cnf);

  DEBUG1 ("mwinReadSetup: End");

  return ret;
}

