/************************************************************************
 *  (C)opyright 1996/1997 by Khamis Abu El-Komboz                       *
 *  xkhfm 1.0 can be copied and distributed freely for any              *
 *  non-commercial purposes. xkhfm can only be incorporated             *
 *  into commercial software with the permission of the current author. *
 *  Mail: khamis@multix.de                                              *
 ************************************************************************/
/*
 * $Author: khamis $
 * $Id: mwinTreeDisp.c,v 1.1 1997/01/07 08:38:27 khamis Exp $
 * $Source: /home/khamis/source/xkhfm/xkhlib/xkhsrc/RCS/mwinTreeDisp.c,v $
 */
#define IS_EXTERN
#define EXTERN		extern

#include <stdio.h>
#include <stdlib.h>
#include <sys/param.h>
#include <Xm/DrawnB.h>
#include "Xinc.h"
#include <X11/keysym.h>

#include "defines.h"
#include "Umem.h"
#include "Uxutil.h"
#include "Umessage.h"
#include "Umenu.h"
#include "Urcnfvoid.h"
#include "Ureadarg.h"
#include "Ufileutil.h"
#include "Ustrutil.h"
#include "Ucnfutil.h"
#include "about.h"
#include "quit.h"
#include "help.h"
#include "mainwin.h"

int mwinTreeDisp (Mainwin_t mwin)
{
  Dchain_t treeList;
  int i;

  DEBUG3 ("mwinTreeDisp: TreeMaxXPaint(%li), TreeCount(%li)\n",
	  mwin->TreeMaxXPaint,
	  mwin->TreeCount);

  for (i=0, mwin->TreeMaxXPaint = 0;
       i<mwin->TreeCount;
       i++)
    {
      mwinDispTreeIcon (mwin, i, False);
    }

  return True;
}

int mwinCalcTree (Mainwin_t mwin)
{
  Dchain_t treeList;
  long wX;
  long wY;
  int i;

  DEBUG1 ("mwinCalcTree:");

  /* Important to know the Window Height */
  mwin->PixmTreeHeight =
    max (pxmHeight [PIXM_tree_dir16],
	 (mwin->fntMainwin->ascent + mwin->fntMainwin->descent));

  wX = mwin->PixmTreeX;
  wY = mwin->PixmTreeY;
  for (i=0, mwin->TreeMaxXPaint = 0;
       i<mwin->TreeCount;
       i++)
    {
      DirTreeWidgets_t trWdg = mwin->dirTreeWidgets[i];
      
      if (trWdg == NULL)
	{
	  WARNING2 ("mwinTreeDisp: num(%i)==NULL\n", i);
	  continue;
	}

      treeList = trWdg->dirTreeEntry;

#ifdef DEBUG
      if (wY < 0)
	DEBUG2 ("mwinTreeDisp: wY=(%li)\n", wY);
#endif

      trWdg->iconWidth  = pxmWidth [PIXM_tree_dir16];
      trWdg->iconHeight = pxmHeight[PIXM_tree_dir16];

      /* name Width and Height */
      trWdg->nameWidth  = XTextWidth(mwin->fntMainwin,
				     dirTreeName (treeList),
				     strlen (dirTreeName (treeList)));
      trWdg->nameHeight = (mwin->fntMainwin->ascent
			   + mwin->fntMainwin->descent);

      /* Icon X/Y coordinates */
      trWdg->iconX 
	= mwin->PixmTreeX
	+ (dirTreeLevel (treeList) * pxmWidth [PIXM_tree_dir16]);
      trWdg->iconY
	= wY
	+ (max (trWdg->iconHeight, trWdg->nameHeight)
	   - trWdg->iconHeight) / 2;

      /* Text X/Y coordinates */
      trWdg->nameX
	= (trWdg->iconX
	   + trWdg->iconWidth
	   + DefTrSpaceBetwIconAndName);

      trWdg->nameY
	= wY
	+ (max (trWdg->nameHeight, trWdg->iconHeight)
	   - trWdg->nameHeight) / 2;

      wY = max (trWdg->iconY + trWdg->iconHeight,
		trWdg->nameY + trWdg->nameHeight)
	+ DefTrSpaceBetwLines;

      mwin->TreeMaxXPaint
	= max (mwin->TreeMaxXPaint,
	       trWdg->nameX + trWdg->nameWidth);
    }
  
  mwin->TreeMaxYPaint = wY;
      
  XResizeWindow (display, XtWindow (mwin->fomTree),
		 max (mwin->dimTreeWidth,  mwin->TreeMaxXPaint),
		 max (mwin->dimTreeHeight, mwin->TreeMaxYPaint));
  mwinSetScrollBars (mwin);
  XMoveWindow (display,
	       XtWindow (mwin->fomTree),
	       mwin->TreeViewX,
	       mwin->TreeViewY);

  return True;
}

int mwinDispTreeIcon (Mainwin_t mwin, int num, int mode)
{
  DirTreeWidgets_t dirWdg = mwin->dirTreeWidgets[num];
  int sel
    = dirTreeSelected     (dirWdg->dirTreeEntry) ||
      dirTreeSoftSelected (dirWdg->dirTreeEntry) ||
      dirTreeCopySelected (dirWdg->dirTreeEntry);
  int i;
  int x1, x2, y1, y2;
  if (mwin->TreeWindow == 0)
    return True;

#ifdef DEBUG
  if (XtWindow (mwin->fomTree) != mwin->TreeWindow)
    {
      ERROR3 ("mwinDispTreeIcon: XtWindow(mwin->fomTree)[%li] != "
	      "mwin->TreeWindow[%li]\n",
	      XtWindow (mwin->fomTree),
	      mwin->TreeWindow);
      exit (1);
    }
#endif

  dirWdg = mwin->dirTreeWidgets[num];

#if 1
  /* not viewed, only the Y-Coordinate */
  if (
      /*
	dirWdg->iconX > mwin->dimTreeWidth  - mwin->TreeViewX ||
	dirWdg->nameX + dirWdg->nameWidth  < - mwin->TreeViewX ||
	*/
      dirWdg->iconY > mwin->dimTreeHeight - mwin->TreeViewY ||
      dirWdg->iconY + dirWdg->iconHeight < - mwin->TreeViewY
      )
    return True;
#endif

  /* delete selection box */
  x1 = dirWdg->iconX - 1;
  x2 = dirWdg->iconWidth + dirWdg->nameWidth + DefTrSpaceBetwIconAndName + 1;
  y1 = min (dirWdg->iconY, dirWdg->nameY) - 1;
  y2 = max (dirWdg->iconHeight, dirWdg->nameHeight) + 1;

#if 1
  /* if icon selcted, and having no focus, delete box on the selected icon*/
  SET_FG (display, mwin->gc, mwin->treeBg);
  XDrawRectangle (display,
		  mwin->TreeWindow,
		  mwin->gc,
		  x1,
		  y1,
		  x2,
		  y2);
#endif
  
  /* display icon */
  SET_FONT (display, mwin->gc, mwin->fntMainwin->fid);
  SET_FG (display, mwin->gc, mwin->treeFg);
  SET_BG (display, mwin->gc, mwin->treeBg);

  /* Display Tree Lines */
  for (i=0; i<dirTreeLevel (dirWdg->dirTreeEntry); i++)
    {
      long x, y, x2, y2;

      /* first draw the Lines for the Icon */
      if (i == 0)
	{
	  int HaveSubDir = dirTreeHaveSubDir (dirWdg->dirTreeEntry);
	  int Divided    = (HaveSubDir ? 4 : 2);

	  /* draw vertical lines */
	  x = x2 = (dirWdg->iconX
		    - (i * dirWdg->iconWidth)
		    - (dirWdg->iconWidth) / 2);
	  y2 = (dirWdg->iconY
		+ dirWdg->iconHeight / Divided);
	  XDrawLine (display,
		     mwin->TreeWindow,
		     mwin->gc,
		     x,
		     min (dirWdg->iconY, dirWdg->nameY) - DefTrSpaceBetwLines,
		     x2,
		     y2);
	    
	  /* if last item on the tree, don't drow the line */
	  if (! dirTreeLastIcon (dirWdg->dirTreeEntry))
	    {
	      y  = (dirWdg->iconY + dirWdg->iconHeight
		    - (dirWdg->iconHeight) / Divided);
	      y2 = min (dirWdg->nameY, dirWdg->iconY)
		 + max (dirWdg->nameHeight, dirWdg->iconHeight);
	      XDrawLine (display,
			 mwin->TreeWindow,
			 mwin->gc,
			 x,
			 y,
			 x2,
			 y2);
	    }

	  /* draw horizontal line */
	  x = (dirWdg->iconX
		    - (i * dirWdg->iconWidth)
		    - dirWdg->iconWidth / Divided);
	  y = dirWdg->iconY + dirWdg->iconHeight / 2;
	  XDrawLine (display,
		     mwin->TreeWindow,
		     mwin->gc,
		     x,
		     dirWdg->iconY + dirWdg->iconHeight / 2,
		     dirWdg->iconX,
		     dirWdg->iconY + dirWdg->iconHeight / 2);

	  /* if have Subdirectories then draw the click on box */
	  if (HaveSubDir)
	    {
	      x = dirWdg->iconX
		- dirWdg->iconWidth
		+ dirWdg->iconWidth / 4;
	      x2 = dirWdg->iconX - dirWdg->iconWidth / 4;
	      y  = dirWdg->iconY + dirWdg->iconHeight / 4;
	      y2 = dirWdg->iconY
		 + dirWdg->iconHeight
		 - dirWdg->iconHeight / 4;
	      XDrawRectangle (display,
			      mwin->TreeWindow,
			      mwin->gc,
			      x,
			      y,
			      dirWdg->iconWidth / 2,
			      dirWdg->iconHeight / 2);

	      /* draw first minus line */
	      y = dirWdg->iconY + dirWdg->iconHeight / 2;
	      XDrawLine (display,
			 mwin->TreeWindow,
			 mwin->gc,
			 x+2,
			 y,
			 x2-2,
			 y);
	      
	      /* draw the vertical line, if the line isn */
	      if (! dirTreeHaveSubTree (dirWdg->dirTreeEntry))
		{
		  y = dirWdg->iconY + dirWdg->iconHeight / 4;
		  x = dirWdg->iconX
		    - dirWdg->iconWidth / 2;
		  XDrawLine (display,
			     mwin->TreeWindow,
			     mwin->gc,
			     x,
			     y+2,
			     x,
			     y2-2);
		}
	    }
	}
      /* Draw the rest of vertical lines on the left side */
      else
	{
	  if (! dirTreeVLineAvail (dirWdg->dirTreeEntry, i))
	    continue;

	  x = (dirWdg->iconX
	       - (i * dirWdg->iconWidth)
	       - dirWdg->iconWidth/ 2);
	  x2 = (dirWdg->iconX
		- (i * dirWdg->iconWidth)
		- dirWdg->iconWidth/ 2);
	  y2 = min (dirWdg->iconY, dirWdg->nameY)
	     + max (dirWdg->iconHeight, dirWdg->nameHeight);
	  XDrawLine (display,
		     mwin->TreeWindow,
		     mwin->gc,
		     x,
		     min (dirWdg->iconY, dirWdg->nameY) - DefTrSpaceBetwLines,
		     x2,
		     y2);
	}
    }

  SET_FG (display, mwin->gc, sel ? mwin->treeBg : mwin->treeFg);
  SET_BG (display, mwin->gc, sel ? mwin->treeFg : mwin->treeBg);

#ifdef HAVE_XPM
  XCopyArea (display,
	     mwin->pixmList [sel ? TREE_OPENDIR : TREE_DIR],
	     mwin->TreeWindow,
	     mwin->gc,
	     0, 0,
	     dirWdg->iconWidth,
	     dirWdg->iconHeight,
	     dirWdg->iconX,
	     dirWdg->iconY);
#else
  XCopyPlane (display,
	      mwin->pixmList [sel ? TREE_OPENDIR : TREE_DIR],
	      mwin->TreeWindow,
	      mwin->gc,
	      0, 0,
	      dirWdg->iconWidth,
	      dirWdg->iconHeight,
	      dirWdg->iconX,
	      dirWdg->iconY,
	      1L /* Depth = 1 Bit */
	      );
#endif
  
#if 1
  XDrawImageString
#else
  XDrawString
#endif
    (display,
     mwin->TreeWindow,
     mwin->gc,
     dirWdg->nameX,
     dirWdg->nameY + mwin->fntMainwin->ascent,
     dirTreeName (dirWdg->dirTreeEntry),
     strlen (dirTreeName (dirWdg->dirTreeEntry)));

  /* if having focus */
  if (sel && mwin->TreeFocus)
    {
      SET_FG (display, mwin->gc, mwin->treeHighLight);
      XDrawRectangle (display,
		      mwin->TreeWindow,
		      mwin->gc,
		      x1,
		      y1,
		      x2,
		      y2);
    }
  
  return True;
}

int mwinPointsToTreeWidget (Mainwin_t mwin, int i,
			int xx1, int yy1, int xx2, int yy2)
{
  DirTreeWidgets_t dirWdg = mwin->dirTreeWidgets[i];
  int
    x1 = min (xx1, xx2),
    y1 = min (yy1, yy2),
    x2 = max (xx1, xx2),
    y2 = max (yy1, yy2);
  return 
    (
#if 0
     /* icon clicked */
     dirWdg->iconX <= x1 &&	  
     dirWdg->iconY <= y1 &&	  
     dirWdg->iconX + dirWdg->iconWidth >= x2 &&	  
     dirWdg->iconY + dirWdg->iconHeight >= y2
     
     ||
     
     /* name clicked */
     dirWdg->nameX <= x1 &&
     dirWdg->nameY <= y1 &&	  
     dirWdg->nameX + dirWdg->nameWidth  >= x2 &&
     dirWdg->nameY + dirWdg->nameHeight >= y2
#else
     dirWdg->iconX <= x1 &&	  
     dirWdg->iconY <= y1 &&	  
     dirWdg->nameX + dirWdg->nameWidth  >= x2 &&
     dirWdg->nameY + dirWdg->nameHeight >= y2
#endif
     );
}
  
int mwinTreePress  (Mainwin_t mwin, XEvent *event)
{
  int i;

  DEBUG1 ("mwinTreePress:");

  /* Verify if the Icon already selected */
  for (i=0; i<mwin->TreeCount; i++)
    {
      if (mwinPointsToTreeWidget (mwin, i,
				  mwin->btnPressX, mwin->btnPressY,
				  mwin->btnPressX, mwin->btnPressY))
	{
	  mwinTreeCopySelectIcon (mwin, i, True);
	  return MOTION_COPY;
	}
    }
  return MOTION_SELECT;
}

int mwinSelectTree (Mainwin_t mwin, XEvent *event)
{
  DirTreeWidgets_t dirWdg;
  int i, j;
  int
    x1 = min (mwin->btnPressX, mwin->btnReleaseX),
    y1 = min (mwin->btnPressY, mwin->btnReleaseY),
    x2 = max (mwin->btnPressX, mwin->btnReleaseX),
    y2 = max (mwin->btnPressY, mwin->btnReleaseY);

  /* Cursor does not point to any icon */
  mwinTreeSoftSelectIcon (mwin, mwin->TreeSoftSelNum, False);
  mwinTreeCopySelectIcon (mwin, mwin->TreeCopySelNum, False);

  /* selection */
  if ((x2 - x1) > CLICK_REGION || (y2 - y1) > CLICK_REGION)
    return False;

  for (i=0; i<mwin->TreeCount; i++)
    {
      dirWdg = mwin->dirTreeWidgets[i];
      if (mwinPointsToTreeWidget (mwin, i, x1, y1, x2, y2))
	{
	  /* already selected, redisplay only */
	  if (dirTreeSelected (dirWdg->dirTreeEntry))
	      break;

	  /* open the new direcotory */
	  return
	    mwinChangePathName
	      (mwin, dirTreePathName (dirWdg->dirTreeEntry));
	  break;
	}

      /* if the Click On box clicked */
      {
	/********************
	** fx,fy ----      **
	**      |-+-|      **
	**      ----       **
	**         fx1,fy1 **
	********************/
	int
	  region = 3,
	  fx  = dirWdg->iconX - dirWdg->iconWidth + dirWdg->iconWidth / 4
	    - region,
	  fx1 = dirWdg->iconX - dirWdg->iconWidth / 4 + region,
	  fy  = dirWdg->iconY + dirWdg->iconHeight / 4 - region,
	  fy1 = dirWdg->iconY + dirWdg->iconHeight - dirWdg->iconHeight / 4
	    + region;
	if (	 
	    fx  <= x1 &&	  
	    fy  <= y1 &&	  
	    fx1 >= x2 &&	  
	    fy1 >= y2 &&
	    dirTreeHaveSubDir (dirWdg->dirTreeEntry))
	    {
	      mwinTreeViewHide (mwin, i);
	      break;
	    }
	}
    }
  
  return True;
}

int mwinTreeSelectIcon (Mainwin_t mwin, int num, int mode)
{
  int i;

  if (num >= mwin->TreeCount)
    return False;

  mwinTreeSoftSelectIcon (mwin, mwin->CurrentTreeNum, False);
  mwinTreeCopySelectIcon (mwin, mwin->CurrentTreeNum, False);

  /* unselect other selected items */
  if (mode && mwin->CurrentTreeNum != num)
    {
      if (mwin->CurrentTreeNum < mwin->TreeCount)
	{
	  dirTreeSelect (mwin->dirTreeWidgets
			 [mwin->CurrentTreeNum]->dirTreeEntry,
			 False);
	  mwinDispTreeIcon  (mwin, mwin->CurrentTreeNum, False);
	}
      mwin->CurrentTreeNum = num;
    }
  
  /* select or unselect an icon, maximal on Item can be selected */
  if (dirTreeSelected (mwin->dirTreeWidgets[num]->dirTreeEntry) != mode)
    {
      dirTreeSelect (mwin->dirTreeWidgets[num]->dirTreeEntry, mode);

      /* Display the selected item */
      mwinDispTreeIcon  (mwin, num, False);
    }
  
  return True;
}

int mwinTreeSoftSelectIcon (Mainwin_t mwin, int num, int mode)
{
  int i;

  if (num >= mwin->TreeCount)
    return False;

  /* unselect other selected items */
  if (mode && mwin->TreeSoftSelNum != num)
    {
      if (mwin->TreeSoftSelNum < mwin->TreeCount)
	{
	  dirTreeSoftSelect (mwin->dirTreeWidgets
			     [mwin->TreeSoftSelNum]->dirTreeEntry,
			     False);
	  mwinDispTreeIcon  (mwin, mwin->TreeSoftSelNum, False);
	}
      mwin->TreeSoftSelNum = num;
    }
  
  /* soft select or unselect an icon, maximal on Item can be selected */
  if (dirTreeSoftSelected (mwin->dirTreeWidgets[num]->dirTreeEntry) != mode)
    {
      dirTreeSoftSelect (mwin->dirTreeWidgets[num]->dirTreeEntry, mode);

      /* Display the soft selected item */
      mwinDispTreeIcon  (mwin, num, False);
    }
  
  return True;
}

int mwinTreeCopySelectIcon (Mainwin_t mwin, int num, int mode)
{
  int i;

  if (num >= mwin->TreeCount)
    return False;

  /* unselect other selected items */
  if (mode && mwin->TreeCopySelNum != num)
    {
      if (mwin->TreeCopySelNum < mwin->TreeCount)
	{
	  dirTreeCopySelect (mwin->dirTreeWidgets
			     [mwin->TreeCopySelNum]->dirTreeEntry,
			     False);
	  mwinDispTreeIcon  (mwin, mwin->TreeCopySelNum, False);
	}
      mwin->TreeCopySelNum = num;
    }

  /* soft select or unselect an icon, maximal on Item can be selected */
  if (dirTreeCopySelected (mwin->dirTreeWidgets[num]->dirTreeEntry) != mode)
    {
      dirTreeCopySelect (mwin->dirTreeWidgets[num]->dirTreeEntry, mode);

      /* Display the soft selected item */
      mwinDispTreeIcon  (mwin, num, False);
    }
  
  return True;
}

int mwinExecTree (Mainwin_t mwin, XEvent *event)
{
  DirTree_t dirTr;
  DirTreeWidgets_t dirWdg;
  int i;
  for (i=0; i<mwin->TreeCount; i++)
    {
      if (dirTreeSelected (mwin->dirTreeWidgets[i]->dirTreeEntry))
	break;
    }
  if (i > mwin->TreeCount)
    return False;

  /* if the pointer doesn't point to selected */
  if (event != NULL && ! mwinPointsToTreeWidget (mwin, i,
						 event->xbutton.x,
						 event->xbutton.y,
						 event->xbutton.x,
						 event->xbutton.y))
    return False;

  BaisyCursor(mwin->shlMainwin);

  dirWdg = mwin->dirTreeWidgets[i];
  dirTr  = UdchClient (dirWdg->dirTreeEntry);

#if 0
  DEBUG2 ("mwinExecTree: exec (%s)\n", dirTr->PathName);
#endif

  if (dirTr->subTree != NULL)
    {
      UdchEmptyAndDel (dirTr->subTree, (int (*)(void*))dirTreeDest);
      dirTr->subTree = NULL;
    }
  else
    {
      dirTr->subTree = dirTreeAddSubTree (dirWdg->dirTreeEntry);
      dirTreeRead (dirTr->subTree, mwin->DirName, mwin->DirFilter, False);
    }

  /* refresh the new structur */
  mwinBuildTree  (mwin, TREE_REFRESH);
  mwinTreeRedisp (mwin);

  /* view the kids of the icon on the screen */
  mwinTreeViewKinds (mwin, i);

  NormalCursor(mwin->shlMainwin);

#if 0
  DEBUG1 ("mwinExecTree: End");
#endif

  return True;
}

#define XKEY_ENTER	36

#define XKEY_HOME	97
#define XKEY_CUP	98
#define XKEY_PUP	99
#define XKEY_CLEFT	100
#define XKEY_CRIGHT	102
#define XKEY_END	103
#define XKEY_CDOWN	104
#define XKEY_PDOWN	105

#define XA	38
#define XB	56
#define XC	54
#define XD	40
#define XE	26
#define XF	41

int mwinTreeKey (Mainwin_t mwin, KeySym key, unsigned int state)
{
  Dchain_t dir;
  int num = mwin->CurrentTreeNum;
  int ChangeDir = False;
  int levels, i;
  int PgLines = max (1,
		     mwin->dimTreeHeight
		     / max (mwin->dirTreeWidgets[0]->nameHeight,
			    mwin->dirTreeWidgets[0]->iconHeight));

  DEBUG2 ("mwinTreeKey: key=(%i)\n", key);

  switch (key)
    {
    case XK_Up:
      if (mwin->CurrentTreeNum <= 0) return False;
      num --;
      ChangeDir = True;
      break;

    case XK_Down:
      if (num == mwin->TreeCount - 1) return False;
      num ++;
      ChangeDir = True;
      break;

      /* one level upper */
    case XK_Left:
      dir = mwin->dirTreeWidgets[num]->dirTreeEntry;
      for (levels=0; dirGetName (mwin->DirName, levels) != NULL; levels++);
      if (levels <= 0) break;
      strcpy (mwin->DirName, dirname (mwin->DirName));
      mwinTreeSelPath (mwin);
      num = mwin->CurrentTreeNum;
      ChangeDir = True;
      break;

      /* one level right */
    case XK_Right:
      dir = mwin->dirTreeWidgets[num]->dirTreeEntry;
      for (levels=0; dirGetName (mwin->DirName, levels) != NULL; levels++);
      if (levels == 0) break;

      for (i=num; i<mwin->TreeCount; i++)
	{
	  dir = mwin->dirTreeWidgets[i]->dirTreeEntry;
	  if (dirGetName (dirTreePathName (dir), levels))
	    break;
	}
      if (i>=mwin->TreeCount)
	break;

      strcpy (mwin->DirName, dirTreePathName (dir));
      mwinTreeSelPath (mwin);
      num = mwin->CurrentTreeNum;
      ChangeDir = True;
      break;

    case XK_Prior:
      if (num <= 0) return False;
      num = max (0, num - 1 - PgLines);
      ChangeDir = True;
      break;

    case XK_Next:
      if (num == mwin->TreeCount - 1) return False;
      num = min (mwin->TreeCount - 1,
		 num - 1 + PgLines);
      ChangeDir = True;
      break;

    case XK_Home:
      if (num == 0) return False;
      num = 0;
      ChangeDir = True;
      break;

    case XK_End:
      if (num == mwin->TreeCount - 1) return False;
      num = mwin->TreeCount - 1;
      ChangeDir = True;
      break;

    case XK_Return:
      mwinExecTree (mwin, NULL);
      break;
    }

  /* printable key pressed, search for entry with this character */
  if (! (state & ControlMask) && key >= 32 && key <= 256)
    {
      int i;
      for (i=num+1;
	   i<mwin->TreeCount;
	   i++)
	{
	  char * p;
	  p = dirName (mwin->dirTreeWidgets[i]->dirTreeEntry);
	  if ((int) p[0] == (int) key)
	    {
	      num = i;
	      ChangeDir = True;
	      break;
	    }
	}
      /* search from beginning */
      if (i==mwin->TreeCount)
	{
	  for (i=0;
	       i<num;
	       i++)
	    {
	      char * p;
	      p = dirName (mwin->dirTreeWidgets[i]->dirTreeEntry);
	      if ((int) p[0] == (int) key)
		{
		  num = i;
		  ChangeDir = True;
		  break;
		}
	    }
	}
      if (! ChangeDir)
	XBell (display, 0);
	    
    }

  if (ChangeDir)
    {
      return
	mwinChangePathName
	(mwin,
	 dirTreePathName (mwin->
			  dirTreeWidgets [num]->
			  dirTreeEntry));
    }

  return True;
}

int mwinTreeViewKinds (Mainwin_t mwin, int index)
{
  DirTree_t dirTr;

  dirTr = UdchClient (MW_TENTRY(mwin, index));
  
  /* scroll down to view the subitems of the (opened) tree item */
  if (dirTr->subTree != NULL &&UdchNext (dirTr->subTree) != NULL)
    {
      int i, firsty, lasty;
      Dchain_t dch;
      /* find the last item in the subtree list */
      for (dch = UdchNext (dirTr->subTree);
	   dch != NULL;
	   dch = UdchNext (dch))
	{
	  if (UdchNext (dch) == NULL)
	    break;
	}
      for (i=0; i<mwin->TreeCount; i++)
	{
	  if (mwin->dirTreeWidgets[i]->dirTreeEntry == dch )
	    break;
	}
      if (i>mwin->TreeCount)
	return False;
#if 0
      lasty  = max (mwin->dirTreeWidgets[i]->iconY + 
		    mwin->dirTreeWidgets[i]->iconHeight,
		    mwin->dirTreeWidgets[i]->nameY +
		    mwin->dirTreeWidgets[i]->nameHeight);
#else
      lasty = max (MW_TICON_Y2(mwin,i), MW_TNAME_Y2(mwin,i));
#endif
      
#if 1
      firsty = min (MW_TICON_Y(mwin,index), MW_TNAME_Y(mwin,index));
#else
      firsty = min (mwin->dirTreeWidgets[index]->iconY,
		    mwin->dirTreeWidgets[index]->nameY);
#endif
      if (firsty > - mwin->TreeViewY &&
	  lasty > - mwin->TreeViewY + mwin->dimTreeHeight)
	scrAddValue (mwin->scrVertTree,
		     (min (firsty
			   - max (MW_TICON_H(mwin,index),
				  MW_TNAME_H(mwin,index))
			   + mwin->TreeViewY,
			   lasty - firsty - max (MW_TICON_H(mwin,index),
						 MW_TNAME_H(mwin,index))
			   + (mwin->PixmTreeHeight+1)
			   ))
		     / (mwin->PixmTreeHeight+2),
		     SCR_DOWN|SCR_REDRAW);
      
    }
  return True;
}

int mwinTreeViewHide (Mainwin_t mwin, int index)
{
  DirTree_t dirTr;
  DirTreeWidgets_t dirWdg, selectedWdg;
  int i, Test = True;

  if (index > mwin->TreeCount)
    return False;

  BaisyCursor (mwin->shlMainwin);

  dirWdg = mwin->dirTreeWidgets[index];
  dirTr  = UdchClient (dirWdg->dirTreeEntry);
  
#if 0
  DEBUG2 ("mwinTreeViewHide: View/Hide (%s)\n", dirTr->PathName);
#endif

  /* close subtree */
  if (dirTr->subTree != NULL)
    {
      if (StringBegins (mwin->DirName, dirTr->PathName) == 0)
	{
	  mwinChangePathName (mwin, dirTr->PathName);
	}
      UdchEmptyAndDel (dirTr->subTree, (int (*)(void*))dirTreeDest);
      dirTr->subTree = NULL;
      Test = False;
    }
  /* open subtree */
  else
    {
      dirTr->subTree = dirTreeAddSubTree (dirWdg->dirTreeEntry);
      dirTreeRead (dirTr->subTree, dirTr->PathName, mwin->DirFilter, False);
    }

  /* refresh the new structur */
  mwinBuildTree  (mwin, TREE_REFRESH | BUILD_DONOT_CHANGE_PATH);
  mwinTreeRedisp (mwin);

  /* view the kinds of the item */
  mwinTreeViewKinds (mwin, index);
  
  /* Select original selected */
  if (Test)
  {
    for (i=0; i<mwin->TreeCount; i++)
    {
      if (dirTreeSelected (mwin->dirTreeWidgets[i]->dirTreeEntry))
      {
        mwin->CurrentTreeNum = i;
        break;
      }
    }
  }
  
  mwinTreeSelectIcon (mwin, mwin->CurrentTreeNum, True);

  NormalCursor (mwin->shlMainwin);

#if 0
  DEBUG1 ("mwinExecTree: End");
#endif

  return True;
}

int mwinTreeCanCopy_x (Mainwin_t mwin, int x, int y, int mode)
{
  static char * data = NULL;
  static int len = 0;
  char * p;
  int i, j, cnt;
  int ret = -1;

  DEBUG1 ("mwinTreeCanCopy:");

  if (data != NULL)
    {
      MFree (data, len+1);
      data = NULL;
    }

  /* if copy command not created */
  if (mode && mwin->Copy == NULL)
    mwin->Copy = cpyMake ();

  /* get file list */
  len = GetDataFromClipboard (&data);

  if (len <= 1)
    return False;

  /* count the files in the clipboard */
  for (cnt=i=0, p = data; p[0] != 0; )
    {
      i = StrChrIndex (p, ' ');
      if (p[0] != ' ') cnt ++;
      if (i<0) break;
      p += i + 1;
    }
  if (cnt == 1 && data[len-1] == ' ')
    data [len-1] = 0;

  if (cnt < 1)
    return False;

  DEBUG3 ("mwinTreeCanCopy: len(%i), cnt(%i)\n", len, cnt);

  /* if x or y point outside the window */
  if (x < 0 || y < 0)
    return False;

  for (i=0; i<mwin->TreeCount; i++)
    {
      if (mwinPointsToTreeWidget (mwin, i, x, y, x, y))
	{
	  mwinTreeSoftSelectIcon (mwin, i, True);

	  /* 1. Fall: the TreeIcon is already copy selected */
	  if (dirTreeCopySelected (mwin->dirTreeWidgets[i]->dirTreeEntry))
	    {
	      ret = False;
	      break;
	    }

	  /* Save the Destination name in the copy structure */
	  if (mode)
	    {
	      strcpy (mwin->Copy->Dest,
		      dirTreePathName (mwin->dirTreeWidgets[i]->dirTreeEntry));
	      /* if there is only one file to copy, it well copy */
	      /* the file/directory to destination with the same */
	      /* name */
	      if (cnt == 1)
		{
		  strcpy (mwin->Copy->Dest+strlen(mwin->Copy->Dest),
			  "/");
		  strcpy (mwin->Copy->Dest+strlen(mwin->Copy->Dest),
			  basename (data));
		}

	    }
	  ret = True;
	  break;
	}
    }

  if (ret == -1)
    {
      /* Cursor does not point to any icon */
      mwinTreeSoftSelectIcon (mwin, mwin->TreeSoftSelNum, False);
      ret = False;
    }

  /* another situations */
  return ret;
}

int mwinTreeCanCopy (Mainwin_t mwin, int x, int y)
{
  return mwinTreeCanCopy_x (mwin, x, y, False);
}

int mwinTreeCopy (Mainwin_t mwin, int x, int y)
{
  static char * files = NULL;
  static int len = 0;
  char * data;
  int ret, i;

  DEBUG1 ("mwinTreeCopy:");

  if (files != NULL)
    {
      MFree (files, len);
      files = NULL;
    }

  /* view copy window */
  ret = mwinTreeCanCopy_x (mwin, x, y, True);

  /* copy can not proceed */
  if (!ret)
    return False;

  /* Remove old values */
  mwin->Copy->Source = StrReplace (mwin->Copy->Source, NULL);

  /* get filelist from Clipboard */
  len = GetDataFromClipboard (&mwin->Copy->Source);
  
  if (len <= 0)
    return False;

  ret = cpyView (mwin->shlMainwin, mwin->Copy);

  DEBUG1 ("mwinTreeCopy: End");

  return ret;
}

int mwinSetScrollBars (Mainwin_t mwin)
{
  /* set scollbar sizes for the new value */
  if (mwin->scrVertTree != NULL && mwin->dimTreeWidth > 0)
    {
      scrSetValues (mwin->scrVertTree,
		    &mwin->TreeViewY,
		    mwin->TreeMaxYPaint
		    /*mwin->TreeCount * (mwin->PixmTreeHeight+2)*/ ,
		    mwin->dimTreeHeight,
		    mwin->PixmTreeHeight+2);

      scrSetValues (mwin->scrHorizTree,
		    &mwin->TreeViewX,
		    mwin->TreeMaxXPaint,
		    mwin->dimTreeWidth,
		    1);
    }
  return True;
}

int mwinTreeRedisp (Mainwin_t mwin)
{
  if (mwin->fomTree == NULL)
    return False;

  mwin->TreeWindow = XtWindow (mwin->fomTree);

#if 1
  DEBUG3 ("mwinTreeRedisp: Resize request width=(%i), height=(%i)\n",
	  mwin->dimTreeWidth,
	  mwin->dimTreeHeight);
#endif

  /*
   * Clear old Entries
   */
  if (mwin->TreeClearIcons)
    {
      XClearArea (display,
		  mwin->TreeWindow,
		  abs (mwin->TreeViewX), abs (mwin->TreeViewY),
		  mwin->dimTreeWidth, mwin->dimTreeHeight,
		  False);
      mwin->TreeClearIcons = False;
    }
  
  /*
   * Fill the DrawingArea with the Directory entries
   */
  mwinTreeDisp (mwin);
  
#if 1
  DEBUG1 ("mwinTreeRedisp: end");
#endif
  return True;
}

int mwinTreeDest (Mainwin_t mwin, int mode)
{
  int i;
  /* free the allocated memory */
  if (mwin->dirTreeWidgets != NULL)
    {
      /* unselect current directory to rebuild */
      if (! (mode & BUILD_DONOT_CHANGE_PATH))
	mwinTreeSelectIcon (mwin, mwin->CurrentTreeNum, False);

      for (i=0; i<mwin->TreeCount; i++)
	{
	  MFree (mwin->dirTreeWidgets[i], sizeof (dirTreeWidgets_t));
	}	
      MFree (mwin->dirTreeWidgets,
	     sizeof (DirTreeWidgets_t) * mwin->TreeCount);
      mwin->dirTreeWidgets = NULL;
      mwin->TreeCount = 0;
    }
  return True;
}


