/************************************************************************
 *  (C)opyright 1996/1997 by Khamis Abu El-Komboz                       *
 *  xkhfm 1.0 can be copied and distributed freely for any              *
 *  non-commercial purposes. xkhfm can only be incorporated             *
 *  into commercial software with the permission of the current author. *
 *  Mail: khamis@multix.de                                              *
 ************************************************************************/
/*
 * $Author: khamis $
 * $Id: winListMake.c,v 1.2 1997/01/15 01:20:51 khamis Exp khamis $
 * $Source: /usr2/xkhfm/xkhlib/xkhsrc/RCS/winListMake.c,v $
 */
#define IS_EXTERN
#define EXTERN		extern

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include "Xinc.h"

#include "defines.h"
#include "Ustrutil.h"
#include "window.h"

/* static */
static int BaisyState = False;

static void * (* xClMake) (void);
static int    (* xClDest) (void*);
static char * (* xClName) (void*);
static Widget (* xClShell)(void*);
static int xType;

WindowList_t winListMakeX (char * Name, unsigned short type)
{
  WindowList_t l = gcalloc (sizeof (windowList_t), __LINE__, __FILE__);

  /****** Strukturfelder auf Defaultwerte festlegen **************************/
  l->Name       = StrReplace (l->Name, Name != NULL ? Name : "mainwin");
  l->x          = -1;
  l->y          = -1;
  l->width      = -1;
  l->height     = -1;
  l->treeWidth  = -1;
  l->filesWidth = -1;
  l->mnList     = winMnuAdd (NULL, NULL);
  
  l->ClMake     = xClMake;
  l->ClDest     = xClDest;
  l->ClName     = xClName;
  l->ClShell    = xClShell;
  
  l->State      = 0x1000;
  
  l->Type       = xType;
  if (l->ClMake != NULL)
    l->wins.obj = l->ClMake();
  else
    l->Type = WIN_TOP;

  return l;
}

WindowList_t winListMake (char * Name)
{
  return winListMakeX (Name, WIN_REGULAR);
}

WindowList_t winListMakeCmd (char * Name)
{
  return winListMakeX (Name, WIN_COMMAND);
}

WindowList_t winListMakeFnd (char * Name)
{
  return winListMakeX (Name, WIN_FIND);
}

WindowList_t winListMakeMkDir (char * Name)
{
  return winListMakeX (Name, WIN_MKDIR);
}

WindowList_t winListMakeCopy (char * Name)
{
  return winListMakeX (Name, WIN_COPY);
}

WindowList_t winListMakeRemove (char * Name)
{
  return winListMakeX (Name, WIN_REMOVE);
}

WindowList_t winListMakeProp (char * Name)
{
  return winListMakeX (Name, WIN_PROP);
}

Widget winGetShellFromList (Dchain_t l)
{
  WindowList_t lst;
  
  if (l == NULL)
    return NULL;
  
  lst = UdchClient (l);

  if (lst->ClShell != NULL)
    return lst->ClShell (lst->wins.obj);
  else
    return NULL;
}

int winListDest (WindowList_t lst)
{
  Dchain_t win;

  if (lst == NULL)
    return False;
    
  win = winListFind (winList, lst->wins.obj);
  
  /* window doesn't exist (already deleted?) */
  if (win == NULL)
    return True;
    
  /* delete this window from menu of others */
  winDelWinMenu (win);

  if (lst->ClDest != NULL)
  {
    if (! lst->ClDest (lst->wins.obj))
      return False;
  }

  /* delete menu entries */
  winMnuRemove (lst->mnList);

  StrReplace (lst->Name, NULL);

  MFree (lst, sizeof (windowList_t));
  return True;
}

int winListComp (WindowList_t liste, char * Name)
{
  return strcmp (liste->Name, Name);
}

int winListCmpWin (WindowList_t liste, void *win)
{
  if (liste->wins.obj == win)
    return 0;
  if (liste->wins.obj < win)
    return -1;
  return 1;
}

Dchain_t winListFind (Dchain_t winList, void *win)
{
  return UdchFind (winList, win, (int (*)(void*,void*)) winListCmpWin);
}

int winListCount (Dchain_t top, int type)
{
  Dchain_t win;
  int cnt = 0;
  for (win=UdchNext (top);
       win != NULL;
       win = UdchNext (win))
    {
      if (((WindowList_t)UdchClient (win))->Type == type || type == 0)
	{
	  cnt ++;
	}
    }
  return cnt;
}

Dchain_t winGetListFromObj (void *obj)
{
  Dchain_t win;
  for (win=UdchNext (winList);
       win != NULL;
       win = UdchNext (win))
    {
      if (((WindowList_t)UdchClient (win))->wins.obj == obj)
	{
	  return win;
	}
    }
  return NULL;
}

Dchain_t winListFindFirst (Dchain_t top, int type)
{
  Dchain_t win = NULL;
  for (win=UdchNext (top);
       win != NULL;
       win = UdchNext (win))
    {
      if (((WindowList_t)UdchClient (win))->Type == type)
	{
	  break;
	}
    }
  return win;
}

Dchain_t winListAdd   (Dchain_t top, char * Name)
{
  xClMake  = (void*(*)(void))   mwinMake;
  xClDest  = (int(*)(void*))    mwinDest;
  xClName  = (char*(*)(void*))  mwinName;
  xClShell = (Widget(*)(void*)) mwinShell;
  xType    = WIN_REGULAR;

  /* create a new entry **/
  return UdchAdd (top, Name,
		  (void * (*) (void *))         winListMake,
		  (int    (*) (void *, void *)) winListComp);
}

/* add command window */
Dchain_t winListAddCmd (Dchain_t top, char * Name)
{
  xClMake  = (void*(*)(void))   cmdMake;
  xClDest  = (int(*)(void*))    cmdDest;
  xClName  = (char*(*)(void*))  cmdName;
  xClShell = (Widget(*)(void*)) cmdShell;
  xType    = WIN_COMMAND;

  /* create a new entry **/
  return UdchAdd (top, Name,
		  (void * (*) (void *))         winListMakeCmd,
		  (int    (*) (void *, void *)) winListComp);
}
Dchain_t winListAddFind (Dchain_t top, char * Name)
{
  xClMake  = (void*(*)(void))   fndMake;
  xClDest  = (int(*)(void*))    fndDest;
  xClName  = (char*(*)(void*))  fndName;
  xClShell = (Widget(*)(void*)) fndShell;
  xType    = WIN_FIND;

  /* create a new entry **/
  return UdchAdd (top, Name,
		  (void * (*) (void *))         winListMakeFnd,
		  (int    (*) (void *, void *)) winListComp);
}

Dchain_t winListAddMkDir (Dchain_t top, char * Name)
{
  xClMake  = (void*(*)(void))   mkdirMake;
  xClDest  = (int(*)(void*))    mkdirDest;
  xClName  = (char*(*)(void*))  mkdirName;
  xClShell = (Widget(*)(void*)) mkdirShell;
  xType    = WIN_MKDIR;

  /* create a new entry **/
  return UdchAdd (top, Name,
		  (void * (*) (void *))         winListMakeMkDir,
		  (int    (*) (void *, void *)) winListComp);
}

Dchain_t winListAddCopy (Dchain_t top, char * Name)
{
  xClMake  = (void*(*)(void))   cpyMake;
  xClDest  = (int(*)(void*))    cpyDest;
  xClName  = (char*(*)(void*))  cpyName;
  xClShell = (Widget(*)(void*)) cpyShell;
  xType    = WIN_COPY;

  /* create a new entry **/
  return UdchAdd (top, Name,
		  (void * (*) (void *))         winListMakeCopy,
		  (int    (*) (void *, void *)) winListComp);
}

Dchain_t winListAddRemove (Dchain_t top, char * Name)
{
  xClMake  = (void*(*)(void))   remMake;;
  xClDest  = (int(*)(void*))    remDest;
  xClName  = (char*(*)(void*))  remName;
  xClShell = (Widget(*)(void*)) remShell;
  xType    = WIN_REMOVE;

  /* create a new entry **/
  return UdchAdd (top, Name,
		  (void * (*) (void *))         winListMakeRemove,
		  (int    (*) (void *, void *)) winListComp);
}

Dchain_t winListAddProp (Dchain_t top, char * Name)
{
  xClMake  = (void*(*)(void))   propMake;;
  xClDest  = (int(*)(void*))    propDest;
  xClName  = (char*(*)(void*))  propName;
  xClShell = (Widget(*)(void*)) propShell;
  xType    = WIN_PROP;

  /* create a new entry **/
  return UdchAdd (top, Name,
		  (void * (*) (void *))         winListMakeProp,
		  (int    (*) (void *, void *)) winListComp);
}

/* functions to manipulate menu entries for all windows */
WinMenu_t winMnuMake (Dchain_t win)
{
  WinMenu_t nm = gcalloc (sizeof (winMenu_t), __LINE__, __FILE__);
  nm->win = (Dchain_t) win;
  return nm;
}
int winMnuDest (WinMenu_t lst)
{
  if (lst != NULL)
    MFree (lst, sizeof (winMenu_t));
  return True;
}
int winMnuComp(WinMenu_t lst, Dchain_t win)
{
  if (lst->win == win)
    return 0;
  return 1;
}

char * winGetName (Dchain_t mw)
{
  static char tmp[MAXPATHLEN];
  WindowList_t win = UdchClient (mw);

  if (win->ClName != NULL)
    strcpy (tmp, win->ClName (win->wins.obj));
  else
    strcpy (tmp, "(unknown)");

  return tmp;
}


int winDelWinMenu (Dchain_t del_win)
{
#ifdef VIEW_WINDOW_LIST
  Dchain_t mw, mnItem;
  WindowList_t win;
  WinMenu_t w_mnu;

  if (del_win == NULL)
    return False;

#if 0
  WARNING2 ("winDelWinMenu:", "");
  WARNING2 ("window [%s]\n", winWin(del_win)->Name);
#endif
  
  for (mw = UdchNext (winList);
       mw != NULL;
       mw = UdchNext (mw))
    {
      win = UdchClient (mw);

      /* delete window list from list */
      if (mw == del_win)
      {
        win->mnPulldown = NULL;
	winMnuEmpty (win->mnList);
      }

      /* Window already deleted and haven't any Widgets */
      if (winGetShellFromList (mw) == NULL)
        continue;

      /* delete entry for this window from other windows */
      if (mw != del_win)
      {
        mnItem = winMnuFind (win->mnList, del_win);
      
	/* delete entry and menu widget */
        if (mnItem != NULL)
          {
            w_mnu = UdchClient (mnItem);
	    if (w_mnu->win == del_win)
	    {
              if (w_mnu->mnuButton != NULL)
              {
                XtDestroyWidget (w_mnu->mnuButton);
	      }
	      w_mnu->mnuButton = NULL;
              winMnuDel (mnItem);
	    }
          }
      }
    }

#if 0
  WARNING1 ("winDelWinMenu: end");
#endif
  
#endif
  return True;
}

void winBuildWinMenuCB (Widget w, Dchain_t win)
{
  Widget shl = winGetShellFromList (win);
  if (shl != NULL)
  {
    XMapRaised (XtDisplay (shl), XtWindow (shl));
  }
}

int winBuildWinMenu (Widget menu_pulldown, void* obj)
{
  Dchain_t mw, mw2, nm, mnLst;
  WindowList_t win, win2;
  Widget menu_button, shlObj;
  Arg args[4];
  char tmp [MAXPATHLEN];
  int n, i;
  
#if 0
  WARNING1 ("winBuildWinMenu:");
#endif
  
  /* add miscellanius callbacks to manage the window events ****/
  if (obj != NULL && winGetListFromObj(obj) != NULL)
  {
    win = winWin (winGetListFromObj(obj));
    if (win != NULL && win->State == 0x1000)
    {
      shlObj = winGetShellFromList (winGetListFromObj (obj));
  
      if (shlObj != NULL)
      {
        /* add readed coordinates from ressource file, if avail. */
        XtAddEventHandler (shlObj, StructureNotifyMask, False,
		     (XtEventHandler) winSetCoordCB,
		     obj);
        /* Merk the state of a window */
        XtAddEventHandler (shlObj, StructureNotifyMask, False,
		     (XtEventHandler) winSetStateCB,
		     obj);
      }
    }
  }
  
#ifdef VIEW_WINDOW_LIST

  /* save pulldown widget for later usage */
  for (mw = UdchNext (winList);
       mw != NULL;
       mw = UdchNext (mw))
    {
      win = winWin (mw);
      if (obj == win->wins.obj)
	{
	  win->mnPulldown = menu_pulldown;
	}
    }

  /* Add the new entry in every (other) window menu "Windwos" */
  for (mw = UdchNext (winList);
       mw != NULL;
       mw = UdchNext (mw))
    {
      win = winWin (mw);
      if (winGetShellFromList (mw) == NULL || win->mnPulldown == NULL)
	continue;

      for (mw2 = UdchNext (winList);
	   mw2 != NULL;
	   mw2 = UdchNext (mw2))
	{
	  WinMenu_t mnItem;
	  win2 = UdchClient (mw2);

	  if (winGetShellFromList (mw2) == NULL)
	    continue;

	  /* get menu entry for this window */
	  mnLst = winMnuFind (win->mnList, mw2);
	  
	  /* if entry already exists, test if there are maked */
	  /* any changes of the window names */
	  if (mnLst != NULL)
	    {
	      XmString lblS = NULL;
              char * lbl = NULL;
	      mnItem = UdchClient (mnLst);
	      
	      if (mnItem->mnuButton == NULL)
	        continue;
	      
	      XtVaGetValues (mnItem->mnuButton, XmNlabelString, &lblS, NULL);

	      /* make sure, that the entry didn't changed */
	      if (lblS != NULL)
		{
		  XmStringGetLtoR (lblS, XmFONTLIST_DEFAULT_TAG, &lbl);
		  XmStringFree (lblS);
		  strcpy (tmp, winGetName (mw2));
		  if (strcmp (tmp, lbl) != 0)
		    {
		      lblS = XmStringCreateSimple (tmp);
		      XtVaSetValues (mnItem->mnuButton,
				     XmNlabelString, lblS,
				     XmNmnemonic, tmp[0],
				     NULL);
		      XmStringFree (lblS);
		    }
		  XtFree (lbl);
		}
	      continue;
	    }

	  /* don't add menu entry for the window itself */
	  if (win->wins.obj == win2->wins.obj)
	    continue;
	  
	  strcpy (tmp, winGetName (mw2));

	  n = 0;
	  XtSetArg (args[n], XmNmnemonic, tmp[0]); n++;
	  XtSetArg (args[n], XmNsensitive, True); n++;
	  menu_button =
	    XmCreatePushButtonGadget (win->mnPulldown,
				      tmp,
				      args, n);
	  if (menu_button == NULL)
	    continue;
	  XtAddCallback (menu_button,
			 XmNactivateCallback,
			 (XtCallbackProc) winBuildWinMenuCB,
			 mw2);
	  XtManageChild (menu_button);

	  nm = winMnuAdd (win->mnList, mw2);
	  ((WinMenu_t)UdchClient(nm))->mnuButton = menu_button;
	}
    }
#endif
  
#if 0
  WARNING1 ("winBuildWinMenu: end");
#endif
  
  return True;
}
 
int xiconize (Widget w)
{
  if (w == NULL)
    return False;
    
  XIconifyWindow (XtDisplay(w), XtWindow(w), 0);
  
  return True;
}

void winIconizeCB (Widget w, void * obj)
{
  Widget shl = winGetShellFromList (winGetListFromObj (obj));
  xiconize (shl);
}

void winIconOtherCB (Widget w, void* obj)
{
  Dchain_t win, nw;
  for (win=UdchNext (winList);
       win != NULL;
       win = UdchNext (win))
    {
      WindowList_t mw = (WindowList_t)UdchClient (win);
      if (winGetShellFromList(win) == NULL)
        continue;
	
      if (mw->wins.obj != obj)
	{
	  xiconize (winGetShellFromList (win));
	}
    }
}

void winRestoreAllCB (Widget w, void *obj, XConfigureEvent * event)
{
  Dchain_t win, topW = NULL;
  Widget objShl;

  if (event->type != MapNotify)
    return;

  BaisyState = True;
    
  XtRemoveEventHandler (
#if 0
                        winGetShellFromList (winGetListFromObj(obj)),
#else
			w,
#endif
  			StructureNotifyMask, False,
			(XtEventHandler) winRestoreAllCB,
			obj);
			
  for (win=UdchNext (winList);
       win != NULL;
       win=UdchNext (win))
    {
      WindowList_t mw = (WindowList_t)UdchClient (win);
      
      if (winGetShellFromList(win) == NULL)
        continue;
      
      /* window was iconized, restore this state */
      if (mw->State1 == WIN_ICONIZED)
      {
        continue;
      }
      
      if (winGetShellFromList (win) != NULL)
        XtMapWidget (winGetShellFromList (win));
      
      if (win == winGetListFromObj(obj))
      	topW = win;
    }
    
    if (topW != NULL && winGetShellFromList(topW) != NULL)
    {
      XMapRaised (XtDisplay (winGetShellFromList(topW)),
                  XtWindow  (winGetShellFromList(topW)));
    }
    
  BaisyState = False;
}

void winIconAllCB (Widget btn, void *obj)
{
  Dchain_t win;

  BaisyState = True;

  XtAddEventHandler (winGetShellFromList (winGetListFromObj(obj)),
  		     StructureNotifyMask, False,
		     (XtEventHandler) winRestoreAllCB,
		     obj);

  for (win=UdchNext (winList);
       win != NULL;
       win=UdchNext (win))
    {
      Widget shl;
      unsigned int state = 0;
      WindowList_t mw = (WindowList_t)UdchClient (win);
      shl = winGetShellFromList(win);
      
      if (shl == NULL)
        continue;
      
      mw->State1 = mw->State;
      
      if (mw->State == WIN_ICONIZED)
        continue;
      
      /* deiconize iconized windows to hide there icons */
      
      if (mw->wins.obj != obj)
	{
	  XtUnmapWidget (shl);
	}
    }
    
  xiconize (winGetShellFromList (winGetListFromObj (obj)));
  
  BaisyState = False;
}
void winCloseOtherCB (Widget w, void* obj)
{
  Dchain_t win, win2;
  for (win = UdchNext (winList);
       win != NULL;
       )
    {
      WindowList_t mw = (WindowList_t)UdchClient (win);
      
      /* search for next regular or command window */
      for (win2 = UdchNext (win);
           win2 != NULL;
	   win2 = UdchNext (win2))
      {
        if (winWin(win2)->Type == WIN_REGULAR || winWin(win2)->Type == WIN_COMMAND)
	  break;
      }
      
      /* delete only command or mainwindows */
      if (mw->Type == WIN_REGULAR || mw->Type == WIN_COMMAND)
      {
        if (mw->wins.obj != obj)
	  {
	    /* delete window from list */
	    winListDel (win);
	  }
      }
      win = win2;
    }
}
void winWindowsCB (Widget w, void *obj)
{
}


/* XEvent Handling *****************/
int winCorrectCoords(void * obj)
{
  Dchain_t winDch;
  WindowList_t win;
  Widget wdg;
  Position x, y;
  Dimension w, h;

  DEBUG1 ("winCorrectCoords:");

  for (winDch=UdchNext (winList);
       winDch != NULL;
       winDch=UdchNext (winDch))
    {
      win = (WindowList_t)UdchClient (winDch);

      if (win->wins.obj != obj)
	continue;

      if (win->x     == -1 &&
	  win->y     == -1 &&
	  win->width == -1 &&
	  win->height== -1)
	break;
      
      wdg = winGetShellFromList (winDch);
      
      if (wdg == NULL)
        break;
	  
      x = max (0, win->x);
      y = max (0, win->y);
      w = max (50, win->width);
      h = max (50, win->height);
      
      if (wdgX (wdg) > x)
	{
	  Position nx = 2 * x - wdgX (wdg);
	  XtVaSetValues (wdg,
			 XmNx, nx,
			 NULL);
	}
      if (wdgY (wdg) > y)
	{
	  Position ny = 2 * y - wdgY (wdg);
	  XtVaSetValues (wdg,
			 XmNy, ny,
			 NULL);
	}
      break;
    }

  DEBUG1 ("winCorrectCoords: End");

  return True;
}

int winSetCoord (void* obj)
{
  Dchain_t winDch;
  WindowList_t win;
  Widget wdg;
  Position x, y;
  Dimension w, h;

  DEBUG1 ("winSetCoord:");

  for (winDch=UdchNext (winList);
       winDch != NULL;
       winDch=UdchNext (winDch))
    {
      Widget shell;
      win = (WindowList_t)UdchClient (winDch);

      if (win->wins.obj == obj)
	{
	  if (win->x     == -1 &&
	      win->y     == -1 &&
	      win->width == -1 &&
	      win->height== -1)
	    break;
	    
          wdg = winGetShellFromList (winDch);
	  if (wdg == NULL)
	    break;
	    
	  x = max (0, win->x);
	  y = max (0, win->y);
	  w = max (50, win->width);
	  h = max (50, win->height);
	  XtVaSetValues (wdg,
		 XmNx,         x,
		 XmNy,         y,
		 XmNwidth,     w,
		 XmNheight,    h,
		 NULL);
	  break;
	}
    }  

  DEBUG1 ("winSetCoord: End");

  return True;
}

void winSetCoordCB (Widget shell, void * obj, XConfigureEvent * event)
{
  XEvent xev;

  if (event->type != ConfigureNotify) 
    return;

  DEBUG1 ("winSetCoordCB:");

  winSetCoord (obj);
  XtRemoveEventHandler (shell, StructureNotifyMask, False,
			(XtEventHandler) winSetCoordCB,
			obj);
  while (XCheckMaskEvent (display,
			  ALL_MASKS,
			  &xev))
    {
      XtDispatchEvent (&xev);
    }
  winCorrectCoords (obj);

  DEBUG1 ("winSetCoordCB: End");

  return;
}

void winSetStateCB (Widget shell, void * obj, XConfigureEvent * event)
{
  Dchain_t winLst;
  WindowList_t win;

  if (BaisyState || event->type != MapNotify && event->type != UnmapNotify)
    return;

  winLst = winGetListFromObj (obj);
  if (winLst == NULL)
     return;
     
  win = winWin (winLst);
  
  switch (event->type)
  {
  case MapNotify:
    win->State = WIN_NORMAL;
    break;

  case UnmapNotify:
    win->State = WIN_ICONIZED;
    break;
  }
}


