/************************************************************************
 *  (C)opyright 1996/1997 by Khamis Abu El-Komboz                       *
 *  xkhfm 1.0 can be copied and distributed freely for any              *
 *  non-commercial purposes. xkhfm can only be incorporated             *
 *  into commercial software with the permission of the current author. *
 *  Mail: khamis@multix.de                                              *
 ************************************************************************/
/*
 * $Author: khamis $
 * $Id: xclipbrd.c,v 1.1 1997/01/07 08:35:29 khamis Exp $
 * $Source: /home/khamis/source/xkhfm/xkhlib/xkhsrc/RCS/xclipbrd.c,v $
 */
#define IS_EXTERN
#define EXTERN		extern

#include <stdio.h>
#include <stdlib.h>
#include <sys/param.h>
#include <string.h>
#include <fcntl.h>
#include "Xinc.h"
#include <X11/Xaw/Paned.h>
#include <X11/Xatom.h>
#include <X11/Xmu/Atoms.h>
#include <X11/Xmu/StdSel.h>
#include <Xm/DrawingA.h>
#include <Xm/ScrollBar.h>
#include <Xm/Command.h>

#include "defines.h"
#include "Umem.h"
#include "Uxutil.h"
#include "Umessage.h"
#include "Umenu.h"
#include "Urcnfvoid.h"
#include "Ureadarg.h"
#include "Ufileutil.h"
#include "Ustrutil.h"
#include "about.h"
#include "quit.h"
#include "help.h"
#include "xmisc.h"
#include "window.h"
#include "mainwin.h"

static char * SelectedFiles = NULL;
static int    SelectedLen = 0;
static int    SelectionCount = 1;
static Atom   SelectionAtoms[1] = {0};
static Time   SelectionTime = 0;
static String SelectionNames[1] = { "XA_PRIMARY" };

int mwinConvertSelection (Widget w,
			  Atom *selection,
			  Atom *target,
			  Atom *type,
			  XtPointer *value,
			  unsigned long *length,
			  int *format);
void mwinSelectionDone (Widget w, Atom *selection, Atom *target);
void mwinLoseSelection (Widget w, Atom *selection);


int mwinOwnSelection (Mainwin_t mwin)
{
    int i, ret;
    Boolean HaveSelection = False;

    SelectedLen = mwinGetSelectedFiles (mwin, &SelectedFiles);

    if (SelectedLen == 0)
      return False;

    WARNING1 ("mwinOwnSelection");

    XmuInternStrings (display,
		      SelectionNames,
		      SelectionCount,
		      SelectionAtoms);
    for (i = 0; i < SelectionCount; i++)
      {
	int cutbuffer;
	switch (SelectionAtoms[i])
	  {
	  case XA_CUT_BUFFER0: cutbuffer = 0; break;
	  case XA_CUT_BUFFER1: cutbuffer = 1; break;
	  case XA_CUT_BUFFER2: cutbuffer = 2; break;
	  case XA_CUT_BUFFER3: cutbuffer = 3; break;
	  case XA_CUT_BUFFER4: cutbuffer = 4; break;
	  case XA_CUT_BUFFER5: cutbuffer = 5; break;
	  case XA_CUT_BUFFER6: cutbuffer = 6; break;
	  case XA_CUT_BUFFER7: cutbuffer = 7; break;
	  default:	       cutbuffer = -1; break;
	  }
	if (cutbuffer >= 0)
	  {
	    WARNING1 ("mwinOwnSelection: XStore");

	    if (SelectedLen > 4*XMaxRequestSize(display)-32)
	      WARNING2 ("mwinOwnSelection: selection too big (%d bytes)\n",
			SelectedLen);
	    else
	      XStoreBuffer (display, SelectedFiles,
			    SelectedLen,
			    cutbuffer);
	  }
	else
	  {
	    WARNING1 ("mwinOwnSelection: XtOwnSelection");

	    HaveSelection |=
	       XtOwnSelection (mwin->shlMainwin,
			       SelectionAtoms[i],
			       SelectionTime,
			       (XtConvertSelectionProc) mwinConvertSelection,
			       mwinLoseSelection,
			       mwinSelectionDone);
	  }
      }

    WARNING1 ("mwinOwnSelection: end");

    return HaveSelection;
}

int mwinConvertSelection (Widget w,
			  Atom *selection,
			  Atom *target,
			  Atom *type,
			  XtPointer *value,
			  unsigned long *length,
			  int *format)
{
  Display* d = XtDisplay(w);

  WARNING1 ("mwinIwnSelection:");

  if (SelectedFiles == NULL || SelectedLen <= 0)
    return False;

  if (*target == XA_TARGETS(d))
    {
      Atom* targetP;
      Atom* std_targets;
      unsigned long std_length;

      WARNING1 ("mwinIwnSelection: XA_TARGETS");

      XmuConvertStandardSelection
	(w,
	 SelectionTime,
	 selection,
	 target, type, (caddr_t*)&std_targets, &std_length, format);
      *length = std_length + 5;
      *value  = XtMalloc (sizeof(Atom)*(*length));
      targetP = *(Atom**)value;
      *targetP++ = XA_STRING;
      *targetP++ = XA_TEXT(d);
      *targetP++ = XA_COMPOUND_TEXT(d);
      *targetP++ = XA_LENGTH(d);
      *targetP++ = XA_LIST_LENGTH(d);
      memmove ((char*)targetP, (char*)std_targets, sizeof(Atom)*std_length);
      XtFree ((char*)std_targets);
      *type = XA_ATOM;
      *format = 32;
      return True;
    }

  if (*target == XA_STRING ||
      *target == XA_TEXT(d) ||
      *target == XA_COMPOUND_TEXT(d))
    {
      WARNING1 ("mwinIwnSelection: XA_STRING");

      if (*target == XA_COMPOUND_TEXT(d))
	*type = *target;
      else
	*type = XA_STRING;
      *value  = SelectedFiles;
      *length = SelectedLen;
      *format = 8;
      return True;
    }
  if (*target == XA_LIST_LENGTH(d))
    {

      WARNING1 ("mwinIwnSelection: XA_LIST");
    
      *value = XtMalloc(4);
      if (sizeof(long) == 4)
	*(long*)*value = 1;
      else
	{
	  long temp = 1;
	  memmove ((char*)*value, ((char*)&temp)+sizeof(long)-4, 4);
	}
      *type = XA_INTEGER;
      *length = 1;
      *format = 32;
      return True;
    }
  if (*target == XA_LENGTH(d))
    {

      WARNING1 ("mwinIwnSelection: XA_LEN");
    
      *value = XtMalloc(4);
      if (sizeof(long) == 4)
	*(long*)*value = SelectedLen;
      else
	{
	  long temp = SelectedLen;
	  memmove ((char*)*value, ((char*)&temp)+sizeof(long)-4, 4);
	}
      *type = XA_INTEGER;
      *length = 1;
      *format = 32;
      return True;
    }
  if (XmuConvertStandardSelection(w,
				  SelectionTime,
				  selection,
				  target, type,
				  (caddr_t *)value, length, format))
    {
      WARNING1 ("mwinIwnSelection: XmuConvertSta");
      return True;
    }
  
  WARNING1 ("mwinIwnSelection: Error !");

  /* else */
  return False;

}

void mwinLoseSelection (Widget w, Atom *selection)
{
    register Atom* atomP;
    int i;

    WARNING1 ("mwinLoseSelection:");

    for (i = 0, atomP = SelectionAtoms;
	 i < SelectionCount; i++, atomP++)
    {
	if (*selection == *atomP) *atomP = (Atom)0;
	switch (*atomP) {
	  case XA_CUT_BUFFER0:
	  case XA_CUT_BUFFER1:
	  case XA_CUT_BUFFER2:
	  case XA_CUT_BUFFER3:
	  case XA_CUT_BUFFER4:
	  case XA_CUT_BUFFER5:
	  case XA_CUT_BUFFER6:
	  case XA_CUT_BUFFER7:	*atomP = (Atom)0;
	}
    }

    for (i = SelectionCount; i>=0; i--) {
	if (SelectionAtoms[i-1] != 0)
	  break;
    }
    SelectionCount = i;

    for (i = 0, atomP = SelectionAtoms;
	 i < SelectionCount; i++, atomP++)
      {
	if (*atomP == (Atom)0)
	  {
	    *atomP = SelectionAtoms [-- SelectionCount];
	  }
      }

    WARNING1 ("mwinLoseSelection: end");
}

void mwinSelectionDone (Widget w, Atom *selection, Atom *target)
{
    WARNING1 ("mwinSelectionDone: end");
    /* empty proc so Intrinsics know we want to keep storage */
}

